#ifndef ClientGUI_H
#define ClientGUI_H

#include "GameObj.H"
#include "ChargeMap.H"
#include <SDL/SDL_thread.h>

using namespace std;

/*
	Struct for handling RGB colors.
*/
struct RGB {
  uint1 R, G, B;
  uint4 col;

  void from(sint4 r, sint4 g, sint4 b=0) {
    R = r; G = g; B = b;
    col = ((((R << 8)+G) << 8) + B);
  }
  
  RGB(sint4 r=0, sint4 g=0, sint4 b=0) : R(r), G(g), B(b) {
    col = ((((R << 8)+G) << 8) + B);    
  }
  
  uint4 color() const { return col; }
  uint4 color_alpha(sint4 alpha=255) const {
    assert(alpha >=0 && alpha < 256);
    return (col << 8) + alpha;
  }  
};

/*
	This class is responsible for displaying a charge map GUI. It is used
	to debug the behaviour of the units.
*/
class ClientGUI {
	public:

		/*
			Constructor.
		*/
		ClientGUI();
		
		/*
			Returns true if the GUI is ready (init finished).
		*/
		bool isReady();

		/*
			Initiates the GUI.
		*/
		void init();

		/*
			Event handler for key events in the GUI window.
		*/
		void event();

		/*
			Updates the ChargeMap.
		*/
		void updateRef(ChargeMap* map);

		/*
			Starts the auto update thread.
		*/
		void startThread();

		/*
			Updates the GUI.
		*/
		void display();

		//True if a quit application event occured
		bool quit;
		//True if a write unit status event occured
		bool writeStatus;

		//Thread run method
		static int execute(void *data);

	protected:

		//Thread
		SDL_Thread *my_thread;
		
	private:

		/*
			Returns a display color for the specified charge.
		*/
		Uint32 getColor(real8 charge);

		//True if the GUI update thread is running
		bool gui_running;

		//The current display mode (see constants below)
		int mode;
		static const int MODE_OFFENSIVE;
		static const int MODE_DEFENSIVE;
		static const int MODE_BASE;
		static const int MODE_OFFENSIVE_UNITS;
		static const int MODE_DEFENSIVE_UNITS;
		static const int MODE_TERRAIN;
		static const int MODE_PLAYER;

		//GUI params
		SDL_Surface *screen;
		sint4 s_x, s_y;
		sint4 gfx_mag_frac;
		sint4 adj_x, adj_y;

		//ChargeMap reference
		ChargeMap* c_map;
		
};

#endif
