#include "TBConfig.H"

using namespace std;

bool config_init = false;

Params TBConfig::params;
string TBConfig::path = "NAN";
string TBConfig::paramsString = "NAN";

/*
	Sets path to the config file to use.
*/
void TBConfig::setConfigPath(string confpath) {
	TBConfig::path = confpath;
	cout << "Using config " << TBConfig::path << endl;
}

/*
	Sets the app params as a string instead of read from file.
*/
void TBConfig::setConfigString(string mParamsString) {
	TBConfig::paramsString = mParamsString;
}

/*
	Returns true if config is initiated, false otherwise.
*/
bool TBConfig::isInitiated() {
	return config_init;
}

/*
	Inits the config by reading a config file and set the
	parameters in the params struct.
*/
void TBConfig::init() {
	//Set default value to params struct
	params.UnitAgent_HPDefensiveLimit = 5;
	params.UnitAgent_LookAheadTiles = 1;
	params.Blackboard_ActionsBeforeClearRestless = 100;
	params.Blackboard_IdleFramesBeforeRestless = 124;
	params.Blackboard_MaxDistBetweenGroupUnits = 30;
	params.Blackboard_MinUnitsInGroup = 10;
	params.Blackboard_HuntUnits_Ratio = 25;
	params.ChargeMap_PathMinThreshold = -50;
	params.ChargeMap_PathMaxThreshold = -1;
	params.ChargeMap_Fill_DistToCheck = 6;
	params.ChargeMap_Fill_MaxBlockedDirections = 9;
	params.Attack_DmgProbability = 0.9;

	if (TBConfig::path != "NAN") {
		cout << "Reading TankBattle config file" << endl;

		if (!TBConfig::path.empty()) {
			//Read config file
			string line;
			ifstream myfile(TBConfig::path.c_str());
			if (myfile.is_open()) {
				while (!myfile.eof()) {
					getline(myfile,line);
					parseLine(line);
				}
				myfile.close();
			}
			else {
				cout << "WARNING! Config file not found! Using default values" << endl; 
			}
		}
		else {
			cout << "WARNING! Config file not set. Using default values" << endl;
		}
	}
	if (TBConfig::paramsString != "NAN") {
		cout << "Setting params from string " << TBConfig::paramsString << endl;

		vector<string> tokens;
		int no = split(TBConfig::paramsString, "#", tokens);
		if (no == 12) {
			params.UnitAgent_HPDefensiveLimit = stringToInt(tokens[0]);
			params.UnitAgent_LookAheadTiles = stringToInt(tokens[1]);
			params.Blackboard_ActionsBeforeClearRestless = stringToInt(tokens[2]);
			params.Blackboard_IdleFramesBeforeRestless = stringToInt(tokens[3]);
			params.Blackboard_MaxDistBetweenGroupUnits = stringToInt(tokens[4]);
			params.Blackboard_MinUnitsInGroup = stringToInt(tokens[5]);
			params.Blackboard_HuntUnits_Ratio = stringToInt(tokens[6]);
			params.ChargeMap_PathMinThreshold = stringToInt(tokens[7]);
			params.ChargeMap_PathMaxThreshold = stringToInt(tokens[8]);
			params.ChargeMap_Fill_DistToCheck = stringToInt(tokens[9]);
			params.ChargeMap_Fill_MaxBlockedDirections = stringToInt(tokens[10]);
			params.Attack_DmgProbability = stringToReal(tokens[11]);
		}
		else {
			cout << "WARNING! Invalid amount of tokens in params string. Found " << no << ", expected 16. Using default values" << endl;
		}
	}

	config_init = true;

	//Debug. Comment out if not needed
	/*cout << "Config params to use:" << endl;
	cout << "UnitAgent_HPDefensiveLimit=" << params.UnitAgent_HPDefensiveLimit << endl;
	cout << "UnitAgent_LookAheadTiles=" << params.UnitAgent_LookAheadTiles << endl;
	cout << "Blackboard_ActionsBeforeClearRestless=" << params.Blackboard_ActionsBeforeClearRestless << endl;
	cout << "Blackboard_IdleFramesBeforeRestless=" << params.Blackboard_IdleFramesBeforeRestless << endl;
	cout << "Blackboard_MaxDistBetweenGroupUnits=" << params.Blackboard_MaxDistBetweenGroupUnits << endl;
	cout << "Blackboard_MinUnitsInGroup=" << params.Blackboard_MinUnitsInGroup << endl;
	cout << "Blackboard_HuntUnits_Ratio=" << params.Blackboard_HuntUnits_Ratio << endl;
	cout << "ChargeMap_PathMinThreshold=" << params.ChargeMap_PathMinThreshold << endl;
	cout << "ChargeMap_PathMaxThreshold=" << params.ChargeMap_PathMaxThreshold << endl;
	cout << "ChargeMap_Fill_DistToCheck=" << params.ChargeMap_Fill_DistToCheck << endl;
	cout << "ChargeMap_Fill_MaxBlockedDirections=" << params.ChargeMap_Fill_MaxBlockedDirections << endl;
	cout << "Attack_DmgProbability=" << params.Attack_DmgProbability << endl;
	//*/

	cout << "Done reading config file" << endl;
}

/*
	Parses a line from the config file and, if it is a parameter line,
	sets the value in the params struct.
*/
void TBConfig::parseLine(string line) {
	if (line.find("#") != 0 && line.length() > 2) {
		vector<string> tokens;
		int no = split(line, "=", tokens);

		if (no == 2) {
			//Set param values to struct
			if (line.find("UnitAgent_HPDefensiveLimit") == 0) {
				params.UnitAgent_HPDefensiveLimit = stringToReal(tokens[1]);
			}
			else if (line.find("UnitAgent_LookAheadTiles") == 0) {
				params.UnitAgent_LookAheadTiles = stringToInt(tokens[1]);
			}
			else if (line.find("Blackboard_IdleFramesBeforeRestless") == 0) {
				params.Blackboard_IdleFramesBeforeRestless = stringToInt(tokens[1]);
			}
			else if (line.find("Blackboard_ActionsBeforeClearRestless") == 0) {
				params.Blackboard_ActionsBeforeClearRestless = stringToInt(tokens[1]);
			}
			else if (line.find("Blackboard_MaxDistBetweenGroupUnits") == 0) {
				params.Blackboard_MaxDistBetweenGroupUnits = stringToInt(tokens[1]);
			}
			else if (line.find("Blackboard_MinUnitsInGroup") == 0) {
				params.Blackboard_MinUnitsInGroup = stringToInt(tokens[1]);
			}
			else if (line.find("Blackboard_HuntUnits_Ratio") == 0) {
				params.Blackboard_HuntUnits_Ratio = stringToInt(tokens[1]);
			}
			else if (line.find("ChargeMap_PathMinThreshold") == 0) {
				params.ChargeMap_PathMinThreshold = stringToReal(tokens[1]);
			}
			else if (line.find("ChargeMap_PathMaxThreshold") == 0) {
				params.ChargeMap_PathMaxThreshold = stringToReal(tokens[1]);
			}
			else if (line.find("ChargeMap_Fill_DistToCheck") == 0) {
				params.ChargeMap_Fill_DistToCheck = stringToInt(tokens[1]);
			}
			else if (line.find("ChargeMap_Fill_MaxBlockedDirections") == 0) {
				params.ChargeMap_Fill_MaxBlockedDirections = stringToInt(tokens[1]);
			}
			else if (line.find("Attack_DmgProbability") == 0) {
				params.Attack_DmgProbability = stringToReal(tokens[1]);
			}
			else {
				cout << "Unknown config parameter: " << tokens[0] << endl;
			}
		}
		else {
			cout << "Invalid format for parameter: " << line << endl;
		}
	}
}

/*
	Converts a string to a real value.
	Returns 0 if string is not a valid real.
*/
real8 TBConfig::stringToReal(string str) {
	real8 value(0);
	std::istringstream iss(str);

	if (iss >> value) {
		return value;
	}
	else {
		return 0;
	}
}

/*
	Converts a string to an int value.
	Returns 0 if string is not a valid int.
*/
int TBConfig::stringToInt(string str) {
	int value(0);
	std::istringstream iss(str);

	if (iss >> value) {
		return value;
	}
	else {
		return 0;
	}
}

/*
	Converts a string to a bool. Returns false
	if not a valid string.
*/
bool TBConfig::stringToBool(string str) {
	if (str.find("true") != string::npos) return true;
	if (str.find("TRUE") != string::npos) return true;
	if (str.find("1") != string::npos) return true;
	return false;
}

/*
	Splits a string using a specified delimiter. The tokens are stored in a
	results vector.
*/
int TBConfig::split(const std::string &input, const std::string &delimiter, std::vector<string> &results) {
    // find the start character of the next token.

	int offset = 0;
	int found = 0;
	bool end = false;

	while (!end) {
		size_t i = input.find_first_not_of(delimiter, offset);
		if (i == string::npos) {
			//No more tokens. End.
			offset = input.length();
			end = true;
			break;
		}

		// find the end of the token.
		size_t j = input.find_first_of(delimiter, i);
		if (j == string::npos) {
			string token = input.substr(i);
			offset = input.length();
			
			if (token.length() > 0) {
				results.push_back(token);
				found++;
			}
		}
		else {
			// to intercept the token and save current position
			string token = input.substr(i, j - i);
			offset = j;
			
			if (token.length() > 0) {
				results.push_back(token);
				found++;
			}
		}
	}

	return found;
}
