#ifndef TBConfig_H
#define TBConfig_H

#include "GameObj.H"

using namespace std;

/*
	Struct containing the changeable parameters.
*/
struct Params {
	real8 UnitAgent_HPDefensiveLimit;
	int UnitAgent_LookAheadTiles;
	int Blackboard_IdleFramesBeforeRestless;
	int Blackboard_ActionsBeforeClearRestless;
	int Blackboard_MaxDistBetweenGroupUnits;
	int Blackboard_MinUnitsInGroup;
	int Blackboard_HuntUnits_Ratio;
	real8 ChargeMap_PathMinThreshold;
	real8 ChargeMap_PathMaxThreshold;
	int ChargeMap_Fill_DistToCheck;
	int ChargeMap_Fill_MaxBlockedDirections;
	real8 Attack_DmgProbability;
};

/*
	This class reads application parameters from file and stores them in a struct.
	It enables the possibility to change the AI behavior with settings in a config file.
	The path to the config file is passed as parameter when starting the app.
*/
class TBConfig {
	public:

		/*
			Sets path to the config file to use.
		*/
		static void setConfigPath(string confpath);

		/*
			Sets the app params as a string instead of read from file.
		*/
		static void setConfigString(string mParamsString);

		/*
			Returns true if config is initiated, false otherwise.
		*/
		static bool isInitiated();

		/*
			Inits the config by reading a config file and set the
			parameters in the params struct.
		*/
		static void init();

		/*
			Parses a line from the config file and, if it is a parameter line,
			sets the value in the params struct.
		*/
		static void parseLine(string line);

		/*
			Splits a string using a specified delimiter. The tokens are stored in a
			results vector.
		*/
		static int split(const string& input, const string& delimiter, vector<string>& results);

		/*
			Converts a string to a real value.
			Returns 0 if string is not a valid real.
		*/
		static real8 stringToReal(string str);

		/*
			Converts a string to an int value.
			Returns 0 if string is not a valid int.
		*/
		static int stringToInt(string str);

		/*
			Converts a string to a bool. Returns false
			if not a valid string.
		*/
		static bool stringToBool(string str);

		//Parameters struct
		static Params params;

		//Path to config file
		static string path;

		//Params as a string instead of file
		static string paramsString;
	
};

#endif
