/** @file ScriptEmitter.C
    @see ScriptEmitter.H

    $Id: ScriptEmitter.C,v 1.20 2006/08/11 20:54:01 dtom Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/gfxclient/src/ScriptEmitter.C,v $
*/

// This is an ORTS file 
// (c) Tim Furtak
// (c) Erik Benner

#include "Global.H"
#include "ScriptEmitter.H"
#include "ScriptModel.H"
#include "GfxEmitter.H"
#include "GfxModule.H"
#include "GfxAnimesh.H"

using namespace std;

const sint4 ScriptEmitter::STYPE_GFXEMITTER_ID = stype::next_id();

//===================================================================

ScriptEmitter::ScriptEmitter()
{
  add("init_box",       &ScriptEmitter::init_box,       0,26, 0);
  add("init_spray",     &ScriptEmitter::init_spray,     0,23, 0);
  add("init_circle",    &ScriptEmitter::init_circle,     0,23, 0);
  add("init_corona",    &ScriptEmitter::init_corona,    0, 4, 0);
  add("init_fire",      &ScriptEmitter::init_fire,      0, 2, 0);
  add("init_shell",     &ScriptEmitter::init_shell,     0, 2, 0);
  add("init_lightning", &ScriptEmitter::init_lightning, 0, 6, 0);

  add("position",       &ScriptEmitter::position,       0, 3, 0);
  add("rotate",         &ScriptEmitter::rotate,         0, 1, 0);
  add("scale",          &ScriptEmitter::scale,          0, 1, 0);
  add("set_rate",       &ScriptEmitter::set_rate,       0, 1, 0);

  //  add("draw_2d",        &ScriptEmitter::draw_2d,        0, 3, 0);
  //  add("draw_overlay",   &ScriptEmitter::draw_overlay,   0, 5, 0);

  add("set_parent",     &ScriptEmitter::set_parent,     1, 0, 0);

  add("enable",         &ScriptEmitter::enable,         0, 1, 0);
  
  emitter = 0;
}

ScriptEmitter::~ScriptEmitter()
{
}

//-------------------------------------------------------------------

void ScriptEmitter::add(const std::string &name, myfunc f, sint4 objs, sint4 p, sint4 p2)
{
  ObjActionDefinition def(name, objs, p, p2);
  ObjAction *a = new ObjAction(static_cast<ObjFunction>(f), def); // fixme: cast is non-portable
  actions.push_back(a);
  act_map[name] = a;
}

//===================================================================

stype ScriptEmitter::init_box(const SActParam &p)
{
  OBJN_CHECK(0);
  VALN_CHECK(26);

  vector<sint4> v;
  FORU (i, p.val.size()) {
    v.push_back(p.val[i].to_int());
  }
  
  real4 fp = 0.01;

  if (emitter) {
    delete emitter;
    emitter = 0;
  }

  GameObj *gob = this->parent->get_GameObj();
  SCRIPT_ASSERT(gob && "GameObj ptr 0");

  GfxBoxEmitter *ge =
    new GfxBoxEmitter(*GUI_gfxm,
                      v[0],                                     // creation rate
                      fp*v[1],                                  // particle life
                      fp*v[2],                                  // time limit
                      Vec3<real4>(fp*v[3], fp*v[4], fp*v[5]),   // position
                      Vec3<real4>(fp*v[6], fp*v[7], fp*v[8]),   // position randomness
                      Vec3<real4>(fp*v[9], fp*v[10],fp*v[11]),  // velocity
                      Vec3<real4>(fp*v[12],fp*v[13],fp*v[14]),  // velocity randomness
                      Vec3<real4>(fp*v[15],fp*v[16],fp*v[17]),  // accelleration
                      fp*v[18], fp*v[19],                   // friction & randomness
                      fp*v[20], fp*v[21],                   // particle size
                      v[22], v[23]);                        // particle image and color textures

  ge->draw_type = static_cast<GfxPListEmitter::DRAW_TYPE>(v[24]);
  ge->blend = static_cast<GfxEmitter::BlendType>(v[25]);

  GfxGameObject *ggo = GUI_gfxm->game2gfx[gob];
  if (!ggo->gfx_obj) {
    ggo->attach_gfxobj(ge);
    GUI_gfxm->add_GfxObject(ge);
  }

  emitter = ge;
  return stype();
}

//-------------------------------------------------------------------

stype ScriptEmitter::init_spray(const SActParam &p)
{
  OBJN_CHECK(0);
  VALN_CHECK(23);

  vector<sint4> v;
  FORU (i, p.val.size()) {
    v.push_back(p.val[i].to_int());
  }
  
  real4 fp = 0.01;

  if (emitter) {
    delete emitter;
    emitter = 0;
  }

  GameObj *gob = this->parent->get_GameObj();
  SCRIPT_ASSERT(gob && "GameObj ptr 0");

  GfxSprayEmitter *ge =
    new GfxSprayEmitter(*GUI_gfxm,
                        v[0],                                 // creation rate
                        fp*v[1],                              // particle life
                        fp*v[2],                              // time limit
                        Vec3<real4>(fp*v[3],fp*v[4],fp*v[5]), // position
                        v[6], v[7],                           // min, max latitude
                        fp*v[8],  fp*v[9],                    // offset & randomness
                        fp*v[10], fp*v[11],                   // speed & randomness
                        Vec3<real4>(fp*v[12],fp*v[13],fp*v[14]), // accelleration
                        fp*v[15], fp*v[16],                   // friction & randomness
                        fp*v[17], fp*v[18],                   // particle size
                        v[19], v[20]);                        // particle image and color textures

  ge->draw_type = static_cast<GfxPListEmitter::DRAW_TYPE>(v[21]);
  ge->blend = static_cast<GfxEmitter::BlendType>(v[22]);

  GfxGameObject *ggo = GUI_gfxm->game2gfx[gob];
  if (!ggo->gfx_obj) {
    ggo->attach_gfxobj(ge);
    GUI_gfxm->add_GfxObject(ge);
  }

  emitter = ge;
  return stype();
}

//-------------------------------------------------------------------

stype ScriptEmitter::init_circle(const SActParam &p)
{
  OBJN_CHECK(0);
  VALN_CHECK(23);

  vector<sint4> v;
  FORU (i, p.val.size()) {
    v.push_back(p.val[i].to_int());
  }
  
  real4 fp = 0.01;

  if (emitter) {
    delete emitter;
    emitter = 0;
  }

  GameObj *gob = this->parent->get_GameObj();
  SCRIPT_ASSERT(gob && "GameObj ptr 0");

  GfxCircleEmitter *ge =
    new GfxCircleEmitter(*GUI_gfxm,
                        v[0],                                 // creation rate
                        fp*v[1],                              // particle life
                        fp*v[2],                              // time limit
                        Vec3<real4>(fp*v[3],fp*v[4],fp*v[5]), // position
                        v[6], v[7],                           // min, max latitude
                        fp*v[8],  fp*v[9],                    // offset & randomness
                        fp*v[10], fp*v[11],                   // speed & randomness
                        Vec3<real4>(fp*v[12],fp*v[13],fp*v[14]), // accelleration
                        fp*v[15], fp*v[16],                   // friction & randomness
                        fp*v[17], fp*v[18],                   // particle size
                        v[19], v[20]);                        // particle image and color textures

  ge->draw_type = static_cast<GfxPListEmitter::DRAW_TYPE>(v[21]);
  ge->blend = static_cast<GfxEmitter::BlendType>(v[22]);

  GfxGameObject *ggo = GUI_gfxm->game2gfx[gob];
  if (!ggo->gfx_obj) {
    ggo->attach_gfxobj(ge);
    GUI_gfxm->add_GfxObject(ge);
  }

  emitter = ge;
  return stype();
}

//-------------------------------------------------------------------

stype ScriptEmitter::init_corona(const SActParam &p)
{
  OBJN_CHECK(0);
  VALN_CHECK(4);

  vector<sint4> v;
  FORU (i, p.val.size()) {
    v.push_back(p.val[i].to_int());
  }
  
  real4 fp = 0.001f;

  if (emitter) {
    delete emitter;
    emitter = 0;
  }

  GameObj *gob = this->parent->get_GameObj();
  SCRIPT_ASSERT(gob && "GameObj ptr 0");

  GfxLightCorona *ge =
    new GfxLightCorona(*GUI_gfxm,
                      fp*v[0],          // Particle Size
                      fp*v[1],          // Color texture variation speed
                      v[2], v[3]);  // Particle image and color textures

  if (!ge) cout << "Unable to create GfxLightCorona emitter!" << endl;

  GfxGameObject *ggo = GUI_gfxm->game2gfx[gob];
  if (!ggo->gfx_obj) {
    ggo->attach_gfxobj(ge);
    GUI_gfxm->add_GfxObject(ge);
  }

  emitter = ge;
  return stype();
}

//-------------------------------------------------------------------

stype ScriptEmitter::init_fire(const SActParam &p)
{
  OBJN_CHECK(0);
  VALN_CHECK(2);

  vector<sint4> v;
  FORU (i, p.val.size()) {
    v.push_back(p.val[i].to_int());
  }
  
  //real4 fp = 0.01f;

  if (emitter) {
    delete emitter;
    emitter = 0;
  }

  GameObj *gob = this->parent->get_GameObj();
  SCRIPT_ASSERT(gob && "GameObj ptr 0");

  GfxFireEmitter *ge = new GfxFireEmitter(*GUI_gfxm,
                                          0.01 * v[0], // Size
                                          1.0 * v[1]); // Speed

  if (!ge) cout << "Unable to create GfxFireEmitter emitter!" << endl;

  GfxGameObject *ggo = GUI_gfxm->game2gfx[gob];
  if (!ggo->gfx_obj) {
    ggo->attach_gfxobj(ge);
    GUI_gfxm->add_GfxObject(ge);
  }

  emitter = ge;
  return stype();
}

//-------------------------------------------------------------------

stype ScriptEmitter::init_shell(const SActParam &p)
{
  OBJN_CHECK(0);
  VALN_CHECK(2);

  vector<sint4> v;
  FORU (i, p.val.size()) {
    v.push_back(p.val[i].to_int());
  }
  
  real4 fp = 0.01f;

  if (emitter) {
    delete emitter;
    emitter = 0;
  }

  GameObj *gob = this->parent->get_GameObj();
  SCRIPT_ASSERT(gob && "GameObj ptr 0");

  GfxShellEmitter *ge =
    new GfxShellEmitter(*GUI_gfxm,
                      fp * v[0], // Size
                      fp * v[1]); // Speed

  if (!ge) cout << "Unable to create GfxShellEmitter emitter!" << endl;

  GfxGameObject *ggo = GUI_gfxm->game2gfx[gob];
  if (!ggo->gfx_obj) {
    ggo->attach_gfxobj(ge);
    GUI_gfxm->add_GfxObject(ge);
  }

  emitter = ge;
  return stype();
}

//-------------------------------------------------------------------

stype ScriptEmitter::init_lightning(const SActParam &p)
{
  OBJN_CHECK(0);
  VALN_CHECK(6);

  vector<sint4> v;
  FORU (i, p.val.size()) {
    v.push_back(p.val[i].to_int());
  }
  
  if (emitter) {
    delete emitter;
    emitter = 0;
  }

  GameObj *gob = this->parent->get_GameObj();
  SCRIPT_ASSERT(gob && "GameObj ptr 0");

  Vec3<real4> targ(v[0] / 16.0, v[1] / 16.0, v[2] / 16.0);

  GfxLightningEmitter *ge = new GfxLightningEmitter(*GUI_gfxm,
                                                    targ,
                                                    v[3] * 0.01,   // size
                                                    v[4] * 0.1,    // frequency
                                                    v[5]);         // num arcs

  if (!ge) cout << "Unable to create GfxLightningEmitter emitter!" << endl;

  GfxGameObject *ggo = GUI_gfxm->game2gfx[gob];
  if (!ggo->gfx_obj) {
    ggo->attach_gfxobj(ge);
    GUI_gfxm->add_GfxObject(ge);
  }

  emitter = ge;
  return stype();
}

//===================================================================

stype ScriptEmitter::set_rate(const SActParam &p)
{
  VALN_CHECK(1);
  SCRIPT_ASSERT(emitter && "no emitter");

  GfxPListEmitter *gpe = dynamic_cast<GfxPListEmitter*>(emitter);
  SCRIPT_ASSERT(gpe && "no GfxPListEmitter");

  gpe->set_rate(p.val[0].to_int());
  return stype();
}

//-------------------------------------------------------------------

stype ScriptEmitter::position(const SActParam &p)
{
  VALN_CHECK(3);
  SCRIPT_ASSERT(emitter && "no emitter");

  real4 x = p.val[0].to_int()/100.0;
  real4 y = p.val[1].to_int()/100.0;
  real4 z = p.val[2].to_int()/100.0;

  emitter->set_position(x, y, z);
  return stype();
}

//-------------------------------------------------------------------

stype ScriptEmitter::rotate(const SActParam &p)
{
  VALN_CHECK(1);
  SCRIPT_ASSERT(emitter && "no emitter");
  real4 y = p.val[0].to_int();
  emitter->set_rotation(y);
  return stype();
}

//-------------------------------------------------------------------

stype ScriptEmitter::scale(const SActParam &p)
{
  VALN_CHECK(1);
  SCRIPT_ASSERT(emitter && "no emitter");
  real4 s = p.val[0].to_int() * 0.01;
  emitter->set_scale(s);
  return stype();
}

//-------------------------------------------------------------------
/*
stype ScriptEmitter::draw_2d(const SActParam &p)
{
  return stype();
}

//-------------------------------------------------------------------

stype ScriptEmitter::draw_overlay(const SActParam &p)
{
  return stype();
}
*/
//-------------------------------------------------------------------

stype ScriptEmitter::set_parent(const SActParam &p)
{
  OBJN_CHECK(1);
  SCRIPT_ASSERT(emitter && "no emitter");

  ScriptObj *foo = p.obj[0];

  ScriptModel *s = dynamic_cast<ScriptModel*>(foo);
  if (s) {
    GfxAnimesh *o = s->get_mesh();
    
    if (p.val.size() == 1) {
      string tag_name = p.val[0].to_cstr();
      o->attachEmitter_tag(emitter, tag_name);
    } else {
      o->attachEmitter(emitter);
    }
    return stype_from_Emitter(emitter);
  }

  ScriptEmitter *e = dynamic_cast<ScriptEmitter*>(foo);
  if (e) {
    GfxEmitter *o = e->get_emitter();
    o->attachEmitter(emitter);
    return stype_from_Emitter(emitter);
  }

  SCRIPT_ERR("cannot attach to non ScriptModel/ScriptEmitter " + foo->bp_name());
  return stype(-1);
}

//-------------------------------------------------------------------

stype ScriptEmitter::enable(const SActParam &p)
{
  VALN_CHECK(1);
  SCRIPT_ASSERT(emitter && "no emitter");  

  if (p.val[0].to_int()) emitter->enable();
  else                   emitter->disable();
  return stype();
}

//===================================================================

ObjAction* ScriptEmitter::action(const std::string &name) const
{
  FIND (act_map, it, name);
  if (it != act_map.end()) return it->second;
  return 0;
}

//===================================================================

void ScriptEmitter::destruct()
{
  if (emitter) emitter->kill();
  emitter = 0;
}

//===================================================================
