#ifndef Map_H
#define Map_H

/** @file Map.H
    Tile container

    $Id: Map.H,v 1.17 2007/03/24 22:44:42 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Map.H,v $
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"
#include "SaveLoad.H"

class ByteStream;

//===================================================================

// (0,0) is the upper left corner (NW)

template <typename TILE> class Map : public SaveLoad
{
protected:
  
  sint4 width, height, num;
  TILE *tiles;

public:

  Map() { width = height = num = 0; tiles = 0; }
  
  ~Map() {
    if (tiles) delete[] tiles;
  }

  void allocate(sint4 width_, sint4 height_) {
    assert(width_ > 0 && height_ > 0);
    delete [] tiles;
    width = width_; height = height_;
    num = width * height;
    tiles = new TILE[num];
  }

  sint4 get_width() const { return width; }
  sint4 get_height() const { return height; }
  sint4 get_num() const { return num; }
  
  // Tile access functions: by index or (x,y)
  TILE &get_tile(sint4 ind) const { assert(ind >= 0 && ind < num); return tiles[ind]; }
  TILE &get_tile(sint4 x, sint4 y) const { return get_tile(xy2ind(x, y)); }

  TILE &operator()(int ind) { return get_tile(ind); }
  const TILE &operator()(int ind) const { return get_tile(ind); }

  TILE &operator()(int x, int y) { return get_tile(x, y); }
  const TILE &operator()(int x, int y) const { return get_tile(x, y); }

  TILE *get_tiles() { return tiles; } // fixme: unsafe
  const TILE *get_tiles() const { return tiles; }

  //-------------------------------------
  
  void read(ByteStream &data) {
    FORS (i, num) tiles[i].read_b(data);
  }

  void write(std::stringstream &data) {
    std::string tile_data;
    FORS (i, num) tiles[i].write(tile_data);
    data.write(tile_data.c_str(), tile_data.size());
    data << std::endl;
    std::cout << "WROTE " << tile_data.size() << std::endl;
  }

  //-------------------------------------

  void save(std::ostream &os) const {
    FORS (i, num) tiles[i].save(os);
  }

  void load(std::istream &is) {
    FORS (i, num) tiles[i].load(is);
  }

  //-------------------------------------

  // write readable representation to std::cout
  
  void display() {
    FORS (i, num) {
      tiles[i].display();
      if ((i+1) % width == 0) std::cout << std::endl;
    }
  }

  void clear() {
    FORS (i, num) tiles[i].clear();
  }

  // index - (x,y) conversion
  
  sint4 ind2x(sint4 ind) const { assert(ind >= 0 && ind < num); return ind % width; }

  sint4 ind2y(sint4 ind) const { assert(ind >= 0 && ind < num); return ind / width; }  
  
  sint4 xy2ind(sint4 x, sint4 y) const {
    assert(x >= 0 && x < width);
    assert(y >= 0 && y < height);
    return y * width + x;
  }

};

//===================================================================

#endif
