// $Id: HeuristicMapWidget.cpp,v 1.1 2006/08/10 17:19:35 nicholas Exp $

// This is an ORTS file (c) Frano Sailer, licensed under the GPL

#include "HeuristicMapWidget.h"
#include "GameStateModule.H"

HeuristicMapWidget::HeuristicMapWidget(Point2i pos, real8 mapScaleSize,
				       real8 gridSquareWidth, real8 gridSquareHeight,
				       real8 alpha, bool overlay)
  : CWidget()
{
  this->pos = pos;
  this->mapScaleSize = mapScaleSize;

  this->gridSquareWidth = gridSquareWidth;
  this->gridSquareHeight = gridSquareHeight;

  showFriendly = true;
  showEnemy = true;
  showNeutral = true;
  currentType = 0;

  setAlpha(alpha);

  this->overlay = overlay;

  lastChangedObjectPos.x = 0;
  lastChangedObjectPos.y = 0;

  heuristicMap = NULL;
  oldHeuristicMap = NULL;
  immobileHeuristicMap = NULL;
  ageMap = NULL;
}

HeuristicMapWidget::~HeuristicMapWidget() {
  deleteMap();
}

//sint4 HeuristicMapWidget::draw() {
stype HeuristicMapWidget::draw() {

  //glEnable(GL_SCISSOR_TEST);
  //glScissor(x-2,y-2,w+3,h+30);

  glColor4f(1, 0, 0, 1);

  //    glEnable(GL_TEXTURE_2D);
  glDisable(GL_TEXTURE_2D);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  // draw the grid
  if (numRows <= 0 || numColumns <= 0 || attributes.size() == 0) {
    printf("Error in HeuristicMapWidget,  columns AND rows need to be > 0 and need to have at least one attribute\n");
    return 0;
  }

  real8 hGap = (real8)h.to_int() / (real8)numRows;
  real8 vGap = (real8)w.to_int() / (real8)numColumns;

  FORU (i,numColumns) {
    FORU (j,numRows) {
      if (attributes[currentAttributeIndex].maxValue == 0.0) {
        glColor4f(0,0,0,alpha);
      } else {
	uint4 row = numRows-j-1;
        // for us, we have green map markings, enemies, get red, neutral get blue
        real8 ourValue = (heuristicMap[i][row][0][currentAttributeIndex] +
			  oldHeuristicMap[i][row][0][currentAttributeIndex] +
			  immobileHeuristicMap[i][row][0][currentAttributeIndex]) /
	  attributes[currentAttributeIndex].maxValue;

        real8 enemyValue = (heuristicMap[i][row][1][currentAttributeIndex] +
			    oldHeuristicMap[i][row][1][currentAttributeIndex] +
			    immobileHeuristicMap[i][row][1][currentAttributeIndex]) /
	  attributes[currentAttributeIndex].maxValue;

        real8 neutralValue = (heuristicMap[i][row][2][currentAttributeIndex] +
			      oldHeuristicMap[i][row][2][currentAttributeIndex] +
			      immobileHeuristicMap[i][row][2][currentAttributeIndex]) /
	  attributes[currentAttributeIndex].maxValue;
	/*
	  if (ourValue < -0.1 || enemyValue < -0.1 || neutralValue < -0.1) {
	  printf("VALUES = %f,%f,%f\n",ourValue,enemyValue,neutralValue);
	  }
	*/
        if (!showFriendly) ourValue = 0.0;
        if (!showEnemy) enemyValue = 0.0;
        if (!showNeutral) neutralValue = 0.0;

        glColor4f(enemyValue, ourValue, neutralValue, alpha);
        //glColor4f(0, ourValue,0, alpha);
      }
      //printf("%f,",heuristicMap[j][numRows-i-1]);
      real8 xStart = (real8)x.to_int()+(vGap*(real8)i);
      real8 yStart = (real8)y.to_int()+(hGap*(real8)j);
      glBegin(GL_POLYGON);
      glVertex2f(xStart,yStart);
      glVertex2f(xStart+vGap, yStart);
      glVertex2f(xStart+vGap, yStart+hGap);
      glVertex2f(xStart, yStart+hGap);
      glEnd();
    }
    //printf("\n");
  }

  // draw the current answers
  for (uint4 i=0; i < queryResults.size(); i++) {

    real8 xStart = (real8)x.to_int()+(vGap*(real8)queryResults[i]->pos.x);
    real8 yStart = (real8)y.to_int()+(hGap*(real8)(getRows()-queryResults[i]->pos.y-1));

    glColor4f(1, 1, 1, 1);
    glBegin(GL_POLYGON);
    glVertex2f(xStart+2,yStart+2);
    glVertex2f(xStart+vGap-2,yStart+2);
    glVertex2f(xStart+vGap-2,yStart+hGap-2);
    glVertex2f(xStart+2,yStart+hGap-2);
    glEnd();
  }

  // draw the outline of the widget
  glLineWidth(2.0);
  glColor4f(1,0,0,1);
  glBegin(GL_LINE_LOOP);
  glVertex2f(x.to_int(),y.to_int());
  glVertex2f(x.to_int()+w.to_int(),y.to_int());
  glVertex2f(x.to_int()+w.to_int(),y.to_int()+h.to_int());
  glVertex2f(x.to_int(),y.to_int()+h.to_int());
  glEnd();

  glRasterPos2f(x.to_int(), y.to_int()+h.to_int()+5);
  //glScalef(2.0f,2.0f,1.0f);
  if (attributes[currentAttributeIndex].isComponent) {
    FORU (i,attributes[currentAttributeIndex].componentName.length()) {
      glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,attributes[currentAttributeIndex].componentName[i]);
    }
    glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,'.');

    FORU (i,attributes[currentAttributeIndex].attributeName.length()) {
      glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,attributes[currentAttributeIndex].attributeName[i]);
    }
  } else {
    FORU (i,attributes[currentAttributeIndex].attributeName.length()) {
      glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,attributes[currentAttributeIndex].attributeName[i]);
    }
  }
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  //glEnable(GL_BLEND);
  //glEnable(GL_LINE_SMOOTH);

  //glLineWidth(2.0);
  //glScalef(0.2f,0.2f,1.0f);
  //for (real8 i=y;i<=y+h;i+=hGap) {
  /*
    glPushMatrix();
    glTranslatef(x+10, y+10, 0);

    glutStrokeCharacter(GLUT_STROKE_ROMAN,'a');
    glPopMatrix();
  */
  //}

  //  glEnable(GL_TEXTURE_2D);
  // glDisable(GL_SCISSOR_TEST);
  return 0;

}

// make sure this is always called immediately after the attributes are added, otherwise, there will be a segfault
// since the map will not be large enough to accomodate the new attributes.

void HeuristicMapWidget::initializeMap() {
  if (heuristicMap != 0 ||
      oldHeuristicMap != 0 ||
      immobileHeuristicMap != 0 ||
      ageMap != 0) {
    deleteMap();
  }

  heuristicMap = new real8***[numColumns];
  oldHeuristicMap = new real8***[numColumns];
  immobileHeuristicMap = new real8***[numColumns];
  ageMap = new uint8*[numColumns];

  FORU (i, numColumns) {
    heuristicMap[i] = new real8**[numRows];
    oldHeuristicMap[i] = new real8**[numRows];
    immobileHeuristicMap[i] = new real8**[numRows];
    ageMap[i] = new uint8[numRows];

    FORU (j, numRows) {
      heuristicMap[i][j] = new real8*[3];
      oldHeuristicMap[i][j] = new real8*[3];
      immobileHeuristicMap[i][j] = new real8*[3];
      ageMap[i][j] = 0;

      FORU (k, 3) {
        heuristicMap[i][j][k] = new real8[attributes.size()];
	oldHeuristicMap[i][j][k] = new real8[attributes.size()];
	immobileHeuristicMap[i][j][k] = new real8[attributes.size()];

	FORU (l, attributes.size()) {
	  heuristicMap[i][j][k][l] = 0.0f;
	  oldHeuristicMap[i][j][k][l] = 0.0f;
	  immobileHeuristicMap[i][j][k][l] = 0.0f;
	}
      }
    }
  }

  currentAttributeIndex = 0;
}

void HeuristicMapWidget::deleteMap() {

  if (heuristicMap != 0) {
    FORU (i, numColumns) {
      FORU (j, numRows) {
	FORU (k, 3) {
	  delete heuristicMap[i][j][k];
	}
	delete[] heuristicMap[i][j];
      }
      delete[] heuristicMap[i];
    }
    delete[] heuristicMap;
    heuristicMap = 0;
  }

  if (oldHeuristicMap != 0) {
    FORU (i, numColumns) {
      FORU (j, numRows) {
	FORU (k,3) {
	  delete oldHeuristicMap[i][j][k];
	}
	delete[] oldHeuristicMap[i][j];
      }
      delete[] oldHeuristicMap[i];
    }
    delete[] oldHeuristicMap;
    oldHeuristicMap = 0;
  }

  if (immobileHeuristicMap != 0) {
    FORU (i, numColumns) {
      FORU (j, numRows) {
	FORU (k, 3) {
	  delete immobileHeuristicMap[i][j][k];
	}
	delete[] immobileHeuristicMap[i][j];
      }
      delete[] immobileHeuristicMap[i];
    }
    delete[] immobileHeuristicMap;
    immobileHeuristicMap = 0;
  }
  if (ageMap != 0) {
    FORU (i, numColumns) {
      delete[] ageMap[i];
    }
    delete[] ageMap;
    ageMap = 0;
  }
}


stype HeuristicMapWidget::event_mouse_motion(stype /*sx*/,stype /*sy*/, stype /*button*/) {
  // if (buttons == mouse wheel) { handel; return 1; } else { return 0;}
  return 0;
}
stype HeuristicMapWidget::event_mouse_hover(stype /*sx*/,stype /*sy*/, stype /*button*/) {
  return 0;
}
stype HeuristicMapWidget::event_mouse_up(stype /*sx*/,stype /*sy*/, stype /*button*/) {
  return 0;
}

//sint4 HeuristicMapWidget::event_mouse_down(sint4 /*sx*/,sint4 /*sy*/, sint4 button) {
stype HeuristicMapWidget::event_mouse_down(stype,stype,stype buttontype){
  sint4 button = buttontype.to_int();

  printf("FRANO: button %d used on mouse\n",(int)button);
  // mouse wheel up
  if (button == 3) {
    printf("mouse wheel up\n");
    currentAttributeIndex++;
    if (currentAttributeIndex >= (sint4)attributes.size()) {
      currentAttributeIndex = 0;
    }
    return 0;

  } else if (button == 4) { // wheel down
    printf("mouse wheel down\n");
    currentAttributeIndex--;
    if (currentAttributeIndex < 0) {
      currentAttributeIndex = attributes.size()-1;
    }
    return 0;
  }

  return 0;
}


//sint4 HeuristicMapWidget::event_key_down(sint4 key,sint4 /*mod*/, sint4 /*x*/,sint4 /*y*/) {
stype HeuristicMapWidget::event_key_down(stype keytype,stype,stype,stype){
  printf("Key pressed in widget!\n");
  sint4 key = keytype.to_int();
  /*
    currentAttributeIndex++;
    if (currentAttributeIndex >= attributes.size()) {
    currentAttributeIndex = 0;
    }
  */
  //printf("Key %d pressed in widget, character value is %c",key,key);

  // add or subtract 10% of current max value, to eith ermake display be brighter, or dimmer
  // 61 is '+' key
  if (key == 'z') {
    attributes[currentAttributeIndex].maxValue -= attributes[currentAttributeIndex].maxValue/10;
    return 1;
  } else if (key == 'x') {// '-' key
    attributes[currentAttributeIndex].maxValue += attributes[currentAttributeIndex].maxValue/10;
    return 1;
  } else if (key == 'f') {
    showFriendly = !showFriendly;
    return 1;
  } else if (key == 'e') {
    showEnemy = !showEnemy;
    return 1;
  } else if (key == 'n') {
    showNeutral = !showNeutral;
    return 1;
  } else if (key == 257) { // 257 = 1, 258 = 2, 259, = 3
    currentType = 0;
    return 1;
  } else if (key == 258) {
    currentType = 1;
    return 1;
  } else if (key == 259) {
    currentType = 2;
    return 1;
  }	else if (key == 'q') {
    getMaxValues(10, -10000.0, 10000.0, currentType,
		 attributes[currentAttributeIndex].attributeName);
    return 1;
  } else if (key == 'w') {
    getNearestValues(lastChangedObjectPos, 10, 10.0, 10000.0, currentType,
		     attributes[currentAttributeIndex].attributeName);
    return 1;
  }    else if (key == 'o'){
    currentAttributeIndex++;
    if (currentAttributeIndex >= (sint4)attributes.size()) {
      currentAttributeIndex = 0;
    }
    return 1;
  }   else if (key == 'p'){
    currentAttributeIndex--;
    if (currentAttributeIndex < 0) {
    currentAttributeIndex = (sint4)attributes.size()-1;
    }
    return 1;
  }

  return 0;
}

stype HeuristicMapWidget::event_key_up(stype /*key*/,stype /*mod*/, stype /*x*/,stype /*y*/) {
  return 0;
}

void HeuristicMapWidget::setGfxModule(GfxModule* gfxm) {
  this->gfxm = gfxm;
}

void HeuristicMapWidget::setGameStateModule(GameStateModule* gsm) {
  this->gsm = gsm;
}

void HeuristicMapWidget::setAlpha(real8 alpha) {
  this->alpha = alpha;
}

void HeuristicMapWidget::setRows(uint4 rows) {
  this->numRows = rows;
}

void HeuristicMapWidget::setColumns(uint4 columns) {
  this->numColumns = columns;
}

void HeuristicMapWidget::addAttribute(const std::string &newAttributeName, real8 maxValue)
{
  Attribute newAttr;
  newAttr.maxValue = maxValue;

  // do all this processing here at the start, so we don't have to do any runtime string finding(since it is SLOOOW)
  sint4 periodIndex = newAttributeName.find(".");

  if (periodIndex != (sint4)std::string::npos) {
    newAttr.componentName = newAttributeName.substr(0, periodIndex);
    newAttr.attributeName = newAttributeName.substr(periodIndex+1, newAttributeName.length());
    newAttr.isComponent = true;
  } else {
    newAttr.attributeName = newAttributeName;
    newAttr.componentName = "";
    newAttr.isComponent = false;
  }

  attributes.push_back(newAttr);
}

Vector<Attribute> HeuristicMapWidget::getAttributes() {
  return attributes;
}

uint4 HeuristicMapWidget::getNumAttributes() {
  return attributes.size();
}

uint4 HeuristicMapWidget::getRows() {
  return numRows;
}

uint4 HeuristicMapWidget::getColumns() {
  return numColumns;
}

real8 HeuristicMapWidget::getGridHeight() {
  return gridHeight;
}

real8 HeuristicMapWidget::getGridWidth() {
  return gridWidth;
}

real8 HeuristicMapWidget::getGridSquareHeight() {
  return gridSquareHeight;
}

real8 HeuristicMapWidget::getGridSquareWidth() {
  return gridSquareWidth;
}

Point2i HeuristicMapWidget::getPos() {
  return pos;
}

uint4 HeuristicMapWidget::getWidth() {
  return width;
}

uint4 HeuristicMapWidget::getHeight() {
  return height;
}

bool HeuristicMapWidget::isOverlay() {
  return overlay;
}


void HeuristicMapWidget::calculateGridDimensions() {

  gridWidth =
    gsm->get_game().get_map().get_width() * gsm->get_game().get_tile_points() /(real8)numColumns;
  gridHeight =
    gsm->get_game().get_map().get_height() * gsm->get_game().get_tile_points() / (real8)numRows;
}

real8 HeuristicMapWidget::getAlpha() {
  return alpha;
}

sint4 HeuristicMapWidget::getAttributeIndex(std::string attributeName) {
  sint4 attributeIndex = -1;
  // first we have to find the attribute
  FORU (i, attributes.size()) {
    if (attributes[i].attributeName == attributeName) {
      attributeIndex = i;
      break;
    }
  }

  return attributeIndex;
}

// queries that can be asked
// the processing of them is VERY inefficient, but there really isn't a much better way of doing it, and
// since the number of grid squares is relatively small, and the code is called VERY infrequently, it should be OK(I hope)


void HeuristicMapWidget::getMaxValues(uint4 numValues, real8 lowerLimit, real8 upperLimit,
				      uint4 type, const std::string &attributeName)
{
  sint4 attributeIndex = getAttributeIndex(attributeName);

  // if that attribute doesn't exist, then don't do anything
  if (attributeIndex < 0) return;

  // delete the old vector
  FORU (i, queryResults.size()) {
    delete queryResults[i];
  }
  queryResults.clear();

  uint4 numEnteredValues = 0;

  FORU (i, getColumns()) {
    FORU (j, getRows()) {

      real8 currentValue = heuristicMap[i][j][type][attributeIndex] +
	oldHeuristicMap[i][j][type][attributeIndex] +
	immobileHeuristicMap[i][j][type][attributeIndex];

      if (currentValue < lowerLimit || currentValue > upperLimit) continue;

      if (numEnteredValues < numValues) {

	QueryResult* newResult = new QueryResult(Point2i(i,j),currentValue,0);
	queryResults.push_back(newResult);
	for (sint4 k = queryResults.size()-1; k >= 1;k--) {
	  // go backwards, and insert as soon as we see a smaller value... so keep it sorted...
	  if (queryResults[k]->value < queryResults[k-1]->value) {
	    QueryResult* tempPtr = queryResults[k-1];
	    queryResults[k-1] = queryResults[k];
	    queryResults[k] = tempPtr;
	  }
	}
	numEnteredValues++;
      } else {
	// if we have filled up our vector, we have to start dropping elements, add another only if it is biggest in our vector
	if (currentValue > queryResults[0]->value) {
	  QueryResult* newResult = new QueryResult(Point2i(i,j),currentValue,0);
	  queryResults.push_back(newResult);
	  delete queryResults[0];
	  queryResults.erase(queryResults.begin());

	  for (sint4 k = queryResults.size()-1; k >= 1;k--) {
	    // go backwards, and insert as soon as we see a smaller value... so keep it sorted...
	    if (queryResults[k]->value < queryResults[k-1]->value) {
	      QueryResult* tempPtr = queryResults[k-1];
	      queryResults[k-1] = queryResults[k];
	      queryResults[k] = tempPtr;
	    }
	  }
	}
      }
      /*
	printf("---------\n");
	for (uint4 i=0;i<queryResults.size();i++) {
	printf("VECTOR: %d, %d, %f\n",queryResults[i]->pos.x,queryResults[i]->pos.y,queryResults[i]->value);
	}
      */
    }
  }

}


// yes, I know this is probably the WORST way to do this... but it's easy to code and it takes just as long as the max search anyways
void HeuristicMapWidget::getNearestValues(Point2i objectPos, uint4 numValues, real8 lowerLimit,
					  real8 upperLimit, uint4 type,
					  const std::string &attributeName)
{
  sint4 attributeIndex = getAttributeIndex(attributeName);

  // if that attribute doesn't exist, then don't do anything
  if (attributeIndex < 0) return;

  // delete the old vector
  FORU (i, queryResults.size()) {
    delete queryResults[i];
  }
  queryResults.clear();

  uint4 numEnteredValues = 0;

  for (uint4 i=0; i < getColumns(); i++) {

    for (uint4 j=0; j < getRows(); j++) {

      real8 currentValue = heuristicMap[i][j][type][attributeIndex] +
	oldHeuristicMap[i][j][type][attributeIndex] +
	immobileHeuristicMap[i][j][type][attributeIndex];

      if (currentValue < lowerLimit || currentValue > upperLimit) continue;

      // fixme: pow -> *
      real8 distance = sqrt(pow((sint4)(i-objectPos.x), 2.0)+pow((sint4)(j-objectPos.y), 2.0));

      if (numEnteredValues < numValues) {

	QueryResult* newResult = new QueryResult(Point2i(i, j), currentValue, distance);
	queryResults.push_back(newResult);
	for (sint4 k = queryResults.size()-1; k >= 1; k--) {
	  // go backwards, and insert as soon as we see a smaller value... so keep it sorted...
	  if (queryResults[k]->distance > queryResults[k-1]->distance) {
	    QueryResult* tempPtr = queryResults[k-1];
	    queryResults[k-1] = queryResults[k];
	    queryResults[k] = tempPtr;
	  }
	}
	numEnteredValues++;
      } else {
	// if we have filled up our vector, we have to start dropping elements, add another only if it is biggest in our vector
	if (distance < queryResults[0]->distance) {
	  QueryResult* newResult = new QueryResult(Point2i(i, j), currentValue, distance);
	  queryResults.push_back(newResult);
	  delete queryResults[0];
	  queryResults.erase(queryResults.begin());

	  for (sint4 k = queryResults.size()-1; k >= 1;k--) {
	    // go backwards, and insert as soon as we see a smaller value... so keep it sorted...
	    if (queryResults[k]->distance > queryResults[k-1]->distance) {
	      QueryResult* tempPtr = queryResults[k-1];
	      queryResults[k-1] = queryResults[k];
	      queryResults[k] = tempPtr;
	    }
	  }
	}
      }
      /*
	printf("---------\n");
	for (uint4 i=0;i<queryResults.size();i++) {
	printf("VECTOR: %d, %d, %f, %f\n",queryResults[i]->pos.x,queryResults[i]->pos.y,queryResults[i]->value,queryResults[i]->distance);
	}
      */
    }
  }
}

