#define OGRE_RELEASE_ASSERT 0
#include "OgreGfxModuleFrameListener.H"


//TODO: REMOVE_THIS

int randInt(int a, int b)//random number function *very* useful
{
	return rand()%(b+1-a)+a;
}

//OgreGfxModuleFrameListener::

void OgreGfxModuleFrameListener::updateStats(void){
  static Ogre::String currFps = "Current FPS: ";
  static Ogre::String avgFps = "Average FPS: ";
  static Ogre::String bestFps = "Best FPS: ";
  static Ogre::String worstFps = "Worst FPS: ";
  static Ogre::String tris = "Triangle Count: ";

  // update stats when necessary
  try {
    Ogre::OverlayElement* guiAvg = Ogre::OverlayManager::getSingleton().getOverlayElement("Core/AverageFps");
    Ogre::OverlayElement* guiCurr = Ogre::OverlayManager::getSingleton().getOverlayElement("Core/CurrFps");
    Ogre::OverlayElement* guiBest = Ogre::OverlayManager::getSingleton().getOverlayElement("Core/BestFps");
    Ogre::OverlayElement* guiWorst = Ogre::OverlayManager::getSingleton().getOverlayElement("Core/WorstFps");
	
    const Ogre::RenderTarget::FrameStats& stats = mWindow->getStatistics();

    guiAvg->setCaption(avgFps + Ogre::StringConverter::toString(stats.avgFPS));
    guiCurr->setCaption(currFps + Ogre::StringConverter::toString(stats.lastFPS));
    guiBest->setCaption(bestFps + Ogre::StringConverter::toString(stats.bestFPS)
			+" "+Ogre::StringConverter::toString(stats.bestFrameTime)+" ms");
    guiWorst->setCaption(worstFps + Ogre::StringConverter::toString(stats.worstFPS)
			 +" "+Ogre::StringConverter::toString(stats.worstFrameTime)+" ms");

    Ogre::OverlayElement* guiTris = Ogre::OverlayManager::getSingleton().getOverlayElement("Core/NumTris");
    guiTris->setCaption(tris + Ogre::StringConverter::toString(stats.triangleCount));

    Ogre::OverlayElement* guiDbg = Ogre::OverlayManager::getSingleton().getOverlayElement("Core/DebugText");
    guiDbg->setCaption(mWindow->getDebugText());
  }
  catch(...){
    // ignore
  }
}

    // Constructor takes a RenderWindow because it uses that to determine input context
OgreGfxModuleFrameListener::OgreGfxModuleFrameListener(Ogre::RenderWindow* win, Ogre::Camera* cam, Ogre::SceneManager* sceneManager, bool useBufferedInputKeys, bool useBufferedInputMouse){

  for(int i=0; i<256; i++){ keys[i]=false; }
 
  mDebugOverlay = Ogre::OverlayManager::getSingleton().getByName("Core/DebugOverlay");
  mUseBufferedInputKeys = useBufferedInputKeys;
  mUseBufferedInputMouse = useBufferedInputMouse;
  mInputTypeSwitchingOn = mUseBufferedInputKeys || mUseBufferedInputMouse;
  mRotateSpeed = 36;
  mMoveSpeed = 100;

  if (mInputTypeSwitchingOn){
    mEventProcessor = new Ogre::EventProcessor();
    mEventProcessor->initialise(win);
    mEventProcessor->startProcessingEvents();
    mEventProcessor->addKeyListener(this);
    mEventProcessor->addMouseListener( this );
    mEventProcessor->addMouseMotionListener( this );
    mInputDevice = mEventProcessor->getInputReader();
  }else{
    mInputDevice = Ogre::PlatformManager::getSingleton().createInputReader();
    mInputDevice->initialise(win,true, true);
  }

  mSceneMgr = sceneManager;
  ObjectManager::getSingleton().gdSceneMgr = mSceneMgr;
  mCamera = cam;
  mWindow = win;
  mStatsOn = true;
  mNumScreenShots = 0;
  mTimeUntilNextToggle = 0;
  mSceneDetailIndex = 0;
  mMoveScale = 1.0f;
  mRotScale = 0.0f;
  mTranslateVector = Ogre::Vector3::ZERO;
  mAniso = 1;
  mFiltering = Ogre::TFO_BILINEAR;
  mRightMousePressed = false;
  mLeftMousePressed = false;
  mRightMousePressedLastFrame = false;
  mLeftMousePressedLastFrame = false;
  mMiddleMousePressed = false;
	
  gui_obj = 0;

  showDebugOverlay(true);

  // Create RaySceneQuery
  mRaySceneQuery = mSceneMgr->createRayQuery( Ogre::Ray() );

 //Initialize all our Fancy GUI stuff that we'll need.
  mGUIRenderer = new CEGUI::OgreCEGUIRenderer(mWindow, Ogre::RENDER_QUEUE_OVERLAY, false, 3000, mSceneMgr);
  mGUISystem = new CEGUI::System(mGUIRenderer);
  CEGUI::SchemeManager::getSingleton().loadScheme((CEGUI::utf8*)"TaharezLookSkin.scheme");
  CEGUI::MouseCursor::getSingleton().setImage("TaharezLook", "MouseArrow");
}

bool OgreGfxModuleFrameListener::init(GameStateModule* gameStateModule){
  gsm = gameStateModule;
  Ogre::SceneNode* mapNode = mSceneMgr->getRootSceneNode()->createChildSceneNode("battleField SceneNode");
  Game* game = &(gsm->get_game());
  int tileRows = game->get_gtiles_y();
  int tileCols = game->get_gtiles_x();
  
  int megaRows = 8;
  int megaCols = 8;

  int rowsPerMegaTile = tileRows/8;
  int colsPerMegatile = tileCols/8;

  float tileWidth = (gsm->get_game()).get_tile_points ();
  float tileHeight= tileWidth;
  //std::cout<<"TileRows = "<<tileRows;
  //std::cout<<" TileCols = "<<tileCols<<std::endl; 
  
  battleField = new OgreGfxModuleTerrainTileGrid(mSceneMgr, mapNode, rowsPerMegaTile, colsPerMegatile,
						 megaRows, megaCols,tileWidth ,tileHeight );
  return true;
}

 OgreGfxModuleFrameListener::~OgreGfxModuleFrameListener(){
  if (mInputTypeSwitchingOn){
    delete mEventProcessor;
  }
  else{
    Ogre::PlatformManager::getSingleton().destroyInputReader( mInputDevice );
  }
  delete mRaySceneQuery;
}

 bool OgreGfxModuleFrameListener::processUnbufferedKeyInput(const Ogre::FrameEvent& evt){
  if (mInputDevice->isKeyDown(Ogre::KC_A)){
    // Move camera left
    mTranslateVector.x = -mMoveScale;
  }

  if (mInputDevice->isKeyDown(Ogre::KC_D)){
    // Move camera RIGHT
    mTranslateVector.x = mMoveScale;
  }

  /* Move camera forward by keypress. */
  if (mInputDevice->isKeyDown(Ogre::KC_UP) || mInputDevice->isKeyDown(Ogre::KC_W) ){
    mTranslateVector.z = -mMoveScale;
  }

  /* Move camera backward by keypress. */
  if (mInputDevice->isKeyDown(Ogre::KC_DOWN) || mInputDevice->isKeyDown(Ogre::KC_S) ){
    mTranslateVector.z = mMoveScale;
    std::cout << "Moving ->" << mMoveScale << std::endl;
  }

  if (mInputDevice->isKeyDown(Ogre::KC_PGUP)){
    // Move camera up
    mTranslateVector.y = mMoveScale;
  }

  if (mInputDevice->isKeyDown(Ogre::KC_PGDOWN)){
    // Move camera down
    mTranslateVector.y = -mMoveScale;
  }

  if (mInputDevice->isKeyDown(Ogre::KC_RIGHT)){
    mCamera->yaw(-mRotScale);
  }
		
  if (mInputDevice->isKeyDown(Ogre::KC_LEFT)){
    mCamera->yaw(mRotScale);
  }

  if( mInputDevice->isKeyDown( Ogre::KC_ESCAPE) ){            
    return false;
  }

  // see if switching is on, and you want to toggle 
  if (mInputTypeSwitchingOn && mInputDevice->isKeyDown(Ogre::KC_M) && mTimeUntilNextToggle <= 0){
    switchMouseMode();
    mTimeUntilNextToggle = 1;
  }

  if (mInputTypeSwitchingOn && mInputDevice->isKeyDown(Ogre::KC_K) && mTimeUntilNextToggle <= 0){
    // must be going from immediate keyboard to buffered keyboard
    switchKeyMode();
    mTimeUntilNextToggle = 1;
  }
  if (mInputDevice->isKeyDown(Ogre::KC_F) && mTimeUntilNextToggle <= 0){
    mStatsOn = !mStatsOn;
    showDebugOverlay(mStatsOn);

    mTimeUntilNextToggle = 1;
  }
  if (mInputDevice->isKeyDown(Ogre::KC_T) && mTimeUntilNextToggle <= 0){
    switch(mFiltering){
    case Ogre::TFO_BILINEAR:
      mFiltering = Ogre::TFO_TRILINEAR;
      mAniso = 1;
      break;
    case Ogre::TFO_TRILINEAR:
      mFiltering = Ogre::TFO_ANISOTROPIC;
      mAniso = 8;
      break;
    case Ogre::TFO_ANISOTROPIC:
      mFiltering = Ogre::TFO_BILINEAR;
      mAniso = 1;
      break;
    default:
      break;
    }
    Ogre::MaterialManager::getSingleton().setDefaultTextureFiltering(mFiltering);
    Ogre::MaterialManager::getSingleton().setDefaultAnisotropy(mAniso);


    showDebugOverlay(mStatsOn);
    
    mTimeUntilNextToggle = 1;
  }

  if (mInputDevice->isKeyDown(Ogre::KC_SYSRQ) && mTimeUntilNextToggle <= 0){
    char tmp[20];
    sprintf(tmp, "screenshot_%d.png", ++mNumScreenShots);
    mWindow->writeContentsToFile(tmp);
    mTimeUntilNextToggle = 0.5;
    mWindow->setDebugText(Ogre::String("Wrote ") + tmp);
  }
		
  if (mInputDevice->isKeyDown(Ogre::KC_R) && mTimeUntilNextToggle <=0){
    mSceneDetailIndex = (mSceneDetailIndex+1)%3 ;
    switch(mSceneDetailIndex) {
    case 0 : mCamera->setPolygonMode(Ogre::PM_SOLID) ; break ;
    case 1 : mCamera->setPolygonMode(Ogre::PM_WIREFRAME) ; break ;
    case 2 : mCamera->setPolygonMode(Ogre::PM_POINTS) ; break ;
    }
    mTimeUntilNextToggle = 0.5;
  }

  static bool displayCameraDetails = false;
  if (mInputDevice->isKeyDown(Ogre::KC_P) && mTimeUntilNextToggle <= 0){
    displayCameraDetails = !displayCameraDetails;
    mTimeUntilNextToggle = 0.5;
    if (!displayCameraDetails)
      mWindow->setDebugText("");
  }
  if (displayCameraDetails){
    // Print camera details
    mWindow->setDebugText("P: " + Ogre::StringConverter::toString(mCamera->getDerivedPosition()) + " " + 
			  "O: " + Ogre::StringConverter::toString(mCamera->getDerivedOrientation()));
  }

  // Return true to continue rendering
  return true;
}

bool OgreGfxModuleFrameListener::processUnbufferedMouseInput(const Ogre::FrameEvent& evt){
  /* Rotation factors, may not be used if the second mouse button is pressed. */

  /* If the second mouse button is pressed, then the mouse movement results in 
     sliding the camera, otherwise we rotate. */
  if( mInputDevice->getMouseButton( 1 ) ){
    mTranslateVector.x += mInputDevice->getMouseRelativeX() * 0.13;
    mTranslateVector.y -= mInputDevice->getMouseRelativeY() * 0.13;
  }
  else{
    mRotX = Ogre::Degree(-mInputDevice->getMouseRelativeX() * 0.13);
    mRotY = Ogre::Degree(-mInputDevice->getMouseRelativeY() * 0.13);
  }

  return true;
}

/*=============================================================================
Function - processKeyChanges()
Called From OgreGrxModuleFrameListener.frameStarted();
Purpose: This function does all nessessary commands  based upon the current keys
pressed.
==============================================================================*/
void OgreGfxModuleFrameListener::processKeyChanges(){
  Game::pick_game(&(gsm->get_game()));  //TODO:find out what this does..
  
  // Make all the changes to the camera
  // Note that YAW direction is around a fixed axis (freelook style) rather than a natural YAW (e.g. airplane)
  mCamera->yaw(mRotX);  
  mCamera->pitch(mRotY);
  if( keys[ 17 ] ){ mTranslateVector.z += -1.0; }  //W Key
  if( keys[ 31 ] ){ mTranslateVector.z += 1.0; }  // S key
  if( keys[ 30 ] ){ mTranslateVector.x += -1.0; } // A key
  if( keys[ 32 ] ){ mTranslateVector.x += 1.0; }  //D Key
  if( keys[0] ){/*views[active_view].zoom(-ZOOM_OFF); views[active_view].write();*/}   // ] and [ key same action??
  if( keys[51]){/*views[active_view].rotate_y(-ROTATE_OFF); views[active_view].write();*/} //, key
  if( keys[52]){/*views[active_view].rotate_y(ROTATE_OFF); views[active_view].write();*/}  //. key
  if( keys[44]){/*views[active_view].rotate_x(-ROTATE_OFF); views[active_view].write();*/} //z key
  if( keys[45]){/*views[active_view].rotate_x(ROTATE_OFF); views[active_view].write();*/} //x
  if( keys[16]){/*GfxView::change_angle(lx_angle, -LROTATE_OFF);*/} //q key
  if( keys[17]){/*GfxView::change_angle(lx_angle, LROTATE_OFF);*/} //w key
  if( keys[18]){/*GfxView::change_angle(ly_angle, -LROTATE_OFF);*/} //e key
  if( keys[19]){/*GfxView::change_angle(ly_angle, LROTATE_OFF);*/} //r key
  if( keys[25]){/*opt.perspective_mode = !opt.perspective_mode; */} //p key
  if( keys[33]){/*wireframe = !wireframe;*/} //f key

  Ogre::Vector3 direction = mCamera->getDirection();
  direction.y = 0;
  direction.normalise();
  Ogre::Vector3 rightV = mCamera->getRight();
  rightV.normalise();

  if( keys[200] ){ mCamera->move(direction*2); } // Up arrow key.
  if( keys[208] ){ mCamera->move(-direction*2); } // Down arrow key
  if( keys[203] ){ mCamera->move(-rightV*2); } // Down arrow key
  if( keys[205] ){ mCamera->move(rightV*2); } // Down arrow key
  if( keys[201] ){ mCamera->move(Ogre::Vector3(0.0, 1.0, 0.0)*2); } // Down arrow key
  if( keys[209] ){ mCamera->move(Ogre::Vector3(0.0, -1.0, 0.0)*2); } // Down arrow key

  mCamera->moveRelative(mTranslateVector);  //move the camera based upon changes in the mTranslateVector.
  mTranslateVector.x=0; mTranslateVector.y=0; mTranslateVector.z =0;  //reseset the translate vector after movement,.
}

void OgreGfxModuleFrameListener::showDebugOverlay(bool show){
  if (mDebugOverlay){
    if (show){
      mDebugOverlay->show();
    }
    else{
      mDebugOverlay->hide();
    }
  }
}

// Override frameStarted event to process that (don't care about frameEnded)
bool OgreGfxModuleFrameListener::frameStarted(const Ogre::FrameEvent& evt){
  ObjectManager::getSingleton().frameUpdate(evt);

  if(mWindow->isClosed())
    return false;
  
  if (!mInputTypeSwitchingOn){
    mInputDevice->capture();
  }

  
  if ( !mUseBufferedInputMouse || !mUseBufferedInputKeys){
    // one of the input modes is immediate, so setup what is needed for immediate mouse/key movement
    if (mTimeUntilNextToggle >= 0) 
      mTimeUntilNextToggle -= evt.timeSinceLastFrame;
    
    // If this is the first frame, pick a speed
    if (evt.timeSinceLastFrame == 0){
      mMoveScale = 1;
      mRotScale = 0.1;
    }
    // Otherwise scale movement units by time passed since last frame
    else{
      // Move about 100 units per second,
      mMoveScale = mMoveSpeed * evt.timeSinceLastFrame;
      // Take about 10 seconds for full rotation
      mRotScale = mRotateSpeed * evt.timeSinceLastFrame;
    }
    mRotX = 0;
    mRotY = 0;
    mTranslateVector = Ogre::Vector3::ZERO;
  }
  
  if (mUseBufferedInputKeys){
    // no need to do any processing here, it is handled by event processor and 
    // you get the results as KeyEvents
    if( keys[1] == true ){ return false; }
  }
  else{
    if (processUnbufferedKeyInput(evt) == false){  return false;  }
  }
  if (mUseBufferedInputMouse){
    // no need to do any processing here, it is handled by event processor and 
    // you get the results as MouseEvents
  }
  else{
    if (processUnbufferedMouseInput(evt) == false){
      return false;
    }
  }

  /*if ( !mUseBufferedInputMouse || !mUseBufferedInputKeys){
    // one of the input modes is immediate, so update the movement vector
    
    moveCamera();
    
  }*/
  processKeyChanges();

  //Check RaySceneQueries.

  if( mLeftMousePressed ){
    processLeftMouseButton();
    mLeftMousePressedLastFrame = true;;
  }else{
    mLeftMousePressedLastFrame = false;
  }
  if( mRightMousePressed ){
    processRightMouseButton();
    mRightMousePressedLastFrame = true;
  }else{
    mRightMousePressedLastFrame = false;
  }
 
  return true;
}

bool OgreGfxModuleFrameListener::processRightMouseButton(){
  if( mRightMousePressedLastFrame ){ return true; } 

  std::cout << "Processing Right Mouse button" << std::endl;

  Ogre::RaySceneQueryResult &result = mRaySceneQuery->execute();
  Ogre::RaySceneQueryResult::iterator itr;
  for ( itr = result.begin( ); itr != result.end(); itr++ ){ // loop through the results
    if ( itr->worldFragment ){  // Is this result a WorldFragment?
      Ogre::Vector3 location = itr->worldFragment->singleIntersection;  
      //At this point we have had terrain clicked on for each selected unit lets move.
      ObjectManager::getSingleton().processMovementRequest( location );
    }
  }
  return true;
}

bool OgreGfxModuleFrameListener::processLeftMouseButton(){
  if( mLeftMousePressedLastFrame ){ return true; } //if we already processed the mouse dont do it again.

  if( !keys[42] ){ ObjectManager::getSingleton().clearSelectedObjects();  } //if we dont click w/ shift clear selectdobj's

  
  Ogre::RaySceneQueryResult &result = mRaySceneQuery->execute();
  Ogre::RaySceneQueryResult::iterator itr;

  for ( itr = result.begin( ); itr != result.end(); itr++ ){ // loop through the results
    if ( itr->worldFragment ){  // Is this result a WorldFragment?
      Ogre::Vector3 location = itr->worldFragment->singleIntersection;
      //std::cout << "WorldFragment: (" << location.x << ", " << location.y << ", " << location.z << ")" << std::endl;
    } //  if worldfragement.
    else if ( itr->movable ){ // Is this result a MovableObject?
      //std::cout << "MovableObject: " << itr->movable->getName() << std::endl;
      BasicObject* tempObj = ObjectManager::getSingleton().getObjectByName(itr->movable->getName());
      if( tempObj != 0 ) { //if temp is not NULL(0), then a object was found.
	  std::cout << "Calling Process L-Mouse Button on validobj->";
        if( ObjectManager::getSingleton().doesSelectedObjectExist(tempObj) ){ //is obj already selected?
	  //object could not be added because it already exists so maybe its suppose to be remvoe check is shift key is down
	  if( keys[42] ){ //if so remove the object from list.
	    //std::cout << "Removed selected objected w/ Shift key called" << std::endl;
	    ObjectManager::getSingleton().removeSelectedObject(tempObj);
	    std::cout << "Removing Selected Object\n";
	  }//if
	} else { //if object doesnt exist.
	   ObjectManager::getSingleton().addSelectedObject(tempObj);
	   std::cout << "Adding Object\n";
	}
	return true;  //since we got a valid obj we are done, 
      }//if object valid.
    } // else if movable object.
  } // for all objects in ray query. 
  return true;
}

bool OgreGfxModuleFrameListener::frameEnded(const Ogre::FrameEvent& evt){
  updateStats();
  return true;
}

void OgreGfxModuleFrameListener::switchMouseMode() {
  mUseBufferedInputMouse = !mUseBufferedInputMouse;
  mInputDevice->setBufferedInput(mUseBufferedInputKeys, mUseBufferedInputMouse);
}
void OgreGfxModuleFrameListener::switchKeyMode() {
  mUseBufferedInputKeys = !mUseBufferedInputKeys;
  mInputDevice->setBufferedInput(mUseBufferedInputKeys, mUseBufferedInputMouse);
}

void OgreGfxModuleFrameListener::keyClicked(Ogre::KeyEvent* e) {
  if (e->getKeyChar() == 'm'){
    switchMouseMode();
  }
  else if (e->getKeyChar() == 'k'){    
    switchKeyMode();
  }
  
}
//Keyboard Events
void OgreGfxModuleFrameListener::keyPressed(Ogre::KeyEvent* e) {
  std::cout << "Key pressed:" << e->getKey() << std::endl; 
  keys[e->getKey()] = true;

  int key = KeyMap::ogre2Ascii(e);

  Game::pick_game(&(gsm->get_game()));
  //my = opt.win_h-1-my;  UNSURE IF NEEDED.
  //int key = KeyMap::convert_key(e->getKey());

  int mod = 0;   //clone glut key modifieds for orts event
  if( e->isAltDown() ){ mod = 4; }
  else if( e->isShiftDown() ){ mod = 1;}
  else if( e->isControlDown() ){ mod = 2;}   

  ObjAction *c = gui_obj ? gui_obj->action("event_key_down") : 0;
  if (c) {
    SActParam p;
    p.obj.push_back(gui_obj);
    p.val.push_back(stype(key));
    p.val.push_back(stype(mod));
    p.val.push_back(stype(lastMousePosX));
    p.val.push_back(stype(lastMousePosY));

    stype result = c->Act(p);
    if (!!result) return;
  }
}

void OgreGfxModuleFrameListener::keyReleased(Ogre::KeyEvent* e) {
  
   keys[e->getKey()] = false;
}

//Mouse Events.
void OgreGfxModuleFrameListener::mouseClicked(Ogre::MouseEvent* e){
  std::cout << "Mouse clicked called\n";
}
void OgreGfxModuleFrameListener::mouseEntered(Ogre::MouseEvent* e) {}
void OgreGfxModuleFrameListener::mouseExited(Ogre::MouseEvent* e)  {}
void OgreGfxModuleFrameListener::mousePressed(Ogre::MouseEvent* e) {
  //Button (From glut) has value for Left(0), Middle(1), Right(2). As ints.
  //This is expected inside p.val.push_back(stype(button))
  int glutKeyPressedAdaptor = 4; 

  Game::pick_game(&(gsm->get_game()));

  // Left mouse button down
  if (e->getButtonID() & Ogre::InputEvent::BUTTON0_MASK){ 
    mLeftMousePressed = true; 
    mRaySceneQuery->setSortByDistance( true );
    glutKeyPressedAdaptor = 0;
  } 

  // Right mouse button down
  else if (e->getButtonID() & Ogre::InputEvent::BUTTON1_MASK){
    CEGUI::MouseCursor::getSingleton().hide( );
    mMiddleMousePressed = true;
    glutKeyPressedAdaptor = 1;
  } 
  else if( e->getButtonID() & Ogre::InputEvent::BUTTON2_MASK){
    mRightMousePressed = true;
    glutKeyPressedAdaptor = 2;
  }

  //Trigger a GUI event from Mouse being pressed.
  //Button (From glut) has value for Left(0), Middle(1), RIght(2). As ints.
  ObjAction *c = gui_obj ? gui_obj->action("event_mouse_down") : 0;
  if (c) {
    SActParam p;
    p.obj.push_back(gui_obj);
    p.val.push_back(stype(lastMousePosX));
    p.val.push_back(stype(lastMousePosY));
    p.val.push_back(stype(glutKeyPressedAdaptor));
    
    stype result = c->Act(p);
    
    /*if (!!result) {
      cur_click_state = JUST_CLICKED;
      return;
    }*/
  }
}
void OgreGfxModuleFrameListener::mouseReleased(Ogre::MouseEvent* e){
  //Button (From glut) has value for Left(0), Middle(1), Right(2). As ints.
  //This is expected inside p.val.push_back(stype(button))
  int glutKeyPressedAdaptor = 4; 

  Game::pick_game(&(gsm->get_game()));

  // Left mouse 
  if (e->getButtonID() & Ogre::InputEvent::BUTTON0_MASK){ 
    glutKeyPressedAdaptor = 0;
    mLeftMousePressed = false;  
  } 
  // Middle Mouse
  else if (e->getButtonID() & Ogre::InputEvent::BUTTON1_MASK){
    glutKeyPressedAdaptor = 1;
    CEGUI::MouseCursor::getSingleton().show( );
    mMiddleMousePressed = false;
  }  //Right Mouse
  else if( e->getButtonID() & Ogre::InputEvent::BUTTON2_MASK){
    glutKeyPressedAdaptor = 2;
    mRightMousePressed = false;
  }

  //Trigger a GUI event from Mouse being released.
  ObjAction *c = gui_obj ? gui_obj->action("event_mouse_up") : 0;
  if (c) {
    SActParam p;
    p.obj.push_back(gui_obj);
    p.val.push_back(stype(lastMousePosX));
    p.val.push_back(stype(lastMousePosY));
    p.val.push_back(stype(glutKeyPressedAdaptor));
    stype result = c->Act(p);
    if (!!result) return;
  }
}

void OgreGfxModuleFrameListener::mouseMoved (Ogre::MouseEvent *e)  {
  lastMousePosX = (int)(e->getRelX());
  lastMousePosY = (int)(e->getRelY());
  CEGUI::System::getSingleton().injectMouseMove(e->getRelX() * mGUIRenderer->getWidth(), e->getRelY() * mGUIRenderer->getHeight());
  Ogre::Ray mouseRay = mCamera->getCameraToViewportRay( e->getX(), e->getY() );   // Setup the ray scene query
  mRaySceneQuery->setRay( mouseRay );    

  if( keys[29] ){  //Ctrl key.
    Ogre::RaySceneQueryResult &result = mRaySceneQuery->execute();
    Ogre::RaySceneQueryResult::iterator itr;
    for ( itr = result.begin( ); itr != result.end(); itr++ ){ // loop through the results
      if ( itr->worldFragment ){  // Is this result a WorldFragment?
        Ogre::Vector3 location = itr->worldFragment->singleIntersection; 
        std::cout << "Intersection with:" << location.x << ", " << location.y << ", " << location.z << std::endl; 
        //At this point we have had terrain clicked on for each selected unit lets move.
        Ogre::Vector3 prevLoc = mCamera->getPosition();
        mCamera->setPosition( location );
        mCamera->yaw( -e->getRelX() * mRotateSpeed );
        mCamera->pitch( -e->getRelY() * mRotateSpeed );
        mCamera->setPosition(prevLoc);
      }
    }
  }
}

void OgreGfxModuleFrameListener::mouseDragged (Ogre::MouseEvent *e){
       CEGUI::System::getSingleton().injectMouseMove(e->getRelX() * mGUIRenderer->getWidth(), e->getRelY() * mGUIRenderer->getHeight());

       // If we are dragging the left mouse button.
       if ( mLeftMousePressed ){}

       // If we are dragging the right mouse button.
       else if ( mMiddleMousePressed ){
           mCamera->yaw( -e->getRelX() * mRotateSpeed );
           mCamera->pitch( -e->getRelY() * mRotateSpeed );
       } 
}

void OgreGfxModuleFrameListener::process_changes(){

  Game* game = &(gsm->get_game());
  const GameChanges* changes = &(gsm->get_changes());
  const ActionChanges* action_changes = &(gsm->get_action_changes());

  //std::cout << "OgreGfxModule::process_changes() called" << std::endl;

  Game::pick_game(game);

#if 0
  std::cout << "new#=" << changes->new_objs.size()
       << " changed#= " << changes->changed_objs.size()
       << " vanished#= " << changes->vanished_objs.size()
       << " dead#= " << changes->dead_objs.size()
       << std::endl;
#endif

  
  // new tiles

  const real4 h_avg = 0.25/GameConst::HEIGHT_MULT;
  
  //FORALL (changes->new_tile_indexes, i) {
   /* terrain->new_tile(*i, game); 
    const GameTile &tile = game->get_map()(*i);
    sint4 h0, h1, h2, h3;
    Tile::Type tw, te;
    Tile::Split split;
    tile.get_topo(h0, h1, h2, h3, split);
    tile.get_type(tw, te);
    minimap->new_tile(game->get_map().ind2x(*i), game->get_map().ind2y(*i), 
                      static_cast<sint4>((h0+h1+h2+h3)*h_avg+0.5), tw);*/
  //std::cout << "New tiles added here" << std::endl;
    //}

  // create new gfx objects

  FORALL (changes->new_objs, i) {
    GameObj *gob = dynamic_cast<GameObj*>(*i);
    Ogre::String objName = ObjectManager::getSingleton().generateUniqueName();
    BasicObject *temp = ObjectManager::getSingleton().addObject(objName, gob);
    Ogre::Vector3 position = temp->objNode->getPosition();
    //now we need to find where the object intersects the terrain and push it that way in the Y direction.
    Ogre::Ray objPosRay( Ogre::Vector3(position.x, 5000.0f, position.z), Ogre::Vector3::NEGATIVE_UNIT_Y );
    mRaySceneQuery->setRay( objPosRay );
    Ogre::RaySceneQueryResult &result = mRaySceneQuery->execute();
    Ogre::RaySceneQueryResult::iterator itr = result.begin( );
    if ( itr != result.end() && itr->worldFragment ){  // Get the results, set the camera height
      Ogre::Real terrainHeight = itr->worldFragment->singleIntersection.y;
      if ((terrainHeight + 10.0f) > position.y)
        temp->objNode->setPosition( position.x, terrainHeight + 15.0f, position.z );
    }
  }
   /* if (gob == 0) continue;
    if (*gob->sod.shape == Object::SHAPE_UNDEF) continue; // don't know how to draw the object
    GfxGameObject *ggo = new GfxGameObject(gob);
    add_GfxGameObject(ggo);
    ObjAction *init = gob->action("GfxInit");
    if (init) {
      SActParam p;
      p.obj.push_back(gob);
      init->Act(p);
    } else {
      cout << gob->bp_name() << " object has no GfxInit function" << endl;
    }*/ 
  // update changed gfx objects
  FORALL (changes->changed_objs, i) {
    GameObj *gob = (*i)->get_GameObj();
    if (gob == 0) continue;
    std::cout << "Ogre Gfx Object Changed" << std::endl;
    ObjectManager::getSingleton().updateChangedObject(gob);
  }
  /*  FIND (game2gfx, j, gob);
    if (j == game2gfx.end()) continue;
    // update object in octree iff position or size has changed   
    if (gob->attr_changed("x") ||
        gob->attr_changed("y") ||
        gob->attr_changed("heading") ||
        gob->attr_changed("last_direction") ||
        gob->attr_changed("radius") ||
        gob->attr_changed("width") ||
        gob->attr_changed("height") ||
        gob->attr_changed("zcat")) {
      GfxGameObject *p = (*j).second;
      GfxObject *gfx = p->gfx_obj;
      if (gfx) octree->remove(gfx); 
      p->game_tick();
      sintptr *xptr = gob->get_int_ptr("x");
      if (!xptr || *xptr < 0) continue;
      
      if (gfx) octree->add(gfx);
    }*/
  //}
  
  // remove vanished gfx objects
  
  //FORALL (changes->vanished_objs, i) {
    /*GameObj *gob = dynamic_cast<GameObj*>(*i);
    if (gob == 0) continue;

    FIND (game2gfx, j, gob);
    if (j == game2gfx.end()) continue;
    
    GfxGameObject *ggo = j->second;
    bool is_static = false;
    if (ggo->gob) {
      sintptr *tmp = ggo->gob->get_int_ptr("static");
      if (tmp) is_static = *tmp;
    }
    
    if (is_static) { // implies that there is a gameobj

      static_objects.push_back(ggo);
      ggo->gob->add_ref();

    } else {
      gob->kill();
      game2gfx.erase(j);
      del_GfxGameObject(ggo);
    }

    ScriptIdSet::invalidate_id(stype(dynamic_cast<ScriptObj*>(gob))); // update id lists*/
  //}

  // remove dead GfxGameObjects
  
  //FORALL (changes->dead_objs, i) {
    /*GameObj *gob = (*i)->get_GameObj();
    if (gob == 0) continue;

    //    cout << "GfxModule::killing obj " << gob->get_refs() << endl;
    gob->kill();
    //    cout << "GfxModule:: " << gob->get_refs() << endl;
    FIND (game2gfx, j, gob);
    if (j == game2gfx.end()) continue;
    GfxGameObject *ggo = j->second;

    game2gfx.erase(j);
    del_GfxGameObject(ggo);
    
    ScriptIdSet::invalidate_id(stype(dynamic_cast<ScriptObj*>(gob))); // update id lists*/
    //}

  /*if (changes->vanished_objs.size() +
      changes->dead_objs.size() > 0)
    update_id_list();

  game->compute_view(game->get_client_player());

  check_static_objects();*/
  
  /*static bool firstTime = true;
  
  if(firstTime){
    firstTime = false;
    const Map<GameTile> playingMap = game->get_map();
  
    unsigned int typeW = 0;
    unsigned int typeE = 0;
    unsigned int typeN = 0;
    unsigned int typeS = 0;
  
    std::cout<<std::endl;
  
    int i = 0;
    int j = 0;
    
    for(int y = 0; y < battleField->rows; y++){
      for(int x = 0; x< battleField->columns; x++){
	const Tile tempTile = playingMap.get_tile(x,y);
	typeW = tempTile.get_typeW();
	typeE = tempTile.get_typeE();
	typeN = tempTile.get_typeN();
	typeS = tempTile.get_typeS();
    
	i = y;
	j = x;
	
	if(typeW || typeE || typeN || typeS){
	  std::cout<<"H";
	  if(!battleField->revealed[i][j]){
	    float foggyTileWidth = 1.0f/((float)(battleField->columns));
	    float foggyTileHieght= 1.0f/((float)(battleField->rows));
	    Ogre::Entity* newEnt = NULL;
	    float U1_offset = (float)(randInt(0,7))/8.0f;
	    float V1_offset = 0.125f;
	    if(!randInt(0,10))
	      V1_offset = 0.0f;
	    float U2_offset =((float)(j))/((float)(battleField->columns));
	    float V2_offset =((float)(i))/((float)(battleField->rows));
	    
	    newEnt = battleField->createNewBasicTile(
		       tempTile.height_at(0.5,0.5),
		       tempTile.height_at(0.5,0),tempTile.height_at(1,0),tempTile.height_at(1,0.5),tempTile.height_at(1,1),
		       tempTile.height_at(0.5,1),tempTile.height_at(0,1),tempTile.height_at(0,0.5),tempTile.height_at(0,0),
		       U1_offset, V1_offset, U1_offset + 0.125f, V1_offset,
		       U1_offset, V1_offset + 0.125f, U1_offset + 0.125f, V1_offset + 0.125f,
		       U2_offset, V2_offset, U2_offset + foggyTileWidth, V2_offset,
		       U2_offset, V2_offset + foggyTileHieght, U2_offset + foggyTileWidth, V2_offset + foggyTileHieght);
	    
	    battleField->addTileToGeometry(i, j, newEnt);
	    }
	  battleField->setVisibility(i,j,true, true);
	}
	else
	  std::cout<<".";
      }
      std::cout<<std::endl;
    }
    
    battleField->refreshGeometry();
    }*/
  battleField->updateTemporalFade(4);

}

