// $Id: OSL_Parser.C,v 1.3 2004/11/21 22:56:10 orts_mburo Exp $

// (c) Jonathan Yip, licensed under the GPL

#include "OSL_Parser.H"
#include "OSL_RecStream.H"
#include "Lexer.H"

using namespace std;

#define DEBUG 1

#define DEBUGSTR errstr << "DEBUG function: " 

#define PARSE_ERR(s) {                                                  \
  errstr << token_stream[token_stream_ptr].file << "(" << token_stream[token_stream_ptr].line_num << "): " << s \
         << endl; exit(20);                                             \
}

#define PARSE_ERR2(s1, s2) {                                                  \
  errstr << token_stream[token_stream_ptr].file << "(" << token_stream[token_stream_ptr].line_num << "): " << s1 << " " << s2 \
         << endl; exit(20);                                             \
}

/*
#define PARSE_ERR2(s1, s2) {                                            \
  errstr << lex->cur_file() << "(" << lex->cur_line_num() << "): "      \
         << s1 << " " << s2 << endl; exit(20);                          \
}
*/

#define PARSE_ERR4(file, line, pos, s) {                                \
  errstr << file << "(" << line << ":" << pos << "): " << s             \
	 << endl; exit(20);                                             \
}

#define PARSE_ERR5(file, line, pos, s1, s2) {                           \
  errstr << file << "(" << line << ":" << pos << "): " << s1 << " "     \
	 << s2 << endl; exit(20);                                       \
}


#define PROGRAM_ERR ERR("Should not reach here, error in C++ program")

namespace OSL {

  Parser::Parser(Program &p, const string &file_name) : prog(p) {
    lex = new Lexer(rs);
    FILE *fp = fopen(file_name.c_str(), "r");
    
    if (!fp) ERR2("File not present:", file_name);
    rs.push(fp, file_name);
    
    curr_class = 0;
    curr_func = 0;
    curr_static = false;
  }

  inline const Token Parser::get_token() {
    Token tk;

    try {
      tk = lex->get_token();
      // fixme pos number
      token_stream.push_back( TokenInfo(tk, lex->cur_line_num(),
					0, lex->cur_file()) );
      tk.write(); cout << endl;
    }
    catch (Lexer::SyntaxErrorException &e) {
      cout << "Syntax error: " << e.msg << endl;
    }
    catch (RecStream::ReadErrorException &e) {
      cout << "Read error: " << e.msg << endl;
    }
    return tk;
  }
  
  inline const Token Parser::peek_token() {
    Token tk;

    try {
      tk = lex->peek_token();
    }
    catch (Lexer::SyntaxErrorException &e) {
      cout << "Syntax error: " << e.msg << endl;
    }
    catch (RecStream::ReadErrorException &e) {
      cout << "Read error: " << e.msg << endl;
    }
    return tk;
  }

  // fixme
  inline const Token Parser::get_token_info() {
    token_stream[token_stream_ptr].tk.write();
    cout << endl;
    return token_stream[token_stream_ptr++].tk;
  }

  inline const Token Parser::peek_token_info() {
    return token_stream[token_stream_ptr].tk;
  }
  
  inline Parser::var_location Parser::find_variable(const string var_name) {
#if DEBUG
    DEBUGSTR << "find_variable(const string var_name)" << endl
	     << "\tParams: *" << var_name << "*" << endl;
#endif
    uint4 start = local_vars.size()-1;
    for (sint4 i=start; i>=0; --i)
      if (local_vars[i].first == var_name) 
	return var_location(VAR_LOCAL, local_vars[i].second);

    start = param_vars.size()-1;
    for (sint4 i=start; i>=0; --i)
      if (param_vars[i].first == var_name) { cout << "W " << param_vars[i].second << endl;
	return var_location(VAR_PARAM, param_vars[i].second); }

    // Look in current class for member variables
    start = curr_class->inst_data.size()-1;
    for (sint4 i=start; i>=0; --i)
      if (curr_class->inst_data[i].name == var_name)
	return var_location(VAR_INST_MEMBER, i);

    start = curr_class->static_data.size()-1;
    for (sint4 i = start; i>=0; --i)
      if (curr_class->static_data[i].name == var_name)
	return var_location(VAR_STATIC_MEMBER, i);

    return var_location(VAR_UNKNOWN, 0);
  }

  inline void Parser::add_block() {
#if DEBUG
    DEBUGSTR << "add_block()" << endl;
#endif
    local_scopes.push(local_vars.size());
  }

  inline void Parser::del_block() {
#if DEBUG
    DEBUGSTR << "del_block()" << endl;
#endif
    uint4 pop_to = local_scopes.top();
    local_scopes.pop();
    
    uint4 pops = local_vars.size() - pop_to;
    // fixme have to check
    cout << "pops " << pops << endl;
    FORU(i, pops) local_vars.pop_back();
    //    while (local_vars.size() != pop_to) {local_vars.pop_back(); cout << "0*" << endl;}

  }

  void Parser::parse() { 
#if DEBUG
    DEBUGSTR << "parse()" << endl;
#endif

    parse_declarations();

    token_stream_ptr = 0;
    parse_definitions();


    // Ensure that all classes are defined and functions defined
    // fixme add check
  }

  void Parser::parse_declarations() { 
#if DEBUG
    DEBUGSTR << "parse_declarations()" << endl;
#endif
    
    FOREVER {
      Token tk = get_token();
      TokenType type = tk.get_type();

      // Function definition outside class definition
      if (valid_type(type)) {
	Token tk2 = get_token();

	// Class member function
	if (tk2.get_type() == TK_ID) {
	  const string class_name = tk.get_value();
	  if (prog.class_ids.find(class_name) != prog.class_ids.end())
	    ERR("Unknown class");

	  curr_class = &prog.classes[prog.class_ids[class_name]];
	  
	  if (get_token().get_type() != TK_SCOPE)
	    ERR("Expecting '::'");

	  tk = get_token();
	  if (tk.get_type() != TK_ID)
	    PARSE_ERR("Expecting function name");

	  const string func_name = tk.get_value();
	  string stored_name = func_name;

	  if (get_token().get_type() != TK_LPAR)
	    ERR("Expecting '('");

	  // Get parameters
	  FOREVER {
	    tk = get_token();
	    if (tk.get_type() == TK_RPAR) break;
	    
	    TypeIndex t;
	    switch (tk.get_type()) {
	    case TK_BOOL: t = Type::TC_sint1; break;
	    case TK_BYTE: t = Type::TC_sint1; break;
	    case TK_UBYTE: t = Type::TC_uint1; break;
	    case TK_SHORT: t = Type::TC_sint2; break;
	    case TK_USHORT: t = Type::TC_uint2; break;
	    case TK_INT: t = Type::TC_sint4; break;
	    case TK_UINT: t = Type::TC_uint4; break;
	    case TK_LONG: t = Type::TC_sint8; break;
	    case TK_ULONG: t = Type::TC_uint8; break;
	    case TK_FLOAT: t = Type::TC_real4; break;
	    case TK_DOUBLE: t = Type::TC_real8; break;
	    case TK_STRING: t = Type::TC_string; break;
	    case TK_ID: {
	      string class_name = tk.get_value();
	      if (prog.class_ids.find(class_name) == prog.class_ids.end())
		prog.add_class(class_name);
	      
	      t = prog.classes[prog.class_ids[class_name]].ref_tind;
	      break;
	    }
	    default:
	      ERR("Unexpected token");
	    }
	    
	    stringstream s;
	    s << "*" << t;
	    stored_name += s.str();

	    tk = get_token();
	    if (tk.get_type() != TK_ID)
	      ERR("Expecting variable name for parameters");
	    
	    if (peek_token().get_type() == TK_COMMA) get_token();
	    else if (peek_token().get_type() != TK_RPAR)
	      ERR("Unexpected token");
	  }

	  if (curr_class->static_func_ids.find(stored_name) ==
	      curr_class->static_func_ids.end() &&
	      curr_class->virt_func_ids.find(stored_name) ==
	      curr_class->virt_func_ids.end())
	    ERR("Unknown function");	  
	}
	// Class constructor
	else if (tk2.get_type() == TK_SCOPE) {
	  const string class_name = tk.get_value();
	  if (prog.class_ids.find(class_name) != prog.class_ids.end())
	    ERR("Unknown class");

	  curr_class = &prog.classes[prog.class_ids[class_name]];

	  if (get_token().get_value() != class_name)
	    ERR("Constructor name not same as class name");

	  if (get_token().get_type() != TK_LPAR)
	    ERR("Expecting '('");

	  // fixme can i go something like a::a(a abc, int x) and
	  // have 'abc' be the 'this' being called?

	  string stored_name = class_name;

	  // Get parameters
	  FOREVER {
	    tk = get_token();
	    	    if (tk.get_type() == TK_RPAR) break;
	    
	    TypeIndex t;
	    switch (tk.get_type()) {
	    case TK_BOOL: t = Type::TC_sint1; break;
	    case TK_BYTE: t = Type::TC_sint1; break;
	    case TK_UBYTE: t = Type::TC_uint1; break;
	    case TK_SHORT: t = Type::TC_sint2; break;
	    case TK_USHORT: t = Type::TC_uint2; break;
	    case TK_INT: t = Type::TC_sint4; break;
	    case TK_UINT: t = Type::TC_uint4; break;
	    case TK_LONG: t = Type::TC_sint8; break;
	    case TK_ULONG: t = Type::TC_uint8; break;
	    case TK_FLOAT: t = Type::TC_real4; break;
	    case TK_DOUBLE: t = Type::TC_real8; break;
	    case TK_STRING: t = Type::TC_string; break;
	    case TK_ID: {
	      string class_name = tk.get_value();
	      if (prog.class_ids.find(class_name) == prog.class_ids.end())
		prog.add_class(class_name);
	      
	      t = prog.classes[prog.class_ids[class_name]].ref_tind;
	      break;
	    }
	    default:
	      ERR("Unexpected token");
	    }
	    
	    stringstream s;
	    s << "*" << t;
	    stored_name += s.str();

	    tk = get_token();
	    if (tk.get_type() != TK_ID)
	      ERR("Expecting variable name for parameters");
	    
	    if (peek_token().get_type() == TK_COMMA) get_token();
	    else if (peek_token().get_type() != TK_RPAR)
	      ERR("Unexpected token");
	  }
	  
	  // fixme check can constructor be static?
	  if (curr_class->virt_func_ids.find(stored_name) ==
	      curr_class->virt_func_ids.end())
	    ERR("Unknown constructor");
	}
	else ERR("Unexpected token");

	if (get_token().get_type() != TK_LBRACE)
	  ERR("Expecting '{'");

	// Skip over the function definition
	sint4 braces = 1;
	FOREVER {
	  tk = get_token();
	  if (tk.get_type() == TK_LBRACE) ++braces;
	  else if (tk.get_type() == TK_RBRACE) --braces;

	  if (braces == 0) break;
	}
      }
      else if (type == TK_CLASS) {
	tk = get_token();
	if (tk.get_type() != TK_ID)
	  ERR("expecting name for class");

	const string class_name = tk.get_value();

	if (prog.class_ids.find(class_name) == prog.class_ids.end())
	  curr_class = prog.add_class(class_name);
	else {
	  if (prog.classes[prog.class_ids[class_name]].defined)
	    ERR("class already used" + class_name);
	}
	tk = get_token();

	if (tk.get_type() == TK_COLON) {
	  tk = get_token();
	  if (tk.get_type() != TK_ID)
	    ERR("expecting name for superclass");

	  const string super_class = tk.get_value();

	  // Need to make sure super_class != derived class
	  if (class_name == super_class)
	    ERR("Derived class cannot be the same as the base class");
	  
	  if (prog.class_ids.find(super_class) == prog.class_ids.end())
	    ERR("Superclass does not exist");
	  
	  curr_class->super_class_index = prog.class_ids[super_class];
	  
	  // inherit
	  curr_class->inherit(prog.classes[curr_class->super_class_index]);

	  tk = get_token();	  
	}


	if (tk.get_type() != TK_LBRACE)
	  ERR("expecting class definition");

	parse_class_def();
	curr_class = 0;
      }
      else if (type == TK_EOF) return;
      else ERR("Unexpected token");
      
    }

  }

  
  void Parser::parse_class_def() {
#if DEBUG
    DEBUGSTR << "parse_class_def()" << endl;
#endif  
    // fixme in future check to see if defined already
    cout << curr_class << endl;
    if (curr_class->defined)
      ERR("Class already defined");
      //PARSE_ERR2("Class already defined", curr_class->name);
    curr_class->defined = true;
    curr_mod = PRIVATE;

    // Right now get public member variables and functions
    FOREVER {
      const Token type = get_token();
      switch(type.get_type()) {
      case TK_PUBLIC:   
	if (get_token().get_type() != TK_COLON)
	  ERR("Expecting ':' after modifier");
	curr_mod = PUBLIC;
	break;
      case TK_PROTECTED: 
	if (get_token().get_type() != TK_COLON)
	  ERR("Expecting ':' after modifier");
	curr_mod = PROTECTED;
	break;
      case TK_PRIVATE:
	if (get_token().get_type() != TK_COLON)
	  ERR("Expecting ':' after modifier");
	curr_mod = PRIVATE; 
	break;
      case TK_STATIC: 
	{
	  const Token type2 = get_token();
	  if (type2.get_type() == TK_VOID) parse_void_func(true);
	  else parse_var_or_array(type2, true);
	  break;
	}
      case TK_VOID:
	parse_void_func(false);
	break;
      case TK_SEMICOLON:
	break;
      case TK_RBRACE:
	return;
      default:
	parse_var_or_array(type, false);
      }
    }

  }

  void Parser::parse_void_func(bool is_static) {
#if DEBUG
    DEBUGSTR << "parse_void_func(bool is_static)" << endl
	     << "\tParams: *" << is_static << "*" << endl;
#endif
    const Token tk = get_token();
    
    // Check for conflicts
    if (tk.get_type() != TK_ID)
      ERR("Expecting function name");

    const string func_name = tk.get_value();

    parse_func_dec(TK_VOID, func_name, is_static, false);
  }

  void Parser::parse_var_or_array(const Token &type, const bool is_static) {
#if DEBUG
    DEBUGSTR << "parse_var_or_array(const Token &type, const bool is_static)" 
	     << endl << "\tParams: *" << type.get_value() << "*" 
	     << is_static << "*" << endl;
#endif
    const Token tk = peek_token();

    switch (type.get_type()) {
    case TK_BOOL:
    case TK_BYTE:
    case TK_UBYTE:
    case TK_SHORT:
    case TK_USHORT:
    case TK_INT:
    case TK_UINT:
    case TK_LONG:
    case TK_ULONG:
    case TK_FLOAT:
    case TK_DOUBLE:
    case TK_STRING:
      if (tk.get_type() == TK_LBRACKET) parse_array_dec(type, is_static);
      else parse_var_dec(type, is_static);
      return;
    case TK_ID:
      // look in classes
      if (prog.class_ids.find(type.get_value()) != 
	  prog.class_ids.end()) {
	if (tk.get_type() == TK_LBRACKET) parse_array_dec(type, is_static);
	else parse_var_dec(type, is_static);
      }
      else ERR("Unexpected token");
      return;
    default:
      ERR("Unexpected error");
    }

  }

  void Parser::parse_array_dec(const Token &type, const bool is_static) {
#if DEBUG
    DEBUGSTR << "parse_array_dec(const Token &type, const bool is_static)"
	     << endl << "\tParams: *" << type.get_value() << "*" 
	     << is_static << "*" << endl;
#endif
    get_token(); // the '['
    Token tk = get_token();
    if (tk.get_type() != TK_INTC && tk.get_type() != TK_UINTC)
      ERR("Expecting size");

    sint4 size = atoi(tk.get_value().c_str());

    if (size < 0) ERR("Array size needs to be bigger than 0");
    
    if (get_token().get_type() != TK_RBRACKET)
      ERR("Expecting ']'");

    // if there is a size...need to add a new array statement, otherwise
    // have to wait for array to be newed before it can be used

    // fixme need to see if name conflicts
    tk = get_token();
    const string name = tk.get_value();


    // fixme
#if 0
    DataMember dm;
    TypeIndex t;
    t.cat = Type::TC_array;
    switch (type.get_type()) {
    case TK_BOOL:   t.T_cat = Type::TC_sint1; break;
    case TK_BYTE:   t.T_cat = Type::TC_sint1; break;
    case TK_UBYTE:  t.T_cat = Type::TC_uint1; break;
    case TK_SHORT:  t.T_cat = Type::TC_sint2; break;
    case TK_USHORT: t.T_cat = Type::TC_uint2; break;
    case TK_INT:    t.T_cat = Type::TC_sint4; break;
    case TK_UINT:   t.T_cat = Type::TC_uint4; break;
    case TK_LONG:   t.T_cat = Type::TC_sint8; break;
    case TK_ULONG:  t.T_cat = Type::TC_uint8; break;
    case TK_FLOAT:  t.T_cat = Type::TC_real4; break;
    case TK_DOUBLE: t.T_cat = Type::TC_real8; break;
    case TK_STRING: t.T_cat = Type::TC_string; break;
    default: // fixme
      ;
    }
    dm.name = name;
    dm.type = t;
    if (is_static) { 
      dm.offset = curr_class->static_data.size();
      dm.is_static = true;
      curr_class->static_data.push_back(dm);
    }
    else {
      dm.offset = curr_class->inst_data.size();
      dm.offset = false;
      curr_class->inst_data.push_back(dm);
    }
#endif
  }




  void Parser::parse_var_dec(const Token &type, const bool is_static) {
#if DEBUG
    DEBUGSTR << "parse_var_dec(const Token &type, const bool is_static)"
	     << endl << "\tParams: *" << type.get_value() << "*" << is_static
	     << "*" << endl;
#endif

    if (type.get_value() == curr_class->name && 
	peek_token().get_type() == TK_LPAR)
      return parse_constructor(type);

    Token tk = get_token();
    if (tk.get_type() != TK_ID)
      ERR("Expecting name for variable");
    string name = tk.get_value();

    // function declaration
    if (peek_token().get_type() == TK_LPAR) {
      parse_func_dec(type, name, is_static, false);
      return;
    }

    FOREVER {
      DataMember dm;
      TypeIndex t;
      switch (type.get_type()) {
      case TK_BOOL:   t = Type::TC_sint1; break;
      case TK_BYTE:   t = Type::TC_sint1; break;
      case TK_UBYTE:  t = Type::TC_uint1; break;
      case TK_SHORT:  t = Type::TC_sint2; break;
      case TK_USHORT: t = Type::TC_uint2; break;
      case TK_INT:    t = Type::TC_sint4; break;
      case TK_UINT:   t = Type::TC_uint4; break;
      case TK_LONG:   t = Type::TC_sint8; break;
      case TK_ULONG:  t = Type::TC_uint8; break;
      case TK_FLOAT:  t = Type::TC_real4; break;
      case TK_DOUBLE: t = Type::TC_real8; break;
      case TK_STRING: t = Type::TC_string; break;

      default:
	ERR("FIXME");
      }

      dm.name = name;
      dm.type = t;
      dm.mod = curr_mod;
      if (is_static) { 
	dm.offset = curr_class->static_data.size();
	dm.is_static = true;
	curr_class->static_data_ids[dm.name] = dm.offset;
	curr_class->static_data.push_back(dm);
      } 
      else {
	dm.offset = curr_class->inst_data.size();
	dm.is_static = false;
	curr_class->inst_data_ids[dm.name] = dm.offset;
	curr_class->inst_data.push_back(dm);
      }

      cout << "NOW SIZE IS " << curr_class->inst_data.size() << endl;

      if (peek_token().get_type() == TK_ASSIGN) {
	// fixme
      }
	
      tk = peek_token();

      if (tk.get_type() == TK_COMMA) {
	get_token(); // the comma
	tk = get_token(); 
	if (tk.get_type() != TK_ID)
	  ERR("Expecting name for variable");
	name = tk.get_value();
      }
      else if (tk.get_type() == TK_SEMICOLON) {
	get_token(); // the semicolon
	return;
      }
      else return;
    }
      
  }


  void Parser::parse_constructor(const Token &type) {
#if DEBUG
    DEBUGSTR << "parse_constructor(const Token &type)" << endl
	     << "\tParams: *" << type.get_value() << "*" << endl;
#endif

    // Get paramaters
    // similar to parse_func_dec() ???

    parse_func_dec(type, type.get_value(), false, true);

      
  }


  void Parser::parse_func_dec(const Token &return_type, 
			      const string &func_name,
			      const bool is_static,
			      const bool is_constructor) {
#if DEBUG
    DEBUGSTR << "parse_func_dec(const Token &return_type, const string " 
	     << "&func_name, const bool is_static, const bool is_constructor)"
	     << endl
	     << "\tParams: *" << return_type.get_value() << "*" << func_name
	     << "*" << is_static << "*" << is_constructor << "*" << endl;
#endif

    // fixme
    FuncMember fm;
    fm.name = func_name;
    fm.mod = curr_mod;
    fm.is_static = is_static;
    // fixme: not always true
    //    fm.class_id = prog.classes.size()-1;
    fm.class_id = prog.class_ids[curr_class->name];


    // fixme void type...
    switch(return_type.get_type()) {
    case TK_BOOL: fm.ret_tind = Type::TC_sint1; break;
    case TK_BYTE: fm.ret_tind = Type::TC_sint1; break;
    case TK_UBYTE: fm.ret_tind = Type::TC_uint1; break;
    case TK_SHORT: fm.ret_tind = Type::TC_sint2; break;
    case TK_USHORT: fm.ret_tind = Type::TC_uint2; break;
    case TK_INT: fm.ret_tind = Type::TC_sint4; break;
    case TK_UINT: fm.ret_tind = Type::TC_uint4; break;
    case TK_LONG: fm.ret_tind = Type::TC_sint8; break;
    case TK_ULONG: fm.ret_tind = Type::TC_uint8; break;
    case TK_FLOAT: fm.ret_tind = Type::TC_real4; break;
    case TK_DOUBLE: fm.ret_tind = Type::TC_real8; break;
    case TK_STRING: fm.ret_tind = Type::TC_string; break;
    case TK_ID:
      if (prog.class_ids.find(return_type.get_value()) == prog.class_ids.end())
	PARSE_ERR("Unknown type");
      fm.ret_tind = prog.classes[prog.class_ids[return_type.get_value()]].ref_tind;
      break;
    default: // fixme
      ERR("fixme");
    }

    if (get_token().get_type() != TK_LPAR)
      PARSE_ERR("'(' expected after function name");


    // fixme this is wrong....
    // no check right now for 'this' pointer so it works
    // Push this pointer for non-static functions
    if (!is_static) {
      TypeIndex t;
      t = 0; // fixme need current function type
      fm.par_types.push_back(t);
    }

    string stored_name = fm.name;

    // Get parameters
    param_vars.clear();
    FOREVER {
      Token tk = get_token();
      if (tk.get_type() == TK_RPAR) break;

      TypeIndex t;
      switch (tk.get_type()) {
      case TK_BOOL: t = Type::TC_sint1; break;
      case TK_BYTE: t = Type::TC_sint1; break;
      case TK_UBYTE: t = Type::TC_uint1; break;
      case TK_SHORT: t = Type::TC_sint2; break;
      case TK_USHORT: t = Type::TC_uint2; break;
      case TK_INT: t = Type::TC_sint4; break;
      case TK_UINT: t = Type::TC_uint4; break;
      case TK_LONG: t = Type::TC_sint8; break;
      case TK_ULONG: t = Type::TC_uint8; break;
      case TK_FLOAT: t = Type::TC_real4; break;
      case TK_DOUBLE: t = Type::TC_real8; break;
      case TK_STRING: t = Type::TC_string; break;
      default: // fixme
	;
      }
      cout << "WHY " << t << endl;
      fm.par_types.push_back(t);
      
      stringstream s;
      s << "*" << t;
      stored_name += s.str();

      // fixme: For now assume var names have to be here
      tk = get_token();
      if (tk.get_type() != TK_ID)
	PARSE_ERR("Expecting variable name for parameters");

      param_vars.push_back(var_info(tk.get_value(), fm.par_types.size()-1));

      if (peek_token().get_type() == TK_COMMA) get_token();
      else if (peek_token().get_type() != TK_RPAR)
	PARSE_ERR2("Unexpected token in param list", peek_token().get_value());
    }

    // fixme need to check for conflicts.... and later on to see if fns defined

    // Set function into current class
    if (is_static) {
      cout << "Setting in func " << stored_name << " in " << prog.classes.size() << " " << curr_class->static_funcs.size() << endl;
      curr_class->static_func_ids[stored_name] = 
	curr_class->static_funcs.size();
      curr_class->static_funcs.push_back(fm);
      
    }
    else {
      cout << "Setting in func " << stored_name << " in " << prog.classes.size() << " " << curr_class->virt_funcs.size() << endl;
      curr_class->virt_func_ids[stored_name] = 
	curr_class->virt_funcs.size();
      curr_class->virt_funcs.push_back(fm);
    }

    cout << "WHAT IS MY SIZE " << curr_class->virt_funcs.size() << endl;

    Token tk = peek_token();
    if (tk.get_type() == TK_SEMICOLON) {
      get_token(); // the semicolon
      return;
    }
    else if (tk.get_type() == TK_LBRACE) {
      get_token(); // the '{'

      sint4 braces = 1;

      FOREVER {
	tk = get_token();
	if (tk.get_type() == TK_LBRACE) ++braces;
	else if (tk.get_type() == TK_RBRACE) --braces;
	
	if (braces == 0) break;
      }

      /*
      uint4 loc;
      if (is_static) {
	loc = curr_class->static_funcs.size() - 1;
	curr_func = &curr_class->static_funcs[loc];
      }
      else {
	loc = curr_class->virt_funcs.size() - 1;
	curr_func = &curr_class->virt_funcs[loc];
      }

      cout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
      cout << curr_func->loc_types.size() << endl;
      parse_func_def();

      if (is_constructor && !has_return)
	curr_func->add_code(Opcodes::RETURN);

      */
      curr_func = 0;
    }
    else
      ERR("Unexpected token");
  }


  //////////////////////////////////////////////////////////////////
  // 2nd pass


  void Parser::parse_definitions() { 
#if DEBUG
    DEBUGSTR << "parse_definitions()" << endl;
#endif
    
    // A Class, class member function
    FOREVER {
      Token tk = get_token_info();
      TokenType type = tk.get_type();

      // fixme for class constructor
      if (valid_type(type)) {
	Token tk2 = get_token_info();
	if (tk2.get_type() == TK_ID) {
	  const string class_name = tk.get_value();
	  if (prog.class_ids.find(class_name) != prog.class_ids.end())
	    PARSE_ERR2("Unknown class", class_name);
	  
	  curr_class = &prog.classes[prog.class_ids[class_name]];
	  
	  // Should be the scope resolution operator
	  if (get_token_info().get_type() != TK_SCOPE)
	    PARSE_ERR("Expecting '::'");
	  
	  tk = get_token_info();
	  if (tk.get_type() != TK_ID)
	    PARSE_ERR("Expecting function name");

	  const string func_name = tk.get_value();
	  string stored_name = func_name;

	  if (get_token_info().get_type() != TK_LPAR)
	    PARSE_ERR("Expecting '('");
	  
	  // Get parameters
	  FOREVER {
	    tk = get_token_info();
	    if (tk.get_type() == TK_RPAR) break;
	    
	    TypeIndex t;
	    switch (tk.get_type()) {
	    case TK_BOOL: t = Type::TC_sint1; break;
	    case TK_BYTE: t = Type::TC_sint1; break;
	    case TK_UBYTE: t = Type::TC_uint1; break;
	    case TK_SHORT: t = Type::TC_sint2; break;
	    case TK_USHORT: t = Type::TC_uint2; break;
	    case TK_INT: t = Type::TC_sint4; break;
	    case TK_UINT: t = Type::TC_uint4; break;
	    case TK_LONG: t = Type::TC_sint8; break;
	    case TK_ULONG: t = Type::TC_uint8; break;
	    case TK_FLOAT: t = Type::TC_real4; break;
	    case TK_DOUBLE: t = Type::TC_real8; break;
	    case TK_STRING: t = Type::TC_string; break;
	    case TK_ID: {
	      string class_name = tk.get_value();
	      if (prog.class_ids.find(class_name) == prog.class_ids.end())
		prog.add_class(class_name);
	      
	      t = prog.classes[prog.class_ids[class_name]].ref_tind;
	      break;
	    }
	    default:
	      PARSE_ERR("Unexpected token");
	    }
	    
	    stringstream s;
	    s << "*" << t;
	    stored_name += s.str();

	    tk = get_token_info();
	    if (tk.get_type() != TK_ID)
	      PARSE_ERR("Expecting variable name for parameters");
	    
	    if (peek_token_info().get_type() == TK_COMMA) get_token_info();
	    else if (peek_token_info().get_type() != TK_RPAR)
	      PARSE_ERR("Unexpected token");
	  }
	  if (curr_class->static_func_ids.find(stored_name) !=
	      curr_class->static_func_ids.end())
	    curr_func = &curr_class->static_funcs[curr_class->static_func_ids[stored_name]];
	  else if (curr_class->virt_func_ids.find(stored_name) !=
		   curr_class->virt_func_ids.end())
	    curr_func = &curr_class->virt_funcs[curr_class->virt_func_ids[stored_name]];
	  else PROGRAM_ERR;
	}
	// Class constructor
	else if (tk2.get_type() == TK_SCOPE) {
	  	  const string class_name = tk.get_value();
	  if (prog.class_ids.find(class_name) != prog.class_ids.end())
	    ERR("Unknown class");

	  curr_class = &prog.classes[prog.class_ids[class_name]];

	  if (get_token_info().get_value() != class_name)
	    PARSE_ERR("Constructor name not same as class name");

	  if (get_token_info().get_type() != TK_LPAR)
	    PARSE_ERR("Expecting '('");

	  // fixme can i go something like a::a(a abc, int x) and
	  // have 'abc' be the 'this' being called?

	  string stored_name = class_name;

	  // Get parameters
	  FOREVER {
	    tk = get_token_info();
	    if (tk.get_type() == TK_RPAR) break;
	    
	    TypeIndex t;
	    switch (tk.get_type()) {
	    case TK_BOOL: t = Type::TC_sint1; break;
	    case TK_BYTE: t = Type::TC_sint1; break;
	    case TK_UBYTE: t = Type::TC_uint1; break;
	    case TK_SHORT: t = Type::TC_sint2; break;
	    case TK_USHORT: t = Type::TC_uint2; break;
	    case TK_INT: t = Type::TC_sint4; break;
	    case TK_UINT: t = Type::TC_uint4; break;
	    case TK_LONG: t = Type::TC_sint8; break;
	    case TK_ULONG: t = Type::TC_uint8; break;
	    case TK_FLOAT: t = Type::TC_real4; break;
	    case TK_DOUBLE: t = Type::TC_real8; break;
	    case TK_STRING: t = Type::TC_string; break;
	    case TK_ID: {
	      string class_name = tk.get_value();
	      if (prog.class_ids.find(class_name) == prog.class_ids.end())
		prog.add_class(class_name);
	      
	      t = prog.classes[prog.class_ids[class_name]].ref_tind;
	      break;
	    }
	    default:
	      PARSE_ERR("Unexpected token");
	    }
	    
	    stringstream s;
	    s << "*" << t;
	    stored_name += s.str();

	    tk = get_token_info();
	    if (tk.get_type() != TK_ID)
	      PARSE_ERR("Expecting variable name for parameters");
	    
	    if (peek_token_info().get_type() == TK_COMMA) get_token_info();
	    else if (peek_token_info().get_type() != TK_RPAR)
	      PARSE_ERR("Unexpected token");
	  }
	  
	  // fixme check can constructor be static?
	  if (curr_class->virt_func_ids.find(stored_name) !=
	      curr_class->virt_func_ids.end())
	    curr_func = &curr_class->virt_funcs[curr_class->virt_func_ids[stored_name]];
	  else PROGRAM_ERR;
	  
	  
	}
	else
	  PARSE_ERR("Expecting class name for function definition");

	// fixme get the func params in order to set curr_func
	if (get_token_info().get_type() != TK_LBRACE)
	  PARSE_ERR("Expecting '{'");
	
	ERR("DO NOT ");
	parse_func_def();
	curr_func = 0;

      }
      else if (type == TK_CLASS) {
	parse_class_dec();
      }
      else if (type == TK_EOF) return;
      else PARSE_ERR2("Unexpected token", tk.get_value());
    }

  }


  
  void Parser::parse_func_def() {
#if DEBUG
    DEBUGSTR << "parse_func_def()" << endl;
#endif

    // Reset the local variable counter
    curr_local = 0;
    local_scopes.push(0);
    if (!local_vars.empty())
      ERR("PROGRAM ERROR");
    local_vars.clear();
    has_return = false;

    // Get statements
    parse_stmt_list(TK_RBRACE);

    // check for void....
    // check for return

    // see what i get
    cout << "DONE FUNCTION " << endl;

    cout << curr_func->name << endl;
    prog.write_code(*curr_func, cout);
  }


  void Parser::parse_func_dec2(const Token &return_type, 
			      const string &func_name,
			      const bool is_static,
			      const bool is_constructor) {
#if DEBUG
    DEBUGSTR << "parse_func_dec2(const Token &return_type, const string " 
	     << "&func_name, const bool is_static, const bool is_constructor)"
	     << endl
	     << "\tParams: *" << return_type.get_value() << "*" << func_name
	     << "*" << is_static << "*" << is_constructor << "*" << endl;
#endif

    // fixme
    FuncMember fm;
    fm.name = func_name;
    fm.mod = curr_mod;
    fm.is_static = is_static;
    // fixme: not always true
    fm.class_id = prog.classes.size()-1;


    // fixme void type...
    switch(return_type.get_type()) {
    case TK_BOOL: fm.ret_tind = Type::TC_sint1; break;
    case TK_BYTE: fm.ret_tind = Type::TC_sint1; break;
    case TK_UBYTE: fm.ret_tind = Type::TC_uint1; break;
    case TK_SHORT: fm.ret_tind = Type::TC_sint2; break;
    case TK_USHORT: fm.ret_tind = Type::TC_uint2; break;
    case TK_INT: fm.ret_tind = Type::TC_sint4; break;
    case TK_UINT: fm.ret_tind = Type::TC_uint4; break;
    case TK_LONG: fm.ret_tind = Type::TC_sint8; break;
    case TK_ULONG: fm.ret_tind = Type::TC_uint8; break;
    case TK_FLOAT: fm.ret_tind = Type::TC_real4; break;
    case TK_DOUBLE: fm.ret_tind = Type::TC_real8; break;
    case TK_STRING: fm.ret_tind = Type::TC_string; break;
    case TK_ID:
      if (prog.class_ids.find(return_type.get_value()) == prog.class_ids.end())
	PARSE_ERR("Unknown type");
      fm.ret_tind = prog.classes[prog.class_ids[return_type.get_value()]].ref_tind;
      break;
    default: // fixme
      ERR("fixme");
    }

    if (get_token_info().get_type() != TK_LPAR)
      PARSE_ERR("'(' expected after function name");


    // fixme this is wrong....
    // no check right now for 'this' pointer so it works
    // Push this pointer for non-static functions
    if (!is_static) {
      TypeIndex t;
      t = 0; // fixme need current function type
      fm.par_types.push_back(t);
    }

    string stored_name = fm.name;

    // Get parameters
    param_vars.clear();
    FOREVER {
      Token tk = get_token_info();
      if (tk.get_type() == TK_RPAR) break;

      TypeIndex t;
      switch (tk.get_type()) {
      case TK_BOOL: t = Type::TC_sint1; break;
      case TK_BYTE: t = Type::TC_sint1; break;
      case TK_UBYTE: t = Type::TC_uint1; break;
      case TK_SHORT: t = Type::TC_sint2; break;
      case TK_USHORT: t = Type::TC_uint2; break;
      case TK_INT: t = Type::TC_sint4; break;
      case TK_UINT: t = Type::TC_uint4; break;
      case TK_LONG: t = Type::TC_sint8; break;
      case TK_ULONG: t = Type::TC_uint8; break;
      case TK_FLOAT: t = Type::TC_real4; break;
      case TK_DOUBLE: t = Type::TC_real8; break;
      case TK_STRING: t = Type::TC_string; break;
      default: // fixme
	;
      }
      cout << "WHY " << t << endl;
      fm.par_types.push_back(t);
      
      stringstream s;
      s << "*" << t;
      stored_name += s.str();

      // fixme: For now assume var names have to be here
      tk = get_token_info();
      if (tk.get_type() != TK_ID)
	PARSE_ERR("Expecting variable name for parameters");
      

      param_vars.push_back(var_info(tk.get_value(), fm.par_types.size()-1));

      if (peek_token_info().get_type() == TK_COMMA) get_token_info();
      else if (peek_token_info().get_type() != TK_RPAR)
	PARSE_ERR2("Unexpected token in param list", peek_token_info().get_value());
    }

    // fixme need to check for conflicts.... and later on to see if fns defined

    // Set function into current class
    /*
    if (is_static) {
      cout << "Setting in func " << stored_name << " in " << prog.classes.size() << " " << curr_class->static_funcs.size() << endl;
      curr_class->static_func_ids[stored_name] = 
	curr_class->static_funcs.size();
      curr_class->static_funcs.push_back(fm);
      
    }
    else {
      cout << "Setting in func " << stored_name << " in " << prog.classes.size() << " " << curr_class->virt_funcs.size() << endl;
      curr_class->virt_func_ids[stored_name] = 
	curr_class->virt_funcs.size();
      curr_class->virt_funcs.push_back(fm);
    }
    */
    cout << "WHAT IS MY SIZE " << curr_class->virt_funcs.size() << endl;

    Token tk = peek_token_info();
    if (tk.get_type() == TK_SEMICOLON) {
      get_token_info(); // the semicolon
      return;
    }
    else if (tk.get_type() == TK_LBRACE) {
      cout << "HI " << curr_class << endl;
      get_token_info(); // the '{'
      //      uint4 loc;
      if (is_static) {
	//	loc = curr_class->static_funcs.size() - 1;
	cout << "WHY " << curr_class->static_funcs.size() << " " << curr_class->name << endl;
	curr_func = &curr_class->static_funcs[curr_class->static_func_ids[stored_name]];
      }
      else {
	//	loc = curr_class->virt_funcs.size() - 1;
	cout << "WHYs " << curr_class->virt_funcs.size() << endl;
	if (curr_class->virt_func_ids.find(stored_name) ==
	    curr_class->virt_func_ids.end()) {
	  ERR("WRONG");
	}
	else {
	  cout << curr_class->virt_func_ids[stored_name] << endl;
	}

	curr_func = &curr_class->virt_funcs[curr_class->virt_func_ids[stored_name]];
      }

      cout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
	   << curr_func->loc_types.size() << endl;
      cout << curr_func->par_types.size() << " ";
      //      cout << curr_func->par_types[0] << "BYE" << endl;
      parse_func_def();

      if (is_constructor && !has_return)
	curr_func->add_code(Opcodes::RETURN);

      curr_func = 0;
    }
  }


  void Parser::parse_class_dec() {
#if DEBUG
    DEBUGSTR << "parse_class_dec()" << endl;
#endif
    // Set default modifier to private

    Token tk = get_token_info();
    
    // Check for existance
    if (tk.get_type() != TK_ID)
      PARSE_ERR("expecting name for class");

    const string class_name = tk.get_value();

    // fixme need to be smarter here...ie def vs dec
    if (prog.class_ids.find(class_name) != prog.class_ids.end())
      curr_class = &prog.classes[prog.class_ids[class_name]];
    else
      ERR("DUMMY");
    //      curr_class = prog.add_class(class_name);

    tk = get_token_info();

    cout << "HOW COME " << curr_class->static_funcs.size() << " " 
	 << curr_class->virt_funcs.size() << endl;

    /* !new! cannot have declaration of class
    if (tk.get_type() == TK_SEMICOLON) {
      curr_class = 0;
      return;
    }
    */
    if (tk.get_type() == TK_LBRACE) parse_func_headers();
    else if (tk.get_type() == TK_COLON) {
      const string super_class = get_token_info().get_value();


      // !new!
      // First pass should have found super class
      curr_class->super_class_index = prog.class_ids[super_class];

      // fixme should check to see that super class exists

      /*
      // See if super class exists
      if (prog.class_ids.find(super_class) != 
	  prog.class_ids.end())
	curr_class->super_class_index = prog.class_ids[super_class];
      else 
	PARSE_ERR("Unknown superclass");
      */

      if (get_token_info().get_type() != TK_LBRACE) 
	PARSE_ERR("Expecting definition of class");

      // Need to copy over stuff

      // fixme now without forward dec... how do we inherit properly?
      cout << curr_class->name << endl;
      cout << curr_class->super_class_index << endl;
      cout << curr_class->inst_data.size() << endl;

      //      curr_class->inherit(prog.classes[curr_class->super_class_index]);

      // hack
      curr_class->inherit2(prog.classes[curr_class->super_class_index]);
      cout << "AFTER" << curr_class->inst_data.size() << endl;

      //      ERR("STOP");
      parse_func_headers();
    }
    else PARSE_ERR("Unexpected token");

    cout << "****************HERE " << curr_class << endl;
    curr_class->init_class();

    curr_class = 0;
  }



  void Parser::parse_func_headers() {
#if DEBUG
    DEBUGSTR << "parse_func_headers()" << endl;
#endif
    
    // Only interested in functions, skip everything else
    FOREVER {
      const Token type = get_token_info();
      switch(type.get_type()) {
      case TK_RBRACE:
	return;
	
      case TK_STATIC: {
	const Token type2 = get_token_info();
	if (type2.get_type() == TK_VOID) {
	  Token tk2 = get_token_info();
	  const string func_name = tk2.get_value();
	  parse_func_dec2(TK_VOID, func_name, true, false);
	}
	else {
	  Token tk = peek_token_info();
	  
	  switch(type2.get_type()) {
	  case TK_BOOL:
	  case TK_BYTE:
	  case TK_UBYTE:
	  case TK_SHORT:
	  case TK_USHORT:
	  case TK_INT:
	  case TK_UINT:
	  case TK_LONG:
	  case TK_ULONG:
	  case TK_FLOAT:
	  case TK_DOUBLE:
	  case TK_STRING:
	    if (tk.get_type() != TK_LBRACKET) {
	      Token tk2 = get_token_info();
	      string name = tk.get_value();
	      if (peek_token_info().get_type() == TK_LPAR) 
		parse_func_dec2(type2, name, true, false);
	    }
	    break;

	  case TK_ID:
	    if (prog.class_ids.find(type2.get_value()) !=
		prog.class_ids.end()) {
	      if (tk.get_type() != TK_LBRACKET)
		if (type2.get_value() == curr_class->name &&
		    peek_token_info().get_type() == TK_LPAR) {
		  parse_func_dec2(type2, type2.get_value(), true, true);
		}
		else {
		  Token tk2 = get_token_info();
		  string name = tk.get_value();
		  if (peek_token_info().get_type() == TK_LPAR) 
		    parse_func_dec2(type2, name, true, false);
		  
		}
	    }
	    else PARSE_ERR("Unexpected token");
	    break;

	  default:
	    PROGRAM_ERR;
	  }
	}
	break;
      }

      case TK_VOID: {
	Token tk2 = get_token_info();
	const string func_name = tk2.get_value();
	parse_func_dec2(TK_VOID, func_name, false, false);
	break;
      }

      case TK_BOOL:
      case TK_BYTE:
      case TK_UBYTE:
      case TK_SHORT:
      case TK_USHORT:
      case TK_INT:
      case TK_UINT:
      case TK_LONG:
      case TK_ULONG:
      case TK_FLOAT:
      case TK_DOUBLE:
      case TK_STRING: {
	const Token tk = peek_token_info();
	if (tk.get_type() != TK_LBRACKET) {
	  Token tk2 = get_token_info();
	  string name = tk.get_value();
	  if (peek_token_info().get_type() == TK_LPAR) 
	    parse_func_dec2(type, name, false, false);
	  else {
	    while (peek_token_info().get_type() == TK_COMMA) {
	      get_token_info(); // the ','
	      get_token_info();
	    }
	  }
	    
	}
	break;
      }
	
      case TK_ID:
	if (prog.class_ids.find(type.get_value()) !=
	    prog.class_ids.end()) {
	  Token tk = peek_token_info();
	  if (tk.get_type() != TK_LBRACKET) 
	    if (type.get_value() == curr_class->name &&
		peek_token_info().get_type() == TK_LPAR) {
	      parse_func_dec2(type, type.get_value(), false, true);
	    }
	    else {
	      Token tk2 = get_token_info();
	      string name = tk.get_value();
	      if (peek_token_info().get_type() == TK_LPAR) 
		parse_func_dec2(type, name, false, false);
	      
	    }
	}
	else PARSE_ERR("Unexpected token");
	break;
	
      default: // none
	;
      }
      
    }
  }
  

  //////////////////////////////////////////////////////////////////////


  void Parser::parse_stmt_list(const TokenType &expected_end_marker) {
#if DEBUG
    DEBUGSTR << "parse_stmt_list(const TokenType &expected_end_marker)" << endl
	     << "\tParams: *" << expected_end_marker << "*" << endl;
#endif
    FOREVER {
      // order of ifs important!
      const TokenType type = peek_token_info().get_type();
      if (type == expected_end_marker) {
	get_token_info(); // the '}'
	if (type == TK_RBRACE) del_block();
	return;
      }
      if (type == TK_EOF) 
	PARSE_ERR("Unexpected end of file");

      parse_stmt();
    }
  }

  void Parser::parse_stmt() {
#if DEBUG
    DEBUGSTR << "parse_stmt()" << endl;
#endif
    const Token tk = get_token_info();
    
    switch(tk.get_type()) {

    case TK_BOOL:
    case TK_BYTE:
    case TK_UBYTE:
    case TK_SHORT:
    case TK_USHORT:
    case TK_INT:
    case TK_UINT:
    case TK_LONG:
    case TK_ULONG:
    case TK_FLOAT:
    case TK_DOUBLE:
    case TK_STRING:
      if (peek_token_info().get_type() == TK_LBRACKET)
      	parse_local_array(tk.get_type());
      else
	parse_local_var(tk.get_type()); 
      return;

    case TK_IF: parse_if(); return;

    case TK_WHILE: parse_while(); return;

    case TK_FOR: parse_for(); return;

    case TK_SWITCH: parse_switch(); return;

    case TK_ID: // fixme need to worry about class inst vars
      // fixme for local class declaration, function call
      if (prog.class_ids.find(tk.get_value()) != prog.class_ids.end())
	parse_local_obj(tk.get_value());
      // fixme hack...
      else if (prog.extf_ids.find(tk.get_value()) !=
	       prog.extf_ids.end())
	parse_extf_call(prog.extf_ids[tk.get_value()], true);
      else if (peek_token_info().get_type() == TK_LPAR)
	parse_func_call(tk.get_value(), true, -1);
      else parse_assign(tk.get_value());
      return;

    case TK_LBRACE:
      add_block();
      parse_stmt_list(TK_RBRACE);
      return;

    case TK_SEMICOLON:
      return;

    case TK_BREAK:
      parse_break();

    case TK_CONT:
      parse_cont();

    case TK_RETURN:
      parse_return();
      return;

    default: //fixme
      PARSE_ERR("Unexpected token");
    }
  }

  void Parser::parse_local_array(const TokenType &type) {
#if DEBUG
    DEBUGSTR << "parse_local_array(const TokenType &type)" << endl
	     << "\tParams: *" << type << "*" << endl;
#endif
    get_token_info(); // the '['
    Token tk = get_token_info();

    bool sized = false;
    sint4 size;

    // either a uint or ']'
    if (tk.get_type() == TK_INTC || tk.get_type() == TK_UINTC) {
      sized = true;
      size = atoi(tk.get_value().c_str());
      
      if (size < 0) PARSE_ERR("Array size needs to be bigger than 0");
      tk = get_token_info();
    }

    if (tk.get_type() != TK_RBRACKET)
      PARSE_ERR("Expecting ']'");
    
    FOREVER {
      Token name = get_token_info();
      if (name.get_type() != TK_ID)
	PARSE_ERR("Expecting variable name");

      // fixme name conflict check

      const string var_name = name.get_value();

      sint4 check_to = local_scopes.top();
      sint4 check_from = local_vars.size()-1;

      for (sint4 i = check_from; i >= check_to; --i)
	if (local_vars[i].first == var_name)
	  PARSE_ERR("Variable name used already in scope");

      TypeIndex t;
      switch(type) {
      case TK_BOOL:   t = Type::TC_sint1;  break;
      case TK_BYTE:   t = Type::TC_sint1;  break;
      case TK_UBYTE:  t = Type::TC_uint1;  break;
      case TK_SHORT:  t = Type::TC_sint2;  break;
      case TK_USHORT: t = Type::TC_uint2;  break;
      case TK_INT:    t = Type::TC_sint4;  break;
      case TK_UINT:   t = Type::TC_uint4;  break;
      case TK_LONG:   t = Type::TC_sint8;  break;
      case TK_ULONG:  t = Type::TC_uint8;  break;
      case TK_FLOAT:  t = Type::TC_real4;  break;
      case TK_DOUBLE: t = Type::TC_real8;  break;
      case TK_STRING: t = Type::TC_string; break;
      default:
	PROGRAM_ERR;
      }

      // fixme for now only single dimension
      prog.add_array(t, 1);
      
      curr_func->loc_types.push_back(prog.types.size()-1);
      local_vars.push_back(var_info(name.get_value(), curr_func->loc_types.size()-1));

      // fixme put in better spot
      curr_func->add_code(Opcodes::NEW_ARRAY);
      curr_func->add_uint4(prog.types.size()-1);
      curr_func->add_uint4((uint4)size);

      curr_func->add_code(Opcodes::STORE_LREF);
      curr_func->add_uint4(curr_func->loc_types.size()-1);

      tk = peek_token_info();
      if (tk.get_type() == TK_COMMA) get_token_info(); // the ','
      else {
	if (tk.get_type() == TK_SEMICOLON)
	  get_token_info(); // the ';'
	return;
      }
    }
  }

  void Parser::parse_local_var(const TokenType &type) {
#if DEBUG
    DEBUGSTR << "parse_local_var(const TokenType &type)" << endl
	     << "\tParams: *" << type << "*" << endl;
#endif
    FOREVER {
      Token tk = get_token_info();
      if (tk.get_type() != TK_ID)
	PARSE_ERR("Expecting variable name");

      // fixme
      // Need to check if it is a class or function name

      const string var_name = tk.get_value();

      // See if it already exists

      sint4 check_to = local_scopes.top();
      sint4 check_from = local_vars.size()-1;
    
      for (sint4 i = check_from; i >= check_to; --i)
	if (local_vars[i].first == var_name)
	  PARSE_ERR("Variable name used already in this scope");
   
      // Set in function
      TypeIndex t;
      switch(type) {
      case TK_BOOL:   t = Type::TC_sint1;  break;
      case TK_BYTE:   t = Type::TC_sint1;  break;
      case TK_UBYTE:  t = Type::TC_uint1;  break;
      case TK_SHORT:  t = Type::TC_sint2;  break;
      case TK_USHORT: t = Type::TC_uint2;  break;
      case TK_INT:    t = Type::TC_sint4;  break;
      case TK_UINT:   t = Type::TC_uint4;  break;
      case TK_LONG:   t = Type::TC_sint8;  break;
      case TK_ULONG:  t = Type::TC_uint8;  break;
      case TK_FLOAT:  t = Type::TC_real4;  break;
      case TK_DOUBLE: t = Type::TC_real8;  break;
      case TK_STRING: t = Type::TC_string; break;
      default:
	PROGRAM_ERR;
      }
      
      curr_func->loc_types.push_back(t);
      local_vars.push_back(var_info(var_name, curr_func->loc_types.size()-1));

      cout << "%%%%%%%%%%%%%%WTh " << curr_func->loc_types.size() 
	   << local_vars.size() << endl;

      
      if (peek_token_info().get_type() == TK_ASSIGN) {
	get_token_info(); // the '='
	parse_exp(TK_SEMICOLON, t);
	switch (t) {
	case Type::TC_sint1: curr_func->add_code(Opcodes::STORE_LS1); break;
	case Type::TC_uint1: curr_func->add_code(Opcodes::STORE_LU1); break;
	case Type::TC_sint2: curr_func->add_code(Opcodes::STORE_LS2); break;
	case Type::TC_uint2: curr_func->add_code(Opcodes::STORE_LU2); break;
	case Type::TC_sint4: curr_func->add_code(Opcodes::STORE_LS4); break;
	case Type::TC_uint4: curr_func->add_code(Opcodes::STORE_LU4); break;
	case Type::TC_sint8: curr_func->add_code(Opcodes::STORE_LS8); break;
	case Type::TC_uint8: curr_func->add_code(Opcodes::STORE_LU8); break;
	case Type::TC_real4: curr_func->add_code(Opcodes::STORE_LR4); break;
	case Type::TC_real8: curr_func->add_code(Opcodes::STORE_LR8); break;
	case Type::TC_string: curr_func->add_code(Opcodes::STORE_LREF); break;
	default: 
	  PROGRAM_ERR;
	}
	curr_func->add_uint4(curr_func->loc_types.size()-1);
      }

      tk = peek_token_info();
      if (tk.get_type() == TK_COMMA) get_token_info(); // the ','
      else {
	if (tk.get_type() == TK_SEMICOLON) 
	  get_token_info(); // the ';'
	return;
      }
    }
  }

  void Parser::parse_local_obj(const string &class_name) {
#if DEBUG
    DEBUGSTR << "parse_local_obj(const string &class_name)" << endl
	     << "\tParams: *" << class_name << "*" << endl;
#endif
    
    uint4 class_index = prog.class_ids[class_name];
    TypeIndex t;

    for(uint4 i=Type::TC_object; i<prog.types.size(); ++i)
      if (prog.types[i].class_index == class_index) {
	t = i;
	break;
      }
    // fixme if not in here should be error but how can be error...must be c++ error not script error

    FOREVER {
      Token tk = get_token_info();
      if (tk.get_type() != TK_ID)
	PARSE_ERR("Expecting variable name");

      // fixme
      // Need to check if it is a class or function name

      const string var_name = tk.get_value();
      
      // See if it already exists
      sint4 check_to = local_scopes.top();
      sint4 check_from = local_vars.size()-1;
    
      cout << "WHY " << check_to << " " << check_from << endl;

      for (sint4 i = check_from; i > check_to; --i)
	if (local_vars[i].first == var_name)
	  PARSE_ERR("Variable name used already in this scope");
      
      curr_func->loc_types.push_back(t);
      local_vars.push_back(var_info(var_name, curr_func->loc_types.size()-1));


#if 0
      if (peek_token_info().get_type() == TK_ASSIGN) {
	get_token_info(); // the '='
	// fixme ...in future can be assignment
	if (get_token_info().get_type() != TK_NEW) 
	  PARSE_ERR("Expecting 'new'");

      }
#endif 

      tk = peek_token_info();
      if (tk.get_type() == TK_COMMA) get_token_info(); // the ','
      else {
	if (tk.get_type() == TK_SEMICOLON) get_token_info(); // the ';'
	return;
      }
    }
  }

  TypeIndex Parser::parse_func_call(const string& func_name, 
				    const bool is_stmt,
				    const sint4 class_index) {
#if DEBUG
    DEBUGSTR << "parse_func_call(const string& func_name, const bool is_stmt, "
	     << "const sint4 class_index)" << endl
	     << "\tParams: *" << func_name << "*" << is_stmt << "*"
	     << class_index << "*" << endl;
#endif
    // need to check for is_static now!!!


    Token tk = get_token_info();
    if (tk.get_type() != TK_LPAR)
      PARSE_ERR("Expecting '(' for function call");

    /*
    vector<TypeIndex> par_types;
    if (is_static)
      par_types = curr_class->static_funcs[func_index].par_types;
    else
      par_types = curr_class->virt_funcs[func_index].par_types;
    */

    //    uint4 end = par_types.size();

    cout << "HOW COME " << endl;

    // Save space in case virtual function

    //uint4 this_param = curr_func->get_addr();
    curr_func->add_code(Opcodes::NOP);
    curr_func->add_code(Opcodes::NOP);

    string stored_name = func_name;

    FOREVER {
      // Push this pointer
      /*
      if (!is_static and i==0) {
	curr_func->add_code(Opcodes::LOAD_PARAM);
	curr_func->add_code(0);
	continue;
      }
      */      

      if (peek_token_info().get_type() == TK_RPAR) {
	get_token_info(); // the '('
	break;
      }
      


      TypeIndex ti=Type::TC_sint4; //fixme hack
      parse_exp(TK_NIL, ti);

      cout << "HI " << ti << endl;

      stringstream s;
      s << "*" << ti;
      stored_name += s.str();
 

      tk = get_token_info();

      if (tk.get_type() == TK_RPAR) break;

      if (tk.get_type() != TK_COMMA)
	PARSE_ERR("Expecting ',' or ')' for function call");
    }

    bool is_static;
    uint4 func_index;
    TypeIndex ret_tind;
    cout << "WHY " << stored_name << endl;

    if (class_index != -1) {
      cout << "HELO" << endl;
      cout << prog.classes[class_index].virt_func_ids.begin()->first << endl;
      if (prog.classes[class_index].virt_func_ids.find(stored_name) !=
	  prog.classes[class_index].virt_func_ids.end()) {
	is_static = false;
	func_index = prog.classes[class_index].virt_func_ids[stored_name];
	ret_tind = prog.classes[class_index].virt_funcs[func_index].ret_tind;
      }
      else if (prog.classes[class_index].static_func_ids.find(stored_name) !=
	       prog.classes[class_index].static_func_ids.end()) {
	is_static = true;
	func_index = prog.classes[class_index].static_func_ids[stored_name];
	ret_tind = prog.classes[class_index].static_funcs[func_index].ret_tind;
      }
      else
	PARSE_ERR2("Unknown function", func_name);
    }
    else if (curr_class->virt_func_ids.find(stored_name) !=
	     curr_class->virt_func_ids.end()) {
      is_static = false;
      func_index = curr_class->virt_func_ids[stored_name];
      ret_tind = curr_class->virt_funcs[func_index].ret_tind;
    }
    else if (curr_class->static_func_ids.find(stored_name) !=
	     curr_class->static_func_ids.end()) {
      is_static = true;
      func_index = curr_class->static_func_ids[stored_name];
      ret_tind = curr_class->static_funcs[func_index].ret_tind;
    }
    else
      PARSE_ERR2("Unknown function", func_name);

    // Check to see if we are calling virtual function from inside
    // a static function
    if (!is_static && curr_static)
      PARSE_ERR("Cannot call virtual function inside static function");    


    if (is_static) {
      curr_func->add_code(Opcodes::CALL_STAT);
      curr_func->add_uint4(class_index);
      curr_func->add_uint4(func_index);
    }
    else {

      //      curr_func->set_code(this_param, Opcodes::LOAD_PARAM);
      // curr_func->set_uint4(this_param + 1, 0);

      if (class_index == -1) {
	curr_func->add_code(Opcodes::LOAD_PARAM);
	curr_func->add_uint4(0);
      }
      else {
	// super big hack
	curr_func->add_code(Opcodes::LOAD_LOCAL);
	curr_func->add_uint4(1);
      }

      curr_func->add_code(Opcodes::CALL_VIRT);
      curr_func->add_uint4(func_index);


    }

    if (is_stmt)
      curr_func->add_code(Opcodes::POP);

    return ret_tind;
  }

  void Parser::parse_extf_call(const uint4 func_index, const bool is_stmt) {
#if DEBUG
    DEBUGSTR << "parse_extf_call(const uint4 func_index, const bool is_stmt)"
	     << endl << "\tParams: *" << func_index << "*" << is_stmt
	     << "*" << endl;
#endif

    Token tk = get_token_info();
    if (tk.get_type() != TK_LPAR)
      PARSE_ERR("Expecting '(' for external function call");

    vector<TypeIndex> par_types = prog.extfs[func_index].par_types;
    uint4 end = par_types.size();

    FORU (i, end) {
      parse_exp(TK_NIL, par_types[i]);

      if (i != end-1) {
	if (get_token_info().get_type() != TK_COMMA)
	  PARSE_ERR("Still expecting parameters for external function call");
      }
      else if (get_token_info().get_type() != TK_RPAR)
	PARSE_ERR("Expecting ')' for external function call");
    }

    curr_func->add_code(Opcodes::CALL_EXT);
    curr_func->add_uint4(func_index);

    if (is_stmt)
      curr_func->add_code(Opcodes::POP);

  }

  void Parser::parse_assign(const string &var_name) {
#if DEBUG
    DEBUGSTR << "parse_assign(const string &var_name)" << endl
	     << "\tParams: *" << var_name << "*" << endl;
#endif
    // fixme class member var, arrays
    var_location loc = find_variable(var_name);
    if (loc.first == VAR_UNKNOWN)
      PARSE_ERR2("Undeclared variable", var_name);
    
    Token tk = get_token_info();

    TypeIndex t;
    if (loc.first == VAR_LOCAL) t = curr_func->loc_types[loc.second];
    else if (loc.first == VAR_PARAM) t = curr_func->par_types[loc.second];
    else if (loc.first == VAR_INST_MEMBER) {
      t = curr_class->inst_data[loc.second].type;
    }
    else // fixme class member 
      ;

    cout << "t is " << t << " " <<Type::TC_object << endl;
    if (t >= Type::TC_object) {

      if (prog.types[t].cat == Type::TC_array) {
	if (tk.get_type() != TK_LBRACKET)
	  PARSE_ERR("Expecting '['");

	Token ind = get_token_info();

	// fixme can be variable
	if (ind.get_type() != TK_INTC && ind.get_type() != TK_UINTC)
	  PARSE_ERR("Expecting index");

	sint4 index = atoi(ind.get_value().c_str());

	if (get_token_info().get_type() != TK_RBRACKET)
	  PARSE_ERR("Expecting ']'");

	if (get_token_info().get_type() != TK_ASSIGN)
	  PARSE_ERR("Expecting '='");

	TypeIndex t2 = prog.types[t].T_cat;


	if (loc.first == VAR_LOCAL)
	  curr_func->add_code(Opcodes::LOAD_LOCAL);
	else if (loc.first == VAR_PARAM)
	  curr_func->add_code(Opcodes::LOAD_PARAM);
	else
	  ERR("FIXME");
	curr_func->add_uint4(loc.second);

	curr_func->add_code(Opcodes::PUSH_CU4);
	curr_func->add_uint4((uint4)index);

	parse_exp(TK_SEMICOLON, t2);

	switch(t2) {
	case Type::TC_sint1: curr_func->add_code(Opcodes::STORE_AS1); break;
	case Type::TC_uint1: curr_func->add_code(Opcodes::STORE_AU1); break;
	case Type::TC_sint2: curr_func->add_code(Opcodes::STORE_AS2); break;
	case Type::TC_uint2: curr_func->add_code(Opcodes::STORE_AU2); break;
	case Type::TC_sint4: curr_func->add_code(Opcodes::STORE_AS4); break;
	case Type::TC_uint4: curr_func->add_code(Opcodes::STORE_AU4); break;
	case Type::TC_sint8: curr_func->add_code(Opcodes::STORE_AS8); break;
	case Type::TC_uint8: curr_func->add_code(Opcodes::STORE_AU8); break;
	case Type::TC_real4: curr_func->add_code(Opcodes::STORE_AR4); break;
	case Type::TC_real8: curr_func->add_code(Opcodes::STORE_AR8); break;
	case Type::TC_string: curr_func->add_code(Opcodes::STORE_AREF); break;
	default: // fixme
	  ERR("FOO4");	
	  ;
	}

	return;
      }

      if (tk.get_type() != TK_ASSIGN && tk.get_type() != TK_DOT)
	PARSE_ERR("Expecting '='");

      if (tk.get_type() == TK_ASSIGN) {
	if (get_token_info().get_type() != TK_NEW)
	  PARSE_ERR("Expecting 'new'");
	tk = get_token_info();
	uint4 class_index = prog.types[t].class_index;
	if (tk.get_value() != prog.classes[class_index].name)
	  PARSE_ERR("Wrong class");
	
	
	curr_func->add_code(Opcodes::NEW_CLASS);
	curr_func->add_uint4(class_index);
	
	if (loc.first == VAR_LOCAL)
	  curr_func->add_code(Opcodes::STORE_LREF);
	else
	  ERR("IMPLEMENT ME");
	
	curr_func->add_uint4(loc.second);

	// Call the constructor

	parse_func_call(prog.classes[class_index].name, false, class_index);
	// fixme right now no args

	/*
	if (get_token_info().get_type() != TK_LPAR)
	  PARSE_ERR("Expecting '('");
	
	if (get_token_info().get_type() != TK_RPAR)
	  PARSE_ERR("Expecting ')'");
	*/

	return;
      }
      else if (tk.get_type() == TK_DOT) {
	TypeIndex t2;
	Token member;
	uint4 class_index;
	FOREVER {
	  member = get_token_info();
	  if (member.get_type() != TK_ID)
	    PARSE_ERR("Expecting member after '.'");
	  
	  class_index = prog.types[t].class_index;

	  cout << "CI " << class_index << endl;
	  cout << prog.classes[class_index].inst_data.size() << endl;
	  
	  if (prog.classes[class_index].inst_data_ids.
	      find(member.get_value()) ==
	      prog.classes[class_index].inst_data_ids.end())
	    PARSE_ERR2("Unknown member: ", member.get_value());
	  
	  t2 = prog.classes[class_index].inst_data[prog.classes[class_index].inst_data_ids[member.get_value()]].type;
	  

	  tk = get_token_info();
	  if (tk.get_type() == TK_ASSIGN) break;
	  
	  // fixme have to check to make sure t2 is an object
	  if (tk.get_type() != TK_DOT)
	    PARSE_ERR2("Unexpected token", tk.get_value());
	}

	
	parse_exp(TK_SEMICOLON, t2);
	
	if (loc.first == VAR_LOCAL)
	  curr_func->add_code(Opcodes::LOAD_LOCAL);
	else if (loc.first == VAR_PARAM)
	  curr_func->add_code(Opcodes::LOAD_PARAM);
	else
	  ERR("FIXME");
	curr_func->add_uint4(loc.second);

			     
	switch(t2) {
	case Type::TC_sint1: curr_func->add_code(Opcodes::STORE_IS1); break;
	case Type::TC_uint1: curr_func->add_code(Opcodes::STORE_IU1); break;
	case Type::TC_sint2: curr_func->add_code(Opcodes::STORE_IS2); break;
	case Type::TC_uint2: curr_func->add_code(Opcodes::STORE_IU2); break;
	case Type::TC_sint4: curr_func->add_code(Opcodes::STORE_IS4); break;
	case Type::TC_uint4: curr_func->add_code(Opcodes::STORE_IU4); break;
	case Type::TC_sint8: curr_func->add_code(Opcodes::STORE_IS8); break;
	case Type::TC_uint8: curr_func->add_code(Opcodes::STORE_IU8); break;
	case Type::TC_real4: curr_func->add_code(Opcodes::STORE_IR4); break;
	case Type::TC_real8: curr_func->add_code(Opcodes::STORE_IR8); break;
	case Type::TC_string: curr_func->add_code(Opcodes::STORE_IREF); break;
	default: // fixme
	  ERR("FOO4");	
	  ;
	}
	curr_func->add_uint4(prog.classes[class_index].inst_data_ids[member.get_value()]);

	return;
      }
    }

    // fixme if this is new then it won't work
    if (tk.get_type() != TK_ASSIGN) PARSE_ERR("Expecting '='");
    parse_exp(TK_SEMICOLON, t);
    
    if (loc.first == VAR_LOCAL) {
      switch(t) {
      case Type::TC_sint1: curr_func->add_code(Opcodes::STORE_LS1); break;
      case Type::TC_uint1: curr_func->add_code(Opcodes::STORE_LU1); break;
      case Type::TC_sint2: curr_func->add_code(Opcodes::STORE_LS2); break;
      case Type::TC_uint2: curr_func->add_code(Opcodes::STORE_LU2); break;
      case Type::TC_sint4: curr_func->add_code(Opcodes::STORE_LS4); break;
      case Type::TC_uint4: curr_func->add_code(Opcodes::STORE_LU4); break;
      case Type::TC_sint8: curr_func->add_code(Opcodes::STORE_LS8); break;
      case Type::TC_uint8: curr_func->add_code(Opcodes::STORE_LU8); break;
      case Type::TC_real4: curr_func->add_code(Opcodes::STORE_LR4); break;
      case Type::TC_real8: curr_func->add_code(Opcodes::STORE_LR8); break;
      case Type::TC_string: curr_func->add_code(Opcodes::STORE_LREF); break;
      default: // fixme
	ERR("FOO3");
	;
      }
      curr_func->add_uint4(loc.second);
    }
    else if (loc.first == VAR_PARAM) {
      switch(t) {
      case Type::TC_sint1: curr_func->add_code(Opcodes::STORE_PS1); break;
      case Type::TC_uint1: curr_func->add_code(Opcodes::STORE_PU1); break;
      case Type::TC_sint2: curr_func->add_code(Opcodes::STORE_PS2); break;
      case Type::TC_uint2: curr_func->add_code(Opcodes::STORE_PU2); break;
      case Type::TC_sint4: curr_func->add_code(Opcodes::STORE_PS4); break;
      case Type::TC_uint4: curr_func->add_code(Opcodes::STORE_PU4); break;
      case Type::TC_sint8: curr_func->add_code(Opcodes::STORE_PS8); break;
      case Type::TC_uint8: curr_func->add_code(Opcodes::STORE_PU8); break;
      case Type::TC_real4: curr_func->add_code(Opcodes::STORE_PR4); break;
      case Type::TC_real8: curr_func->add_code(Opcodes::STORE_PR8); break;
      case Type::TC_string: curr_func->add_code(Opcodes::STORE_PREF); break;
      default: // fixme
	ERR("FOO4");	
	;
      }
      curr_func->add_uint4(loc.second);
    }
    else if (loc.first == VAR_INST_MEMBER) {

      curr_func->add_code(Opcodes::LOAD_PARAM);
      curr_func->add_uint4(0);
      
      switch(t) {
      case Type::TC_sint1: curr_func->add_code(Opcodes::STORE_IS1); break;
      case Type::TC_uint1: curr_func->add_code(Opcodes::STORE_IU1); break;
      case Type::TC_sint2: curr_func->add_code(Opcodes::STORE_IS2); break;
      case Type::TC_uint2: curr_func->add_code(Opcodes::STORE_IU2); break;
      case Type::TC_sint4: curr_func->add_code(Opcodes::STORE_IS4); break;
      case Type::TC_uint4: curr_func->add_code(Opcodes::STORE_IU4); break;
      case Type::TC_sint8: curr_func->add_code(Opcodes::STORE_IS8); break;
      case Type::TC_uint8: curr_func->add_code(Opcodes::STORE_IU8); break;
      case Type::TC_real4: curr_func->add_code(Opcodes::STORE_IR4); break;
      case Type::TC_real8: curr_func->add_code(Opcodes::STORE_IR8); break;
      case Type::TC_string: curr_func->add_code(Opcodes::STORE_IREF); break;
      default: // fixme
	ERR("FOO4");	
	;
      }
      curr_func->add_uint4(loc.second);
    }
    else // fixme class member
      ;
    
		
  }

  void Parser::parse_if() {
#if DEBUG
    DEBUGSTR << "parse_if()" << endl;
#endif
    // Get the expression
    if (get_token_info().get_type() != TK_LPAR)
      PARSE_ERR("'(' expected after 'if'");
    TypeIndex t = Type::TC_unknown;
    parse_exp(TK_RPAR, t);
    curr_func->add_code(Opcodes::JUMP_FALSE);

    // Holder for end of if body
    uint4 jump = curr_func->get_addr();
    curr_func->add_addr(0);

    // Get the statements
    parse_stmt();

    // Fix jump position for false if condition
    curr_func->set_addr(jump, curr_func->get_addr());

    // See if there is an else clause
    if (peek_token_info().get_type() == TK_ELSE)
      parse_stmt();
  }

  void Parser::parse_while() {
#if DEBUG
    DEBUGSTR << "parse_while()" << endl;
#endif
    // Holder for start of while
    uint4 cont = curr_func->get_addr();

    // Get the expression
    if (get_token_info().get_type() != TK_LPAR) PARSE_ERR("'(' expected after 'while'");

    TypeIndex t = Type::TC_sint4;
    parse_exp(TK_RPAR, t);
    curr_func->add_code(Opcodes::JUMP_FALSE);

    // Holder  for end of while body
    uint4 jump = curr_func->get_addr();
    curr_func->add_addr(0);
    
    // Get the statements
    parse_stmt();

    // Set jump to beginning of while
    curr_func->add_code(Opcodes::JUMP);
    curr_func->add_addr(cont);

    // Fix jump position for false while condition
    curr_func->set_addr(jump, curr_func->get_addr());
  }

  void Parser::parse_for() {
#if DEBUG
    DEBUGSTR << "parse_for()" << endl;
#endif
    add_block();
    
    if (get_token_info().get_type() != TK_LPAR)
      PARSE_ERR("'(' expected after 'for'");
    
    // fixme variable decs
    Token tk = get_token_info();
    switch(tk.get_type()) {
    case TK_BOOL:
    case TK_BYTE:
    case TK_UBYTE:
    case TK_SHORT:
    case TK_USHORT:
    case TK_INT:
    case TK_UINT:
    case TK_LONG:
    case TK_ULONG:
    case TK_FLOAT:
    case TK_DOUBLE:
    case TK_STRING:
      parse_local_var(tk.get_type());
      break;
    default: // fixme can this be a class
      ERR("NO");
    }

    // Stop condition
    TypeIndex t = Type::TC_unknown;
    uint4 cont = curr_func->get_addr();
    parse_exp(TK_SEMICOLON, t);

    curr_func->add_code(Opcodes::JUMP_FALSE);
    uint4 jump = curr_func->get_addr();
    curr_func->add_addr(0);

    // fixme get increments and save somehow
    ERR("Implement me!");
    

    if (get_token_info().get_type() != TK_RPAR)
      PARSE_ERR("')' expected at end of 'for' increments");

    // Get the statement(s) for the for loop
    parse_stmt();
    
    // fixme here we add the increments to code

    curr_func->add_code(Opcodes::JUMP);
    curr_func->add_addr(cont);

    curr_func->set_addr(jump, curr_func->get_addr());
  }

  void Parser::parse_switch() {
#if DEBUG
    DEBUGSTR << "parse_switch()" << endl;
#endif
    // fixme
    ERR("Implement me!");
  }

  void Parser::parse_break() {
#if DEBUG
    DEBUGSTR << "parse_break()" << endl;
#endif
    if (break_addrs.empty())
      PARSE_ERR("'break' outside of loop");

    uint4 break_to = break_addrs.top();
    curr_func->add_code(Opcodes::JUMP);
    curr_func->add_addr(break_to);
  }

  void Parser::parse_cont() {
#if DEBUG
    DEBUGSTR << "parse_cont()" << endl;
#endif
    if (cont_addrs.empty())
      PARSE_ERR("'continue ' outside of loop");
    
    uint4 cont_to = cont_addrs.top();
    curr_func->add_code(Opcodes::JUMP);
    curr_func->add_addr(cont_to);
  }

  void Parser::parse_return() {
#if DEBUG
    DEBUGSTR << "parse_return()" << endl;
#endif
    has_return = true;
    if (!curr_func)
      PARSE_ERR("Return statement outside function");

    if (curr_func->ret_tind == Type::TC_void) {
      curr_func->add_code(Opcodes::RETURN);
      return;
    }

    parse_exp(TK_NIL, curr_func->ret_tind);
    curr_func->add_code(Opcodes::RETURN_VAL);
  }

  void Parser::parse_exp(const TokenType &expected_end_marker,
			 const TypeIndex &expected_type) {
#if DEBUG
    DEBUGSTR << "parse_exp(const TokenType &expected_end_marker, "
	     << "const TypeIndex &expected_type)" << endl
	     << "\tParams: *" << expected_end_marker << "*"
	     << expected_type << "*" << endl;
#endif
    parse_till_low_op(0, expected_type);
    
    if (expected_end_marker == TK_NIL) return;
    if (expected_end_marker == TK_SEMICOLON) {
      if (peek_token_info().get_type() == TK_SEMICOLON) get_token_info(); // the ';'
      return;
    }

    if (get_token_info().get_type() != expected_end_marker)
      PARSE_ERR("Unexepected end of expression marker");
  }

  
  void Parser::parse_till_low_op(sint4 prec, const TypeIndex &expected_type) {
#if DEBUG
    DEBUGSTR << "parse_till_low_op(sint4 prec, const TypeIndex &expected_type)"
	     << endl << "\tParams: *" << prec << "*" << expected_type << "*"
	     << endl;
#endif
    // the left term
    TypeIndex found_type=expected_type; // fixme
    get_term(found_type);

    FOREVER {
      Token op = peek_token_info();
      cout << "WHY " << op.get_value() << " " << op.get_type() << " " << TK_ADD << endl;
      sint4 op_prec = get_precedence_of(op.get_type());
      cout << op_prec << " and " << prec << endl;
      if (op_prec <= prec) return;

      get_token_info(); // the operator
      
      // the right term
      parse_till_low_op(op_prec, expected_type);

      if (found_type >= Type::TC_END)
	PARSE_ERR("Doing binary operation on non simple type");
      
      if (found_type == Type::TC_sint1 || found_type == Type::TC_sint2 ||
	  found_type == Type::TC_sint4) {
	switch (op.get_type()) {
	case TK_ADD:    curr_func->add_code(Opcodes::ADD_SS4); break;
	case TK_SUB:    curr_func->add_code(Opcodes::SUB_SS4); break;
	case TK_MUL:    curr_func->add_code(Opcodes::MUL_SS4); break;
	case TK_DIV:    curr_func->add_code(Opcodes::DIV_SS4); break;
	case TK_MOD:    curr_func->add_code(Opcodes::MOD_SS4); break;
	case TK_LSHIFT: curr_func->add_code(Opcodes::SHL_I4);  break;
	case TK_RSHIFT: curr_func->add_code(Opcodes::SHR_I4);  break;
	case TK_EQ:     curr_func->add_code(Opcodes::EQ_SS4);  break;
	case TK_NEQ:    curr_func->add_code(Opcodes::NEQ_SS4); break;
	case TK_LT:     curr_func->add_code(Opcodes::LT_SS4);  break;
	case TK_LEQ:    curr_func->add_code(Opcodes::LEQ_SS4); break;
	case TK_GT:     curr_func->add_code(Opcodes::GT_SS4);  break;
	case TK_GEQ:    curr_func->add_code(Opcodes::GEQ_SS4); break;
	case TK_AND:    curr_func->add_code(Opcodes::AND_I4);  break;
	case TK_ANDAND: //fixme
	case TK_OR:     curr_func->add_code(Opcodes::OR_I4);   break;
	case TK_OROR: // fixme
	case TK_XOR:    curr_func->add_code(Opcodes::XOR_I4);  break;
	default:
	  PARSE_ERR("Unknown binary op");
	}
      }
      else if (found_type == Type::TC_uint1 || found_type == Type::TC_uint2 ||
	       found_type == Type::TC_uint4) {
	switch (op.get_type()) {
	case TK_ADD:    curr_func->add_code(Opcodes::ADD_UU4); break;
	case TK_SUB:    curr_func->add_code(Opcodes::SUB_UU4); break;
	case TK_MUL:    curr_func->add_code(Opcodes::MUL_UU4); break;
	case TK_DIV:    curr_func->add_code(Opcodes::DIV_UU4); break;
	case TK_MOD:    curr_func->add_code(Opcodes::MOD_UU4); break;
	case TK_LSHIFT: curr_func->add_code(Opcodes::SHL_I4);  break;
	case TK_RSHIFT: curr_func->add_code(Opcodes::SHR_I4);  break;
	case TK_EQ:     curr_func->add_code(Opcodes::EQ_UU4);  break;
	case TK_NEQ:    curr_func->add_code(Opcodes::NEQ_UU4); break;
	case TK_LT:     curr_func->add_code(Opcodes::LT_UU4);  break;
	case TK_LEQ:    curr_func->add_code(Opcodes::LEQ_UU4); break;
	case TK_GT:     curr_func->add_code(Opcodes::GT_UU4);  break;
	case TK_GEQ:    curr_func->add_code(Opcodes::GEQ_UU4); break;
	case TK_AND:    curr_func->add_code(Opcodes::AND_I4);  break;
	case TK_ANDAND: //fixme
	case TK_OR:     curr_func->add_code(Opcodes::OR_I4);   break;
	case TK_OROR: // fixme
	case TK_XOR:    curr_func->add_code(Opcodes::XOR_I4);  break;
	default:
	  PARSE_ERR("Unknown binary op");
	}
      }
      else if (found_type == Type::TC_sint8) {
	switch (op.get_type()) {
	case TK_ADD:    curr_func->add_code(Opcodes::ADD_SS8); break;
	case TK_SUB:    curr_func->add_code(Opcodes::SUB_SS8); break;
	case TK_MUL:    curr_func->add_code(Opcodes::MUL_SS8); break;
	case TK_DIV:    curr_func->add_code(Opcodes::DIV_SS8); break;
	case TK_MOD:    curr_func->add_code(Opcodes::MOD_SS8); break;
	case TK_LSHIFT: curr_func->add_code(Opcodes::SHL_I8);  break;
	case TK_RSHIFT: curr_func->add_code(Opcodes::SHR_I8);  break;
	case TK_EQ:     curr_func->add_code(Opcodes::EQ_SS8);  break;
	case TK_NEQ:    curr_func->add_code(Opcodes::NEQ_SS8); break;
	case TK_LT:     curr_func->add_code(Opcodes::LT_SS8);  break;
	case TK_LEQ:    curr_func->add_code(Opcodes::LEQ_SS8); break;
	case TK_GT:     curr_func->add_code(Opcodes::GT_SS8);  break;
	case TK_GEQ:    curr_func->add_code(Opcodes::GEQ_SS8); break;
	case TK_AND:    curr_func->add_code(Opcodes::AND_I8);  break;
	case TK_ANDAND: //fixme
	case TK_OR:     curr_func->add_code(Opcodes::OR_I8);   break;
	case TK_OROR: // fixme
	case TK_XOR:    curr_func->add_code(Opcodes::XOR_I8);  break;
	default:
	  PARSE_ERR("Unknown binary op");
	}
      }
      else if (found_type == Type::TC_uint8) {
	switch (op.get_type()) {
	case TK_ADD:    curr_func->add_code(Opcodes::ADD_UU8); break;
	case TK_SUB:    curr_func->add_code(Opcodes::SUB_UU8); break;
	case TK_MUL:    curr_func->add_code(Opcodes::MUL_UU8); break;
	case TK_DIV:    curr_func->add_code(Opcodes::DIV_UU8); break;
	case TK_MOD:    curr_func->add_code(Opcodes::MOD_UU8); break;
	case TK_LSHIFT: curr_func->add_code(Opcodes::SHL_I8);  break;
	case TK_RSHIFT: curr_func->add_code(Opcodes::SHR_I8);  break;
	case TK_EQ:     curr_func->add_code(Opcodes::EQ_UU8);  break;
	case TK_NEQ:    curr_func->add_code(Opcodes::NEQ_UU8); break;
	case TK_LT:     curr_func->add_code(Opcodes::LT_UU8);  break;
	case TK_LEQ:    curr_func->add_code(Opcodes::LEQ_UU8); break;
	case TK_GT:     curr_func->add_code(Opcodes::GT_UU8);  break;
	case TK_GEQ:    curr_func->add_code(Opcodes::GEQ_UU8); break;
	case TK_AND:    curr_func->add_code(Opcodes::AND_I8);  break;
	case TK_ANDAND: //fixme
	case TK_OR:     curr_func->add_code(Opcodes::OR_I8);   break;
	case TK_OROR: // fixme
	case TK_XOR:    curr_func->add_code(Opcodes::XOR_I8);  break;
	default:
	  PARSE_ERR("Unknown binary op");
	}
      }
      else ERR("HUH");
    }
  }

  // Default safe conversions
  // - TC_sint1 -> any TC_sint > 1, any real
  // - TC_uint1 -> any TC_sint > 1, any TC_uint > 1, any real
  // - TC_sint2 -> any TC_sint > 2, any real
  // - TC_uint2 -> any TC_sint > 2, and TC_uint > 2, any real
  // - TC_sint4 -> TC_sint8, any real
  // - TC_uint4 -> TC_uint8, TC_sint8, any real
  // - TC_sint8 -> any real
  // - TC_uint8 -> any real
  // - TC_real4 -> TC_real8

  void Parser::get_term(TypeIndex &type) {
#if DEBUG
    DEBUGSTR << "get_term(TypeIndex &type)" << endl
	     << "\tParams: *" << type << "*" << endl;
#endif
    // fixme

    Token tk = get_token_info();
    // fixme    Type cast;
    stack<sint4> unary;
    
    // fixme bool has_cast = false;
    
    // first check for casting
    // fixme

    // unary operators

    // fixme can be of different types....
    FOREVER {
      // Possibilities: S4, U4, R4, R8
      if (tk.get_type() == TK_SUB) {
	unary.push(Opcodes::NEG_S4);
	tk = get_token_info();
      }
      // Possibilities: S4, U4
      else if (tk.get_type() == TK_COMPLEMENT) {
	unary.push(Opcodes::COMPL_S4);
	tk = get_token_info();
      }
      // Possibilities: S4, U4
      else if (tk.get_type() == TK_INVERSE) {
	unary.push(Opcodes::INV_S4);
	tk = get_token_info();
      }
      else break;
    }
    
    // fixme function call, constant, variable, exp, class var
    var_location loc;
    char* cptr;

    switch(type) {
    case Type::TC_unknown:
      // means i can get anything here, but it should be sint4 or 8?
      switch(tk.get_type()) {
      case TK_INTC:
	if (!valid_sint4(tk.get_value()))
	  PARSE_ERR("Invalid integer");
	
	type = Type::TC_sint4;
	curr_func->add_code(Opcodes::PUSH_CS4);
	curr_func->add_sint4(atoi(tk.get_value().c_str()));
	break;
      case TK_UINTC:
	if (!valid_uint4(tk.get_value()))
	  PARSE_ERR("Invalid integer");

	type = Type::TC_uint4;
	curr_func->add_code(Opcodes::PUSH_CU4);
	// fixme atoi incorrect
	curr_func->add_uint4(atoi(tk.get_value().c_str()));
	break;
      case TK_LONGC:
	if (!valid_sint8(tk.get_value()))
	  PARSE_ERR("Invalid long");

	type = Type::TC_sint8;
	curr_func->add_code(Opcodes::PUSH_CS8);
	curr_func->add_sint8(atoll(tk.get_value().c_str()));
	break;
      case TK_ULONGC:
	if (!valid_uint8(tk.get_value()))
	  PARSE_ERR("Invalid integer");

	type = Type::TC_uint8;
	curr_func->add_code(Opcodes::PUSH_CU8);
	// fixme atoi incorrect
	curr_func->add_uint4(atoll(tk.get_value().c_str()));
	break;
      case TK_STRINGC:
	type = Type::TC_string;
	curr_func->add_code(Opcodes::PUSH_CARR);
	curr_func->add_uint4(tk.get_value().length());
	FORU(i, tk.get_value().length())
	  curr_func->add_uint4(tk.get_value()[i]);
	break;

      case TK_ID:
	// static class variable or function
	if (prog.class_ids.find(tk.get_value()) !=
	    prog.class_ids.end()) {
	  ; // fixme
	}
	// fixme hack
	else if (prog.extf_ids.find(tk.get_value()) !=
		 prog.extf_ids.end()) {
	  type = prog.extfs[prog.extf_ids[tk.get_value()]].ret_tind;
	  parse_extf_call(prog.extf_ids[tk.get_value()], false);
	}
	else if (peek_token_info().get_type() == TK_LPAR) {
	  type = parse_func_call(tk.get_value(), false, -1);
	}
	else {
	  loc = find_variable(tk.get_value());
	  if (loc.first == VAR_LOCAL) {
	    if (curr_func->loc_types[loc.second] > Type::TC_object) {
	      if (get_token_info().get_type() != TK_DOT)
		PARSE_ERR("Expecting '.'");

	      curr_func->add_code(Opcodes::LOAD_LOCAL);
	      curr_func->add_uint4(loc.second);
	      
	      Token var = get_token_info();
	      // fixme what about params?
	      if (peek_token_info().get_type() == TK_LPAR) {
		get_token_info(); // the '('
		// get params
		if (get_token_info().get_type() != TK_RPAR) 
		  PARSE_ERR("Expecting ')'");

	      // push params first for f call

		//		uint4 f_index = prog.classes[loc.second].virt_func_ids[var.get_value()];
		uint4 f_index = loc.second;
		curr_func->add_code(Opcodes::CALL_VIRT);
		curr_func->add_uint4(f_index);
	      }
	      else 
		ERR("FIXME");
	    }
	    else if (curr_func->loc_types[loc.second] != Type::TC_sint4)
	      PARSE_ERR("Expecting a sint4");
	    curr_func->add_code(Opcodes::LOAD_LOCAL);
	    curr_func->add_uint4(loc.second);
	  }
	  else if (loc.first == VAR_PARAM) {
	    if (curr_func->par_types[loc.second] != Type::TC_sint4)
	      PARSE_ERR("Expecting a sint4");
	    type = Type::TC_sint4;
	    curr_func->add_code(Opcodes::LOAD_PARAM);
	    curr_func->add_uint4(loc.second);
	  }
	  else if (loc.first == VAR_INST_MEMBER) {
	    parse_class_var(loc.second, type);
	    // fixme class member
	  }
	  else PROGRAM_ERR;
	}
	break;
      
      default:
	ERR("FOO");
      }
      break;
    case Type::TC_sint1:
      switch(tk.get_type()) {
      case TK_ID:
	// either a variable or a class variable or an array variable or function
	break;
      case TK_INTC:
	if (valid_sint1(tk.get_value())) 
	  PARSE_ERR(tk.get_value() + " is not a valid uint1");
	curr_func->add_code(Opcodes::PUSH_CS4);
	curr_func->add_sint4(atoi(tk.get_value().c_str()));
	break;
      default:
	PARSE_ERR("expecting sint1");
      }
      break;

    case Type::TC_uint1:
      switch(tk.get_type()) {
      case TK_INTC:
      case TK_UINTC:
	if (!valid_uint1(tk.get_value()))
	  PARSE_ERR(tk.get_value() + " is not a valid uint1");
	curr_func->add_code(Opcodes::PUSH_CU4);
	curr_func->add_uint4(atoi(tk.get_value().c_str()));
	break;
      default: // fixme
	;
      }
      break;
      
    case Type::TC_sint2:
      break;

    case Type::TC_uint2:
      break;

    case Type::TC_sint4:

      switch(tk.get_type()) {
	
      case TK_ID:
	cout << "I'm here " << tk.get_value() << endl;
	// static class variable or function
	if (prog.class_ids.find(tk.get_value()) != 
	    prog.class_ids.end()) {
	  // fixme
	}
	// fixme hack
	else if (prog.extf_ids.find(tk.get_value()) !=
		 prog.extf_ids.end())
	  parse_extf_call(prog.extf_ids[tk.get_value()], false);
	else if (peek_token_info().get_type() == TK_LPAR) {
	  if (Type::TC_sint4 != parse_func_call(tk.get_value(), false, -1))
	    PARSE_ERR("Unmatched types");
	}
	else {
	  loc = find_variable(tk.get_value());
	  cout << "MY var type is " << loc.first << endl;
	  if (loc.first == VAR_LOCAL) {
	    cout << "WHY " << curr_func->loc_types[loc.second] << endl;
	    if (curr_func->loc_types[loc.second] >= Type::TC_object) {
	      if (prog.types[curr_func->loc_types[loc.second]].cat == Type::TC_array) {
		if (get_token_info().get_type() != TK_LBRACKET)
		  PARSE_ERR("Expecting '['");

		// fixme can be variable
		Token ind = get_token_info();
		if (ind.get_type() != TK_INTC && ind.get_type() != TK_UINTC)
		  PARSE_ERR("Expecting index");

		sint4 index = atoi(ind.get_value().c_str());

		if (get_token_info().get_type() != TK_RBRACKET)
		  PARSE_ERR("Expecting ']'");
		
		curr_func->add_code(Opcodes::LOAD_LOCAL);
		curr_func->add_uint4(loc.second);		
		curr_func->add_code(Opcodes::PUSH_CU4);
		curr_func->add_uint4((uint4)index);

		curr_func->add_code(Opcodes::LOAD_ARRAY);

	      }
	      else {
		uint4 class_index = prog.types[curr_func->loc_types[loc.second]].class_index;

		if (get_token_info().get_type() != TK_DOT)
		  PARSE_ERR("Expecting '.'");

		curr_func->add_code(Opcodes::LOAD_LOCAL);
		curr_func->add_uint4(loc.second);

		Token var = get_token_info();
		if (peek_token_info().get_type() == TK_LPAR) {
		  get_token_info(); // the '('
		  // get params
		  if (get_token_info().get_type() != TK_RPAR) 
		    PARSE_ERR("Expecting ')'");

		  // push params first for f call
		  
		  // fixme need to check for params
		  uint4 f_index = prog.classes[class_index].virt_func_ids[var.get_value()];

		  cout << var.get_value() << " " << class_index << endl;
		  cout << "&&&&&&&&&&&&&" << endl;
		  
		  curr_func->add_code(Opcodes::CALL_VIRT);
		  curr_func->add_uint4(f_index);
		}
		else {
		  uint4 class_index = prog.types[curr_func->loc_types[loc.second]].class_index;
		  if (prog.classes[class_index].inst_data_ids.find(var.get_value()) == prog.classes[class_index].inst_data_ids.end())
		    PARSE_ERR2("Unknown member", var.get_value());
		  
		  uint4 member_index = prog.classes[class_index].inst_data_ids[var.get_value()];
		  
		  curr_func->add_code(Opcodes::LOAD_INST);
		  curr_func->add_uint4(member_index);
		}
	      }
	    }
	    else {
	      if (curr_func->loc_types[loc.second] != Type::TC_sint4)
		PARSE_ERR("Expecting a sint4");
	      curr_func->add_code(Opcodes::LOAD_LOCAL);
	      curr_func->add_uint4(loc.second);
	      cout << "*************" << endl;
	    }
	  }
	  else if (loc.first == VAR_PARAM) {
	    cout << "HIOY " << loc.second <<  " " 
		 << curr_func->par_types[loc.second] << " "
		 << Type::TC_sint4 << endl;
	    if (curr_func->par_types[loc.second] != Type::TC_sint4)
	      PARSE_ERR("Expecting a sint4");
	    curr_func->add_code(Opcodes::LOAD_PARAM);
	    curr_func->add_uint4(loc.second);
	  }
	  else if (loc.first == VAR_INST_MEMBER) {
	    if (curr_class->inst_data[loc.second].type != Type::TC_sint4)
	      PARSE_ERR("Expecting a sint4");
	    curr_func->add_code(Opcodes::LOAD_PARAM);
	    curr_func->add_uint4(0);
	    curr_func->add_code(Opcodes::LOAD_INST);
	    curr_func->add_uint4(loc.second);

	    //	    parse_class_var(loc.second, type);
	    // fixme class member
	  }
	  else if (loc.first == VAR_STATIC_MEMBER) {
	    ERR("FIXME");
	  }
	  else PARSE_ERR("fixme");
	}
	break;

      case TK_INTC:
	if (!valid_sint4(tk.get_value()))
	  PARSE_ERR(tk.get_value() + " is not a valid sint4");
	curr_func->add_code(Opcodes::PUSH_CS4);
	curr_func->add_sint4(atoi(tk.get_value().c_str()));
	break;
	
      default: //fixme
	tk.write();
	ERR("FOO2");
      }
      break;
      
    case Type::TC_uint4:

      switch (tk.get_type()) {

      case TK_ID:
	// static class variable or function
	if (prog.class_ids.find(tk.get_value()) !=
	    prog.class_ids.end()) {
	  // fixme
	}
	// fixme check for correct return type
	else if (peek_token_info().get_type() == TK_LPAR) {
	  if (Type::TC_uint4 != parse_func_call(tk.get_value(), false, -1))
	    PARSE_ERR("Unmatched types");
	}
	else if (prog.extf_ids.find(tk.get_value()) !=
		 prog.extf_ids.end())
	  parse_extf_call(prog.extf_ids[tk.get_value()], false);
	else {
	  loc = find_variable(tk.get_value());
	  if (loc.first == VAR_LOCAL) {
	    if (curr_func->loc_types[loc.second] != Type::TC_uint4)
	      PARSE_ERR("Expecting a uint4");
	    curr_func->add_code(Opcodes::LOAD_LOCAL);
	    curr_func->add_uint4(loc.second);
	  }
	  else if (loc.first == VAR_PARAM) {
	    if (curr_func->par_types[loc.second] != Type::TC_uint4)
	      PARSE_ERR("Expecting a uint4");
	    curr_func->add_code(Opcodes::LOAD_PARAM);
	    curr_func->add_uint4(loc.second);
	  }
	  else if (loc.first == VAR_INST_MEMBER) {
	    // fixme class member
	    ;
	  }
	  else PROGRAM_ERR;
	}
	break;

	// fixme how do we convert string to uint4 ?
      case TK_INTC:
      case TK_UINTC:
	if (!valid_uint4(tk.get_value()))
	  PARSE_ERR(tk.get_value() + " is not a valid uint4");
	curr_func->add_code(Opcodes::PUSH_CU4);
	curr_func->add_uint4(atoi(tk.get_value().c_str()));
	break;
	
      default: //fixme
	tk.write();
	ERR("FOO");
      }     
      break;

    case Type::TC_sint8:
      switch(tk.get_type()) {
	
      case TK_ID:
	// static class variable or function
	if (prog.class_ids.find(tk.get_value()) != 
	    prog.class_ids.end()) {
	  // fixme
	}
	// fixme check for correct return type
	else if (peek_token_info().get_type() == TK_LPAR) {
	  if (Type::TC_sint8 != parse_func_call(tk.get_value(), false, -1))
	    PARSE_ERR("Unmatched types");
	}
	else if (prog.extf_ids.find(tk.get_value()) !=
		 prog.extf_ids.end())
	  parse_extf_call(prog.extf_ids[tk.get_value()], false);
	else {
	  loc = find_variable(tk.get_value());
	  if (loc.first == VAR_LOCAL) {
	    if (curr_func->loc_types[loc.second] != Type::TC_sint4)
	      PARSE_ERR("Expecting a sint8");
	    curr_func->add_code(Opcodes::LOAD_LOCAL);
	    curr_func->add_uint4(loc.second);
	  }
	  else if (loc.first == VAR_PARAM) {
	    if (curr_func->par_types[loc.second] != Type::TC_sint4)
	      PARSE_ERR("Expecting a sint8");
	    curr_func->add_code(Opcodes::LOAD_PARAM);
	    curr_func->add_uint4(loc.second);
	  }
	  else // fixme class member
	    ;
	}
	break;

      case TK_INTC:
      case TK_UINTC:
      case TK_LONGC:
	if (!valid_sint8(tk.get_value()))
	  PARSE_ERR(tk.get_value() + " is not a valid sint8");
	curr_func->add_code(Opcodes::PUSH_CS8);
	curr_func->add_sint8(atoll(tk.get_value().c_str()));
	break;
	
      default: //fixme
	tk.write();
	ERR("FOO2");
      }
      break;

    case Type::TC_uint8:
      switch (tk.get_type()) {

      case TK_ID:
	// static class variable or function
	if (prog.class_ids.find(tk.get_value()) !=
	    prog.class_ids.end()) {
	  // fixme
	}
	// fixme check for correct return type
	else if (peek_token_info().get_type() == TK_LPAR) {
	  if (Type::TC_uint8 != parse_func_call(tk.get_value(), false, -1))
	    PARSE_ERR("Unmatched types");
	}
	else if (prog.extf_ids.find(tk.get_value()) !=
		 prog.extf_ids.end())
	  parse_extf_call(prog.extf_ids[tk.get_value()], false);
	else {
	  loc = find_variable(tk.get_value());
	  if (loc.first == VAR_LOCAL) {
	    if (curr_func->loc_types[loc.second] != Type::TC_uint4)
	      PARSE_ERR("Expecting a uint8");
	    curr_func->add_code(Opcodes::LOAD_LOCAL);
	    curr_func->add_uint4(loc.second);
	  }
	  else if (loc.first == VAR_PARAM) {
	    if (curr_func->par_types[loc.second] != Type::TC_uint4)
	      PARSE_ERR("Expecting a uint8");
	    curr_func->add_code(Opcodes::LOAD_PARAM);
	    curr_func->add_uint4(loc.second);
	  }
	  else // fixme class member
	    ;
	}
	break;

	// fixme how do we convert string to uint8 ?
      case TK_INTC:
      case TK_UINTC:
	if (!valid_uint8(tk.get_value()))
	  PARSE_ERR(tk.get_value() + " is not a valid uint8");
	curr_func->add_code(Opcodes::PUSH_CU8);
	curr_func->add_uint8(atoll(tk.get_value().c_str()));
	break;
	
      default: //fixme
	tk.write();
	ERR("FOO");
      }     
      break;

    case Type::TC_real4:
      switch(tk.get_type()) {

      case TK_ID:
	// static class variable or function
	if (prog.class_ids.find(tk.get_value()) != 
	    prog.class_ids.end()) {
	  // fixme
	}
	// fixme check for correct return type
	else if (peek_token_info().get_type() == TK_LPAR) {
	  if (Type::TC_real4 != parse_func_call(tk.get_value(), false, -1))
	    PARSE_ERR("Unmatched types");
	}
	else if (prog.extf_ids.find(tk.get_value()) !=
		 prog.extf_ids.end())
	  parse_extf_call(prog.extf_ids[tk.get_value()], false);
	else {
	  loc = find_variable(tk.get_value());
	  if (loc.first == VAR_LOCAL) {
	    if (curr_func->loc_types[loc.second] != Type::TC_real4)
	      PARSE_ERR("Expecting a real4");
	    curr_func->add_code(Opcodes::LOAD_LOCAL);
	    curr_func->add_uint4(loc.second);
	  }
	  else if (loc.first == VAR_PARAM) {
	    if (curr_func->par_types[loc.second] != Type::TC_real4)
	      PARSE_ERR("Expecting a real4");
	    curr_func->add_code(Opcodes::LOAD_PARAM);
	    curr_func->add_uint4(loc.second);
	  }
	  else // fixme class member
	    ;
	}
	break;

      case TK_INTC:
      case TK_UINTC:
      case TK_LONGC:
      case TK_ULONGC:
      case TK_FLOATC:
	curr_func->add_code(Opcodes::PUSH_CR4);
	curr_func->add_real4(strtof(tk.get_value().c_str(), &cptr));
	break;
	
      default: //fixme
	tk.write();
	ERR("FOO5");
      }
      break;

    case Type::TC_real8:
      switch(tk.get_type()) {

      case TK_ID:
	// static class variable or function
	if (prog.class_ids.find(tk.get_value()) != 
	    prog.class_ids.end()) {
	  // fixme
	}
	// fixme check for correct return type
	else if (peek_token_info().get_type() == TK_LPAR) {
	  if (Type::TC_real8 != parse_func_call(tk.get_value(), false, -1))
	    PARSE_ERR("Unmatched types");
	}
	else if (prog.extf_ids.find(tk.get_value()) !=
		 prog.extf_ids.end())
	  parse_extf_call(prog.extf_ids[tk.get_value()], false);
	else {
	  loc = find_variable(tk.get_value());
	  if (loc.first == VAR_LOCAL) {
	    if (curr_func->loc_types[loc.second] != Type::TC_real8)
	      PARSE_ERR("Expecting a real8");
	    curr_func->add_code(Opcodes::LOAD_LOCAL);
	    curr_func->add_uint4(loc.second);
	  }
	  else if (loc.first == VAR_PARAM) {
	    if (curr_func->par_types[loc.second] != Type::TC_real8)
	      PARSE_ERR("Expecting a real8");
	    curr_func->add_code(Opcodes::LOAD_PARAM);
	    curr_func->add_uint4(loc.second);
	  }
	  else // fixme class member
	    ;
	}
	break;
	
      case TK_INTC:
      case TK_UINTC:
      case TK_LONGC:
      case TK_ULONGC:
      case TK_FLOATC:
      case TK_DOUBLEC:
	curr_func->add_code(Opcodes::PUSH_CR8);
	curr_func->add_real8(strtod(tk.get_value().c_str(), &cptr));
	break;
	
      default: //fixme
	tk.write();
	ERR("FOO5");
      }
      break;

    case Type::TC_class:
      break;
    case Type::TC_array:
      break;
    case Type::TC_string:
      switch(tk.get_type()) {
      case TK_ID:
	// static class variable or function
	if (prog.class_ids.find(tk.get_value()) !=
	    prog.class_ids.end()) {
	  // fixme
	}
	// fixme hack
	else if (prog.extf_ids.find(tk.get_value()) !=
		 prog.extf_ids.end())
	  parse_extf_call(prog.extf_ids[tk.get_value()], false);
	else if (peek_token_info().get_type() == TK_LPAR) {
	  if (Type::TC_string != parse_func_call(tk.get_value(), false, -1))
	    PARSE_ERR("Unmatched types");
	}
	else {
	  loc = find_variable(tk.get_value());
	  if (loc.first == VAR_LOCAL) {
	    if (curr_func->loc_types[loc.second] >= Type::TC_object) {
	      if (prog.types[curr_func->loc_types[loc.second]].cat == Type::TC_array) {
		if (get_token_info().get_type() != TK_LBRACKET)
		  PARSE_ERR("Expecting '['");
		
		// fixme can be variable
		Token ind = get_token_info();
		if (ind.get_type() != TK_INTC && ind.get_type() != TK_UINTC)
		  PARSE_ERR("Expecting index");
		
		sint4 index = atoi(ind.get_value().c_str());
		
		if (get_token_info().get_type() != TK_RBRACKET)
		  PARSE_ERR("Expecting ']'");
		
		curr_func->add_code(Opcodes::LOAD_LOCAL);
		curr_func->add_uint4(loc.second);		
		curr_func->add_code(Opcodes::PUSH_CU4);
		curr_func->add_uint4((uint4)index);
		
		curr_func->add_code(Opcodes::LOAD_ARRAY);
		
	      }
	      else {
		uint4 class_index = prog.types[curr_func->loc_types[loc.second]].class_index;
		
		if (get_token_info().get_type() != TK_DOT)
		  PARSE_ERR("Expecting '.'");
		
		curr_func->add_code(Opcodes::LOAD_LOCAL);
		curr_func->add_uint4(loc.second);
		
		Token var = get_token_info();
		if (peek_token_info().get_type() == TK_LPAR) {
		  get_token_info(); // the '('
		  // get params
		  if (get_token_info().get_type() != TK_RPAR) 
		    PARSE_ERR("Expecting ')'");
		  
		  // push params first for f call
		
		  // fixme need to check for params
		  uint4 f_index = prog.classes[class_index].virt_func_ids[var.get_value()];
		  curr_func->add_code(Opcodes::CALL_VIRT);
		  curr_func->add_uint4(f_index);
		}
		else {
		  uint4 class_index = prog.types[curr_func->loc_types[loc.second]].class_index;
		  if (prog.classes[class_index].inst_data_ids.find(var.get_value()) == prog.classes[class_index].inst_data_ids.end())
		    PARSE_ERR2("Unknown member", var.get_value());
		
		  uint4 member_index = prog.classes[class_index].inst_data_ids[var.get_value()];
		
		  curr_func->add_code(Opcodes::LOAD_INST);
		  curr_func->add_uint4(member_index);
		}
	      }
	    }
	    else {
	      if (curr_func->loc_types[loc.second] != Type::TC_string)
		PARSE_ERR("Expecting a string");
	      curr_func->add_code(Opcodes::LOAD_LOCAL);
	      curr_func->add_uint4(loc.second);
	      cout << "*************" << endl;
	    }
	  }
	  else if (loc.first == VAR_PARAM) {
	    if (curr_func->par_types[loc.second] != Type::TC_string)
	      PARSE_ERR("Expecting a string");
	    curr_func->add_code(Opcodes::LOAD_PARAM);
	    curr_func->add_uint4(loc.second);
	  }
	  else if (loc.first == VAR_INST_MEMBER) {
	    if (curr_class->inst_data[loc.second].type != Type::TC_string)
	      PARSE_ERR("Expecting a string");
	    curr_func->add_code(Opcodes::LOAD_PARAM);
	    curr_func->add_uint4(0);
	    curr_func->add_code(Opcodes::LOAD_INST);
	    curr_func->add_uint4(loc.second);

	    //	    parse_class_var(loc.second, type);
	    // fixme class member
	  }
	  else if (loc.first == VAR_STATIC_MEMBER) {
	    ERR("FIXME");
	  }
	  else PARSE_ERR("Fixme");
	}
	break;

      case TK_STRINGC:
	curr_func->add_code(Opcodes::PUSH_CARR);
	curr_func->add_uint4(tk.get_value().length());
	FORU(i, tk.get_value().length())
	  curr_func->add_uint4(tk.get_value()[i]);
	break;
      default: //fixme
	tk.write();
	ERR("FOO2");
      }
      break;
      
    default:
      break;
    }
    
    while (!unary.empty()) {
      curr_func->add_code(unary.top());
      unary.pop();
    }
  }

  void Parser::parse_class_var(uint4 loc, TypeIndex &expected_type) {
#if DEBUG
    DEBUGSTR << "parse_class_var(uint4 loc, TypeIndex &expected_type)" << endl
	     << "\tParams: *" << loc << "*" << expected_type << "*" << endl;
#endif

    // Load this pointer
    curr_func->add_code(Opcodes::LOAD_PARAM);
    curr_func->add_uint4(0);
    
    Token tk = get_token_info();

    if (tk.get_type() == TK_DOT) {
      ERR("NOT IMP YET");
    }
    else if (tk.get_type() == TK_ASSIGN) {
      parse_exp(TK_SEMICOLON, expected_type);
      switch(expected_type) {
      case Type::TC_sint1: curr_func->add_code(Opcodes::STORE_IS1); break;
      case Type::TC_uint1: curr_func->add_code(Opcodes::STORE_IU1); break;
      case Type::TC_sint2: curr_func->add_code(Opcodes::STORE_IS2); break;
      case Type::TC_uint2: curr_func->add_code(Opcodes::STORE_IU2); break;
      case Type::TC_sint4: curr_func->add_code(Opcodes::STORE_IS4); break;
      case Type::TC_uint4: curr_func->add_code(Opcodes::STORE_IU4); break;
      case Type::TC_sint8: curr_func->add_code(Opcodes::STORE_IS8); break;
      case Type::TC_uint8: curr_func->add_code(Opcodes::STORE_IU8); break;
      case Type::TC_real4: curr_func->add_code(Opcodes::STORE_IR4); break;
      case Type::TC_real8: curr_func->add_code(Opcodes::STORE_IR8); break;
      case Type::TC_string:
      default: // fixme
	ERR("FOO4");	
	;
      }
      curr_func->add_uint4(loc);
    } 
  }

  sint4 Parser::get_precedence_of(const sint4 &op)
  {
    if (op == TK_OROR)    return 1;
    if (op == TK_ANDAND)  return 2;
    if (op == TK_OR)      return 3;
    if (op == TK_XOR)     return 4;
    if (op == TK_AND)     return 5;
    if (op == TK_EQ || 
	op == TK_NEQ)     return 6;
    if (op == TK_LT || 
	op == TK_LEQ ||
	op == TK_GT || 
	op == TK_GEQ)     return 7;
    if (op == TK_LSHIFT ||
	op == TK_RSHIFT)  return 8;
    if (op == TK_ADD || 
	op == TK_SUB)     return 9;
    if (op == TK_MUL || 
	op == TK_DIV ||
	op == TK_MOD)     return 10;
    
    return 0;
  }

  inline bool Parser::valid_type(const TokenType &t) {
    return 
      t == TK_VOID   ||
      t == TK_BOOL   ||
      t == TK_BYTE   ||
      t == TK_UBYTE  ||
      t == TK_SHORT  ||
      t == TK_USHORT ||
      t == TK_INT    ||
      t == TK_UINT   ||
      t == TK_LONG   ||
      t == TK_ULONG  ||
      t == TK_FLOAT  ||
      t == TK_DOUBLE ||
      t ==TK_STRING;
  }

  inline bool Parser::valid_sint1(const string &s) {
    sint4 num = atoi(s.c_str());
    return num > min_sint1 && num < max_sint1;
  }
  
  inline bool Parser::valid_uint1(const string &s) {
    sint4 num = atoi(s.c_str());
    return num > min_uint1 && num < max_uint1;
  }

  inline bool Parser::valid_sint2(const string &s) {
    sint4 num = atoi(s.c_str());
    return num > min_sint2 && num < max_sint2;
  }

  inline bool Parser::valid_uint2(const string &s) {
    sint4 num = atoi(s.c_str());
    return num > min_uint2 && num < max_uint2;
  }

  inline bool Parser::valid_sint4(const string &s) {
    sint8 num = atoll(s.c_str());
    return num > min_sint4 && num < max_sint4;
  }

  inline bool Parser::valid_uint4(const string &s) {
    sint8 num = atoll(s.c_str());
    return num > min_uint4 && num < max_uint4;
  }

  inline bool Parser::valid_sint8(const string &s) {
    real8 num = atof(s.c_str());
    return num > min_sint8 && num < max_sint8;
  }

  inline bool Parser::valid_uint8(const string &s) {
    real8 num = atof(s.c_str());
    return num > min_uint8 && num > max_uint8;
  }
}
