/*
 * $Id: graphAbstraction.h,v 1.1 2006/12/31 04:09:40 orts_furtak Exp $
 *
 *  graphAbstraction.h
 *  HOG
 *
 *  Created by Nathan Sturtevant on 1/11/05.
 *  Copyright 2005 Nathan Sturtevant. All rights reserved.
 *
 */

#include "graph.h"

#ifndef graphAbstraction_H
#define graphAbstraction_H
#include <vector>
#ifdef OS_MAC
#include <OpenGL/gl.h>
#else
#include <GL/glut.h>
#endif
#include "constants.h"

/**
 * A generic class for basic operations on graph abstractions.
 */
namespace HOG{

class graphAbstraction {
public:
	graphAbstraction() :abstractions() {}
	virtual ~graphAbstraction() {}

	// basic graph functions
	/** is there a legal path between these 2 nodes? */
  virtual bool pathable(node *from, node *to) = 0;
	/** given 2 nodes, find as much of their hierarchy that exists in the graph */
  void getParentHierarchy(node *from, node *to, std::vector<node *> &fromChain, std::vector<node *> &toChain);
	/** return the abstract graph at the given level */
  graph* getAbstractGraph(int level) { return abstractions[level]; }
	/** return the total number of graphs in the hierarchy */
  unsigned int getNumAbstractGraphs() { return abstractions.size(); }
	/** heuristic cost between any two nodes */
  virtual double h(node *a, node *b) = 0;
	/** return nth level parent of which or null if it doesn't exist */
	node *getNthParent(node *which, int n);

	inline node *getParent(node *which) { return abstractions[getAbstractionLevel(which)+1]->getNode(which->getLabelL(kParent)); }
	inline long getNumChildren(node *which) { return which->getLabelL(kNumAbstractedNodes); }
	inline node *getNthChild(node *which, int n) { return abstractions[getAbstractionLevel(which)-1]->getNode(which->getLabelL(kFirstData+n)); }
	inline long getAbstractionLevel(node *which) { return which->getLabelL(kAbstractionLevel); }
	inline graph* getAbstractGraph(node *which) { return abstractions[which->getLabelL(kAbstractionLevel)]; }
	// utility functions
	/** verify that the hierarchy is consistent */
	virtual void verifyHierarchy() = 0;
	/// rebuild hierarchy from original domain */
  // virtual void rebuild() = 0;
	/// get current revision of hierarchy -- indicates if changes have been made */
	//virtual int getRevision() = 0;

	// hierarchical modifications
	/** remove node from abstraction */
	virtual void removeNode(node *n) = 0;
	/** remove edge from abstraction */
  virtual void removeEdge(edge *e, unsigned int absLevel) = 0;
	/** add node to abstraction */
	virtual void addNode(node *n) = 0;
	/** add edge to abstraction */
	virtual void addEdge(edge *e, unsigned int absLevel) = 0;
	/** This must be called after any of the above add/remove operations. But the
		operations can be stacked followed by a single repairAbstraction call. */
  virtual void repairAbstraction() = 0;

protected:
		std::vector<graph *> abstractions;
private:

};

// this class can be defined if you want to solve multiple domains...
// 
// class domainAbstraction : public graphAbstraction {
// 	void domainAbstraction(abstractableDomain *) = 0;
// };
}
#endif
