/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "CompActStd.H"

Action CompActStd::action(const vector<Prop> &vicinity)
{
  Action action;

  uint4 n = vicinity.size();
  if (n == 0) ERR("vicinity.size=0");

  const Prop &f = vicinity[0];

#if 1

  // if there are enemies in attack range, attack the weakest

  sint4 min_hp = max_sint4, min_i = -1;
  sint4 self = vicinity[0].sv_owner;
  
  for (uint4 i=1; i < n; i++) {
    const Prop &fi = vicinity[i];
    if (self != fi.sv_owner) {
      sint8 dx = fi.sv_pos_x - f.sv_pos_x;
      sint8 dy = fi.sv_pos_y - f.sv_pos_y;	
      sint8 dist2 = dx*dx+dy*dy;
      sint8 maxd = f.sv_max_attack_range + fi.sv_radius;
      sint8 mind = f.sv_min_attack_range - fi.sv_radius;
      if (mind < 0) mind = 0;

      // dist >= min_attack_range - r &&
      // dist <= max_attack_range + r

      if (dist2 >= (mind*mind) && dist2 <= (maxd*maxd)) {

	// cout << "orig: " << dist2 << " " << (mind*mind) << " " << (maxd*maxd) << " " << endl;

	if (fi.sv_hit_points < min_hp) {
	  min_hp = fi.sv_hit_points;
	  min_i = i;
	}
      }
    }
  }

  if (min_i >= 0) {
    action.action = Action::ATTACK;
    action.target_id = vicinity[min_i].sv_id;
    return action;
  }

#endif

#if 1
    
  vector<Prop> enemies;
    
  for (uint4 i=1; i < n; i++) {
    const Prop &fi = vicinity[i];
    if (self != fi.sv_owner) enemies.push_back(fi);
  }

  sint4 mini = -1;
  real8 min_dx=0, min_dy=0;
  sint8 mind = max_sint8;

  FORU (i, enemies.size()) {
    sint8 dx = enemies[i].sv_pos_x - f.sv_pos_x;
    sint8 dy = enemies[i].sv_pos_y - f.sv_pos_y;
    sint8 dist2 = dx*dx+dy*dy;
    if (dist2 < mind) {
      mini = i;
      real8 l = sqrt(real8(dx*dx+dy*dy));
      min_dx = dx/l;
      min_dy = dy/l;	
      mind = dist2;
    }
  }

  if (mini >= 0) {

    sint4 best_i = -1;
    real8 best_diff = 1e200;
      
    for (sint4 i=0; i < Action::DIR_N; i++) {
      real8 angle = (2.0*M_PI*i)/Action::DIR_N;
      real8 dx = ::cos(angle); if (fabs(dx) < .001) dx = 0.0;
      real8 dy = ::sin(angle); if (fabs(dy) < .001) dy = 0.0;
      real8 ddx = min_dx-dx;
      real8 ddy = min_dy-dy;
      real8 diff = square(ddx) + square(ddy);
      if (diff < best_diff) { best_diff = diff; best_i = i; }
    }

    action.action = Action::MOVE;

#if 1
    // randomize direction a little to decrease congestion

    if ((random() % 4) == 0) {
      best_i += (random() % 3) - 1;
      if (best_i < 0) best_i = Action::DIR_N-1;
      if (best_i >= Action::DIR_N) best_i = 0;
    }
#endif
    
    action.move_direction = best_i;
    return action;
  }
   
#endif

#if 1
  {
    vector<Prop> friends;
    
    for (uint4 i=1; i < n; i++) {
      const Prop &fi = vicinity[i];
      if (self == fi.sv_owner) friends.push_back(fi);
    }

    sint4 mini = -1;
    real8 min_dx=0, min_dy=0;
    sint8 mind = max_sint8;

    FORU (i, friends.size()) {
      sint8 dx = friends[i].sv_pos_x - f.sv_pos_x;
      sint8 dy = friends[i].sv_pos_y - f.sv_pos_y;
      sint8 dist2 = dx*dx+dy*dy;
      if (dist2 < mind) {
	mini = i;
	real8 l = sqrt(real8(dx*dx+dy*dy));
	min_dx = dx/l;
	min_dy = dy/l;	
	mind = dist2;
      }
    }

    if (mini >= 0) {

      sint4 best_i = -1;
      real8 best_diff = 1e200;
      
      for (sint4 i=0; i < Action::DIR_N; i++) {
	real8 angle = (2.0*M_PI*i)/Action::DIR_N;
	real8 dx = ::cos(angle); if (fabs(dx) < .001) dx = 0.0;
	real8 dy = ::sin(angle); if (fabs(dy) < .001) dy = 0.0;
	real8 ddx = min_dx-dx;
	real8 ddy = min_dy-dy;	
	real8 diff = square(ddx) + square(ddy);
	if (diff < best_diff) { best_diff = diff; best_i = i; }
      }

      action.action = Action::MOVE;
      sint8 desired_dist2 = (f.sv_max_attack_range + rand() % f.sv_max_attack_range)/2;
      if (mind < square(desired_dist2)) {
	best_i += Action::DIR_N/2;
	best_i %= Action::DIR_N;
      }

#if 1
      // randomize direction a little to decrease congestion
      best_i += (random() % 3) - 1;
      if (best_i < 0) best_i = Action::DIR_N-1;
      if (best_i >= Action::DIR_N) best_i = 0;
#endif
    
      action.move_direction = best_i;
      return action;
    }
  }
#endif
  
  if (!f.sv_moving) {
    // pick random direction
    action.action = Action::MOVE;
    action.move_direction = random() % Action::DIR_N;
  }

  return action;
}
