/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Options.H"

bool OptT<bool>::from_string(const std::string &s)
{
  if (s.empty()) return true;
  if (s[0] == '0') value = false;
  else if (s[0] == '1') value = true;
  else return true;
  return false;
}

bool OptT<sint4>::from_string(const std::string &s)
{
  sint4 r = sscanf(s.c_str(), "%d", &value);
  return r == 0 || r == EOF;
}

bool OptT<real8>::from_string(const std::string &s)
{
  sint4 r = sscanf(s.c_str(), "%lf", &value);
  return r == 0 || r == EOF;
}

bool OptT<string>::from_string(const std::string &s)
{
  value = s;
  return false;
}

bool Options::process(sint4 argc, char **argv, ostream &oserr, const char *remark)
{
  for (sint4 i=1; i < argc; i += 2) {

    string key = argv[i];

    OMapIt oit = omap.find(key);

    if (oit == omap.end()) {
      if (remark) oserr << remark << endl;
      oserr << "options: unknown option " << key << endl;
      oserr << "choices are:" << endl;
      FORALL (OMap, omap, i) {
	form(oserr, "%8s : %s\n", (*i).first.c_str(), (*i).second->remark.c_str());
      }
      return true;
    }

    if (i == argc-1) { oserr << "options: missing value" << endl; return true; }

    string val = argv[i+1];

    Opt *opt = oit->second;
    if (opt->from_string(val)) {
      oserr << "options: value syntax error for option "
	    << key << " (" << val << ")" << endl;
      return true;
    }
  }

  return false;
}

